#############################################################################
##
##  This file is part of recog, a package for the GAP computer algebra system
##  which provides a collection of methods for the constructive recognition
##  of groups.
##
##  This files's authors include Max Neunhöffer, Ákos Seress.
##
##  Copyright of recog belongs to its developers whose names are too numerous
##  to list here. Please refer to the COPYRIGHT file for details.
##
##  SPDX-License-Identifier: GPL-3.0-or-later
##
##
##  Code to recognise (simple) groups by their two largest element orders.
##  At least recognise the "natural" characteristic.
##
#############################################################################

# If you change the following, rerun RECOG.AddInformationAboutDegree below
# to recreate RECOG.TableOfTwoLargestGroupElementOrdersWithDegreeBound!

RECOG.TableOfTwoLargestGroupElementOrders :=
[ [3,2,["l",2,5]],[4,3,["l",2,7]],[5,3,["l",2,4]],
  [5,3,["l",2,5]],[5,3,["s",4,2]],[5,4,["l",2,9]],
  [5,4,["s",4,2]],[5,4,["s",4,3]],[6,5,["l",2,11]],
  [6,5,["s",4,2]],[6,5,["s",4,3]],[7,4,["l",2,7]],
  [7,4,["l",3,2]],[7,5,["l",3,4]],[7,6,["l",2,13]],
  [8,7,["u",3,3]],[8,7,["u",3,5]],[8,7,["u",4,3]],
  [9,5,["s",4,3]],[9,5,["u",4,2]],[9,6,["s",4,3]],
  [9,6,["u",4,2]],[9,7,["l",2,8]],[9,8,["l",2,17]],
  [9,8,["u",4,3]],[10,8,["u",3,5]],
  [10,9,["l",2,19]],[11,6,["l",2,11]],
  [11,8,["M11"]],
  [11,8,["M22"]],
  [11,10,["M12"]],
  [12,5,["s",4,3]],[12,6,["s",4,3]],
  [12,8,["u",3,3]],[12,8,["u",4,3]],
  [12,9,["s",4,3]],[12,9,["u",4,2]],
  [12,9,["u",4,3]],[12,11,["l",2,23]],
  [13,5,["2F4",2]],[13,6,["2F4",2]],[13,7,["2B2",8]],
  [13,7,["l",2,13]],[13,8,["2F4",2]],[13,8,["G2",3]],
  [13,8,["l",3,3]],[13,9,["G2",3]],[13,10,["2F4",2]],
  [13,12,["2F4",2]],[13,12,["G2",3]],
  [13,12,["l",2,25]],[13,12,["s",4,5]],
  [14,13,["l",2,27]],[15,7,["l",4,2]],
  [15,9,["o+",8,2]],[15,9,["s",6,2]],
  [15,10,["o+",8,2]],[15,10,["s",6,2]],
  [15,11,["u",5,2]],[15,11,["u",6,2]],
  [15,12,["o+",8,2]],[15,12,["s",6,2]],
  [15,12,["u",5,2]],[15,12,["u",6,2]],
  [15,12,["J2"]],
  [15,13,["s",4,5]],[15,13,["u",3,4]],
  [15,14,["l",2,29]],[16,13,["2F4",2]],
  [16,15,["l",2,31]],[17,9,["l",2,17]],
  [17,15,["l",2,16]],[17,15,["s",4,4]],
  [18,15,["u",5,2]],[18,15,["u",6,2]],
  [19,10,["l",2,19]],
  [19,15,["J1"]],
  [19,16,["l",3,7]],
  [19,17,["J3"]],
  [19,18,["l",2,37]],[20,13,["l",4,3]],
  [20,13,["s",4,5]],[20,14,["o",7,3]],
  [20,14,["o+",8,3]],[20,14,["s",6,3]],
  [20,15,["o",7,3]],[20,15,["o+",8,3]],
  [20,15,["s",4,5]],[20,15,["s",6,3]],
  [20,15,["HS"]],
  [20,18,["o",7,3]],[20,18,["o+",8,3]],
  [20,18,["s",6,3]],[21,13,["3D4",2]],
  [21,14,["3D4",2]],[21,15,["G2",4]],[21,17,["F4",2]],
  [21,17,["o-",8,2]],[21,17,["s",8,2]],
  [21,18,["3D4",2]],[21,18,["F4",2]],
  [21,18,["s",8,2]],[21,19,["u",3,8]],
  [21,20,["F4",2]],[21,20,["l",2,41]],
  [21,20,["s",8,2]],[22,21,["l",2,43]],
  [23,12,["l",2,23]],
  [23,15,["M23"]],
  [23,21,["M24"]],
  [24,20,["s",6,3]],
  [24,21,["F4",2]],[24,21,["s",8,2]],
  [24,21,["Suz"]],
  [24,23,["l",2,47]],[25,24,["l",2,49]],
  [25,24,["s",4,7]],[27,26,["l",2,53]],
  [28,21,["3D4",2]],[28,21,["F4",2]],
  [28,21,["He"]],
  [28,24,["F4",2]],
  [28,25,["s",4,7]],[29,15,["l",2,29]],
  [29,26,["Ru"]],
  [30,13,["s",4,5]],[30,15,["s",4,5]],
  [30,15,["McL"]],
  [30,20,["s",4,5]],[30,20,["s",6,3]],
  [30,21,["F4",2]],[30,21,["o-",8,2]],
  [30,21,["s",8,2]],[30,24,["F4",2]],
  [30,24,["s",6,3]],[30,24,["s",8,2]],
  [30,24,["Co3"]],
  [30,24,["Fi22"]],
  [30,28,["F4",2]],
  [30,28,["Co2"]],
  [30,29,["l",2,59]],
  [31,16,["l",2,31]],[31,21,["l",5,2]],
  [31,24,["G2",5]],[31,24,["l",3,5]],
  [31,25,["G2",5]],
  [31,28,["O'N"]],
  [31,30,["G2",5]],
  [31,30,["l",2,61]],[33,31,["l",2,32]],
  [34,33,["l",2,67]],[35,33,["2E6",2]],
  [35,33,["o-",10,2]],[36,20,["s",6,3]],
  [36,24,["s",6,3]],[36,30,["s",6,3]],
  [36,35,["l",2,71]],[37,19,["l",2,37]],
  [37,26,["2G2",27]],[37,36,["l",2,73]],
  [39,31,["l",4,5]],
  [39,36,["Th"]],
  [40,35,["HN"]],
  [40,37,["u",3,11]],
  [40,39,["l",2,79]],[41,21,["l",2,41]],
  [41,31,["2B2",32]],[41,40,["l",2,81]],
  [41,40,["s",4,9]],[42,25,["s",4,7]],
  [42,28,["s",4,7]],[42,41,["l",2,83]],
  [43,22,["l",2,43]],[44,40,["u",3,11]],
  [45,44,["l",2,89]],[47,24,["l",2,47]],
  [48,43,["u",3,7]],[49,48,["l",2,97]],
  [51,45,["o+",10,2]],[51,45,["s",10,2]],
  [51,50,["l",2,101]],[52,41,["o",9,3]],
  [52,41,["o-",8,3]],[52,41,["s",8,3]],
  [52,42,["o",9,3]],[52,42,["s",8,3]],
  [52,45,["o",9,3]],[52,45,["s",8,3]],
  [52,51,["l",2,103]],[53,27,["l",2,53]],
  [54,53,["l",2,107]],[55,54,["l",2,109]],
  [56,25,["s",4,7]],[56,28,["s",4,7]],
  [56,42,["s",4,7]],[56,48,["u",3,7]],
  [57,48,["G2",7]],[57,49,["G2",7]],[57,56,["G2",7]],
  [57,56,["l",2,113]],[59,30,["l",2,59]],
  [60,42,["Co1"]],
  [60,42,["Fi23"]],
  [60,45,["Fi24'"]],
  [60,51,["o+",10,2]],[60,51,["s",10,2]],
  [60,52,["o",9,3]],[60,52,["o-",8,3]],
  [60,52,["s",8,3]],[61,31,["l",2,61]],
  [61,56,["l",3,13]],[61,60,["l",2,121]],
  [61,60,["s",4,11]],[63,31,["l",6,2]],
  [63,45,["u",7,2]],[63,52,["u",4,5]],
  [63,60,["u",4,5]],[63,60,["u",7,2]],
  [63,62,["l",2,125]],[64,63,["l",2,127]],
  [65,51,["u",4,4]],[65,51,["u",5,4]],
  [65,63,["l",2,64]],[65,63,["s",4,8]],
  [66,44,["J4"]],
  [66,61,["s",4,11]],[66,63,["u",7,2]],
  [66,65,["l",2,131]],[67,34,["l",2,67]],
  [67,42,["Ly"]],
  [69,68,["l",2,137]],
  [70,66,["B"]],
  [70,69,["l",2,139]],
  [71,36,["l",2,71]],[72,52,["s",8,3]],
  [72,60,["s",8,3]],[73,37,["l",2,73]],
  [73,63,["G2",8]],[73,63,["l",3,8]],
  [75,74,["l",2,149]],[76,75,["l",2,151]],
  [78,52,["s",8,3]],[78,60,["s",8,3]],
  [78,63,["o",7,5]],[78,63,["s",6,5]],
  [78,65,["o",7,5]],[78,65,["s",6,5]],
  [78,72,["s",8,3]],[79,40,["l",2,79]],
  [79,78,["l",2,157]],[80,61,["u",5,3]],
  [80,65,["o-",10,3]],[80,73,["u",3,9]],
  [82,81,["l",2,163]],[83,42,["l",2,83]],
  [84,52,["s",8,3]],[84,60,["s",8,3]],
  [84,72,["s",8,3]],[84,78,["s",8,3]],
  [84,80,["o-",10,3]],[84,80,["u",5,3]],
  [84,83,["l",2,167]],[85,63,["l",4,4]],
  [85,65,["s",6,4]],[85,84,["l",2,169]],
  [85,84,["s",4,13]],[86,75,["u",4,7]],
  [86,84,["u",4,7]],[87,86,["l",2,173]],
  [89,45,["l",2,89]],[90,52,["s",8,3]],
  [90,60,["s",8,3]],[90,72,["s",8,3]],
  [90,78,["s",8,3]],[90,84,["s",8,3]],
  [90,89,["l",2,179]],[91,80,["l",3,9]],
  [91,85,["l",3,16]],[91,85,["s",4,13]],
  [91,90,["l",2,181]],[93,65,["o-",12,2]],
  [93,84,["o-",12,2]],[93,91,["E6",2]],
  [96,91,["u",3,17]],[96,95,["l",2,191]],
  [97,49,["l",2,97]],[97,96,["l",2,193]],
  [99,98,["l",2,197]],[100,99,["l",2,199]],
  [101,51,["l",2,101]],[102,93,["o-",12,2]],
  [102,96,["u",3,17]],[103,52,["l",2,103]],
  [104,73,["3D4",3]],[104,78,["3D4",3]],
  [104,84,["3D4",3]],[105,85,["o+",12,2]],
  [105,93,["s",12,2]],[105,102,["s",12,2]],
  [106,105,["l",2,211]],[107,54,["l",2,107]],
  [109,55,["l",2,109]],[110,61,["s",4,11]],
  [110,66,["s",4,11]],[112,111,["l",2,223]],
  [113,57,["l",2,113]],[114,113,["l",2,227]],
  [115,114,["l",2,229]],[117,116,["l",2,233]],
  [119,110,["M"]],
  [120,78,["s",6,5]],[120,105,["s",12,2]],
  [120,119,["l",2,239]],[121,104,["l",5,3]],
  [121,120,["l",2,241]],[122,91,["u",6,3]],
  [122,120,["u",6,3]],[122,121,["l",2,243]],
  [126,93,["E6",2]],[126,125,["l",2,251]],
  [127,64,["l",2,127]],[127,105,["l",7,2]],
  [127,120,["l",3,19]],[129,99,["u",8,2]],
  [129,99,["u",9,2]],[129,126,["u",8,2]],
  [129,126,["u",9,2]],[129,127,["l",2,128]],
  [129,128,["l",2,257]],[130,78,["s",6,5]],
  [130,120,["s",6,5]],[131,66,["l",2,131]],
  [132,61,["s",4,11]],[132,66,["s",4,11]],
  [132,110,["s",4,11]],[132,129,["u",8,2]],
  [132,129,["u",9,2]],[132,131,["l",2,263]],
  [133,120,["G2",11]],[133,120,["l",3,11]],
  [133,121,["G2",11]],[133,132,["G2",11]],
  [135,134,["l",2,269]],[136,135,["l",2,271]],
  [137,69,["l",2,137]],[139,70,["l",2,139]],
  [139,138,["l",2,277]],[141,140,["l",2,281]],
  [142,141,["l",2,283]],[145,127,["2B2",128]],
  [145,144,["l",2,289]],[145,144,["s",4,17]],
  [147,146,["l",2,293]],[149,75,["l",2,149]],
  [151,76,["l",2,151]],[153,145,["s",4,17]],
  [154,153,["l",2,307]],[156,78,["o+",8,5]],
  [156,85,["s",4,13]],[156,91,["s",4,13]],
  [156,155,["l",2,311]],[157,79,["l",2,157]],
  [157,156,["l",2,313]],[159,158,["l",2,317]],
  [163,82,["l",2,163]],[164,140,["o+",10,3]],
  [164,140,["s",10,3]],[164,156,["o+",10,3]],
  [164,156,["s",10,3]],[166,165,["l",2,331]],
  [167,84,["l",2,167]],[168,157,["u",3,13]],
  [169,168,["l",2,337]],[172,171,["l",2,343]],
  [173,87,["l",2,173]],[174,173,["l",2,347]],
  [175,174,["l",2,349]],[176,169,["u",3,23]],
  [177,176,["l",2,353]],[179,90,["l",2,179]],
  [180,164,["s",10,3]],[180,179,["l",2,359]],
  [181,91,["l",2,181]],[181,180,["l",2,361]],
  [181,180,["s",4,19]],[182,85,["s",4,13]],
  [182,91,["s",4,13]],[182,121,["l",6,3]],
  [182,156,["s",4,13]],[182,168,["u",3,13]],
  [183,168,["G2",13]],[183,169,["G2",13]],
  [183,182,["G2",13]],[184,176,["u",3,23]],
  [184,183,["l",2,367]],[187,186,["l",2,373]],
  [190,181,["s",4,19]],[190,189,["l",2,379]],
  [191,96,["l",2,191]],[192,191,["l",2,383]],
  [193,97,["l",2,193]],[195,165,["o+",14,2]],
  [195,186,["o+",14,2]],[195,194,["l",2,389]],
  [197,99,["l",2,197]],[199,100,["l",2,199]],
  [199,198,["l",2,397]],[200,171,["l",4,7]],
  [200,172,["o",7,7]],[200,172,["s",6,7]],
  [200,175,["o",7,7]],[200,175,["s",6,7]],
  [201,200,["l",2,401]],[204,195,["o+",14,2]],
  [205,182,["l",4,9]],[205,204,["l",2,409]],
  [210,209,["l",2,419]],[211,106,["l",2,211]],
  [211,210,["l",2,421]],[216,215,["l",2,431]],
  [217,208,["l",3,25]],[217,216,["l",2,433]],
  [220,219,["l",2,439]],[222,221,["l",2,443]],
  [223,112,["l",2,223]],[225,224,["l",2,449]],
  [227,114,["l",2,227]],[229,115,["l",2,229]],
  [229,228,["l",2,457]],[231,230,["l",2,461]],
  [232,231,["l",2,463]],[233,117,["l",2,233]],
  [234,164,["s",10,3]],[234,180,["s",10,3]],
  [234,233,["l",2,467]],[239,120,["l",2,239]],
  [240,164,["s",10,3]],[240,180,["s",10,3]],
  [240,234,["s",10,3]],[240,239,["l",2,479]],
  [241,121,["l",2,241]],[244,243,["l",2,487]],
  [246,164,["s",10,3]],[246,180,["s",10,3]],
  [246,234,["s",10,3]],[246,240,["s",10,3]],
  [246,245,["l",2,491]],[250,249,["l",2,499]],
  [251,126,["l",2,251]],[252,164,["s",10,3]],
  [252,180,["s",10,3]],[252,234,["s",10,3]],
  [252,240,["s",10,3]],[252,246,["s",10,3]],
  [252,251,["l",2,503]],[255,85,["o+",8,4]],
  [255,155,["o-",14,2]],[255,195,["s",14,2]],
  [255,204,["s",14,2]],[255,210,["o-",14,2]],
  [255,210,["s",14,2]],[255,217,["l",8,2]],
  [255,241,["u",3,16]],[255,254,["l",2,509]],
  [257,129,["l",2,257]],[257,255,["l",2,256]],
  [257,255,["s",4,16]],[261,260,["l",2,521]],
  [262,261,["l",2,523]],[263,132,["l",2,263]],
  [265,264,["l",2,529]],[265,264,["s",4,23]],
  [269,135,["l",2,269]],[271,136,["l",2,271]],
  [271,270,["l",2,541]],[272,145,["s",4,17]],
  [272,153,["s",4,17]],[273,255,["G2",16]],
  [274,273,["l",2,547]],[276,265,["s",4,23]],
  [277,139,["l",2,277]],[279,278,["l",2,557]],
  [280,271,["u",3,29]],[281,141,["l",2,281]],
  [282,281,["l",2,563]],[283,142,["l",2,283]],
  [285,284,["l",2,569]],[286,285,["l",2,571]],
  [289,288,["l",2,577]],[290,280,["u",3,29]],
  [293,147,["l",2,293]],[294,293,["l",2,587]],
  [297,296,["l",2,593]],[300,299,["l",2,599]],
  [301,300,["l",2,601]],[304,303,["l",2,607]],
  [306,145,["s",4,17]],[306,153,["s",4,17]],
  [306,272,["s",4,17]],[307,154,["l",2,307]],
  [307,288,["G2",17]],[307,288,["l",3,17]],
  [307,289,["G2",17]],[307,306,["G2",17]],
  [307,306,["l",2,613]],[309,308,["l",2,617]],
  [310,309,["l",2,619]],[311,156,["l",2,311]],
  [313,157,["l",2,313]],[313,312,["l",2,625]],
  [313,312,["s",4,25]],[315,241,["3D4",4]],
  [315,257,["o-",8,4]],[316,315,["l",2,631]],
  [317,159,["l",2,317]],[321,320,["l",2,641]],
  [322,321,["l",2,643]],[324,323,["l",2,647]],
  [327,326,["l",2,653]],[330,329,["l",2,659]],
  [331,166,["l",2,331]],[331,320,["l",3,31]],
  [331,330,["l",2,661]],[333,305,["u",4,11]],
  [333,330,["u",4,11]],[336,200,["s",6,7]],
  [337,169,["l",2,337]],[337,336,["l",2,673]],
  [339,338,["l",2,677]],[341,315,["l",5,4]],
  [341,331,["u",3,32]],[342,181,["s",4,19]],
  [342,190,["s",4,19]],[342,341,["l",2,683]],
  [346,345,["l",2,691]],[347,174,["l",2,347]],
  [349,175,["l",2,349]],[350,200,["s",6,7]],
  [350,336,["s",6,7]],[351,350,["l",2,701]],
  [353,177,["l",2,353]],[355,354,["l",2,709]],
  [359,180,["l",2,359]],[360,343,["u",3,19]],
  [360,359,["l",2,719]],[364,363,["l",2,727]],
  [365,328,["u",4,9]],[365,364,["l",2,729]],
  [365,364,["s",4,27]],[367,184,["l",2,367]],
  [367,366,["l",2,733]],[370,369,["l",2,739]],
  [372,313,["o-",8,5]],[372,313,["s",8,5]],
  [372,315,["s",8,5]],[372,371,["l",2,743]],
  [373,187,["l",2,373]],[376,375,["l",2,751]],
  [379,190,["l",2,379]],[379,378,["l",2,757]],
  [380,181,["s",4,19]],[380,190,["s",4,19]],
  [380,342,["s",4,19]],[380,360,["u",3,19]],
  [381,380,["l",2,761]],[383,192,["l",2,383]],
  [385,384,["l",2,769]],[387,386,["l",2,773]],
  [389,195,["l",2,389]],[390,372,["o-",8,5]],
  [390,372,["s",8,5]],[394,393,["l",2,787]],
  [397,199,["l",2,397]],[399,398,["l",2,797]],
  [401,201,["l",2,401]],[405,404,["l",2,809]],
  [406,405,["l",2,811]],[409,205,["l",2,409]],
  [410,365,["s",6,9]],[411,410,["l",2,821]],
  [412,411,["l",2,823]],[414,413,["l",2,827]],
  [415,414,["l",2,829]],[419,210,["l",2,419]],
  [420,419,["l",2,839]],[421,211,["l",2,421]],
  [421,420,["l",2,841]],[421,420,["s",4,29]],
  [427,426,["l",2,853]],[429,428,["l",2,857]],
  [430,429,["l",2,859]],[431,216,["l",2,431]],
  [432,431,["l",2,863]],[433,217,["l",2,433]],
  [435,421,["s",4,29]],[439,220,["l",2,439]],
  [439,438,["l",2,877]],[441,440,["l",2,881]],
  [442,441,["l",2,883]],[443,222,["l",2,443]],
  [444,443,["l",2,887]],[449,225,["l",2,449]],
  [454,453,["l",2,907]],[455,341,["l",6,4]],
  [456,455,["l",2,911]],[457,229,["l",2,457]],
  [460,459,["l",2,919]],[461,231,["l",2,461]],
  [463,232,["l",2,463]],[465,464,["l",2,929]],
  [467,234,["l",2,467]],[469,456,["l",3,37]],
  [469,468,["l",2,937]],[471,470,["l",2,941]],
  [474,473,["l",2,947]],[477,476,["l",2,953]],
  [479,240,["l",2,479]],[481,480,["l",2,961]],
  [481,480,["s",4,31]],[484,410,["s",12,3]],
  [484,420,["s",12,3]],[484,468,["s",12,3]],
  [484,483,["l",2,967]],[486,485,["l",2,971]],
  [487,244,["l",2,487]],[489,488,["l",2,977]],
  [491,246,["l",2,491]],[492,484,["s",12,3]],
  [492,491,["l",2,983]],[496,481,["s",4,31]],
  [496,495,["l",2,991]],[499,250,["l",2,499]],
  [499,498,["l",2,997]],[503,252,["l",2,503]],
  [505,504,["l",2,1009]],[506,265,["s",4,23]],
  [506,276,["s",4,23]],[507,506,["l",2,1013]],
  [509,255,["l",2,509]],[510,509,["l",2,1019]],
  [511,465,["l",9,2]],[511,510,["l",2,1021]],
  [513,387,["u",10,2]],[513,455,["u",4,8]],
  [513,510,["u",10,2]],[513,511,["l",2,512]],
  [516,513,["u",10,2]],[516,515,["l",2,1031]],
  [517,516,["l",2,1033]],[520,519,["l",2,1039]],
  [521,261,["l",2,521]],[523,262,["l",2,523]],
  [525,524,["l",2,1049]],[526,525,["l",2,1051]],
  [531,530,["l",2,1061]],[532,531,["l",2,1063]],
  [535,534,["l",2,1069]],[541,271,["l",2,541]],
  [544,543,["l",2,1087]],[546,545,["l",2,1091]],
  [547,274,["l",2,547]],[547,546,["l",2,1093]],
  [549,548,["l",2,1097]],[552,265,["s",4,23]],
  [552,276,["s",4,23]],[552,506,["s",4,23]],
  [552,551,["l",2,1103]],[553,528,["l",3,23]],
  [555,554,["l",2,1109]],[557,279,["l",2,557]],
  [559,558,["l",2,1117]],[560,547,["u",3,41]],
  [562,561,["l",2,1123]],[563,282,["l",2,563]],
  [565,564,["l",2,1129]],[569,285,["l",2,569]],
  [571,286,["l",2,571]],[574,560,["u",3,41]],
  [576,575,["l",2,1151]],[577,289,["l",2,577]],
  [577,576,["l",2,1153]],[582,581,["l",2,1163]],
  [585,511,["l",4,8]],[586,585,["l",2,1171]],
  [587,294,["l",2,587]],[591,590,["l",2,1181]],
  [593,297,["l",2,593]],[594,593,["l",2,1187]],
  [595,549,["l",4,13]],[597,596,["l",2,1193]],
  [599,300,["l",2,599]],[601,301,["l",2,601]],
  [601,600,["l",2,1201]],[607,304,["l",2,607]],
  [607,606,["l",2,1213]],[609,608,["l",2,1217]],
  [612,611,["l",2,1223]],[613,307,["l",2,613]],
  [615,614,["l",2,1229]],[616,615,["l",2,1231]],
  [617,309,["l",2,617]],[619,310,["l",2,619]],
  [619,618,["l",2,1237]],[620,372,["s",8,5]],
  [620,390,["s",8,5]],[624,521,["u",5,5]],
  [624,521,["u",6,5]],[624,601,["u",3,25]],
  [625,624,["l",2,1249]],[630,372,["s",8,5]],
  [630,390,["s",8,5]],[630,620,["s",8,5]],
  [630,624,["u",5,5]],[630,624,["u",6,5]],
  [630,629,["l",2,1259]],[631,316,["l",2,631]],
  [631,616,["l",3,43]],[639,638,["l",2,1277]],
  [640,639,["l",2,1279]],[641,321,["l",2,641]],
  [642,641,["l",2,1283]],[643,322,["l",2,643]],
  [645,644,["l",2,1289]],[646,645,["l",2,1291]],
  [647,324,["l",2,647]],[649,648,["l",2,1297]],
  [651,650,["l",2,1301]],[652,651,["l",2,1303]],
  [653,327,["l",2,653]],[654,653,["l",2,1307]],
  [659,330,["l",2,659]],[660,659,["l",2,1319]],
  [661,331,["l",2,661]],[661,660,["l",2,1321]],
  [664,663,["l",2,1327]],[666,665,["l",2,1331]],
  [673,337,["l",2,673]],[677,339,["l",2,677]],
  [681,680,["l",2,1361]],[683,342,["l",2,683]],
  [684,683,["l",2,1367]],[685,684,["l",2,1369]],
  [685,684,["s",4,37]],[687,686,["l",2,1373]],
  [691,346,["l",2,691]],[691,690,["l",2,1381]],
  [700,699,["l",2,1399]],[701,351,["l",2,701]],
  [703,685,["s",4,37]],[705,704,["l",2,1409]],
  [709,355,["l",2,709]],[712,711,["l",2,1423]],
  [714,713,["l",2,1427]],[715,714,["l",2,1429]],
  [717,716,["l",2,1433]],[719,360,["l",2,719]],
  [720,484,["s",12,3]],[720,492,["s",12,3]],
  [720,719,["l",2,1439]],[724,723,["l",2,1447]],
  [726,484,["s",12,3]],[726,492,["s",12,3]],
  [726,720,["s",12,3]],[726,725,["l",2,1451]],
  [727,364,["l",2,727]],[727,726,["l",2,1453]],
  [728,560,["u",7,3]],[728,560,["u",8,3]],
  [728,703,["u",3,27]],[730,729,["l",2,1459]],
  [732,484,["s",12,3]],[732,492,["s",12,3]],
  [732,665,["l",4,11]],[732,666,["s",6,11]],
  [732,671,["s",6,11]],[732,720,["s",12,3]],
  [732,726,["s",12,3]],[732,728,["u",7,3]],
  [732,728,["u",8,3]],[733,367,["l",2,733]],
  [736,721,["u",3,47]],[736,735,["l",2,1471]],
  [738,484,["s",12,3]],[738,492,["s",12,3]],
  [738,720,["s",12,3]],[738,726,["s",12,3]],
  [738,732,["s",12,3]],[739,370,["l",2,739]],
  [741,740,["l",2,1481]],[742,741,["l",2,1483]],
  [743,372,["l",2,743]],[744,601,["3D4",5]],
  [744,620,["3D4",5]],[744,630,["3D4",5]],
  [744,743,["l",2,1487]],[745,744,["l",2,1489]],
  [747,746,["l",2,1493]],[750,749,["l",2,1499]],
  [751,376,["l",2,751]],[752,736,["u",3,47]],
  [756,755,["l",2,1511]],[757,379,["l",2,757]],
  [757,728,["l",3,27]],[761,381,["l",2,761]],
  [762,761,["l",2,1523]],[766,765,["l",2,1531]],
  [769,385,["l",2,769]],[772,771,["l",2,1543]],
  [773,387,["l",2,773]],[775,774,["l",2,1549]],
  [777,776,["l",2,1553]],[780,779,["l",2,1559]],
  [781,744,["l",5,5]],[784,783,["l",2,1567]],
  [786,785,["l",2,1571]],[787,394,["l",2,787]],
  [790,789,["l",2,1579]],[792,791,["l",2,1583]],
  [797,399,["l",2,797]],[799,798,["l",2,1597]],
  [801,800,["l",2,1601]],[804,803,["l",2,1607]],
  [805,804,["l",2,1609]],[807,806,["l",2,1613]],
  [809,405,["l",2,809]],[810,809,["l",2,1619]],
  [811,406,["l",2,811]],[811,810,["l",2,1621]],
  [812,421,["s",4,29]],[812,435,["s",4,29]],
  [814,813,["l",2,1627]],[817,800,["l",3,49]],
  [819,818,["l",2,1637]],[821,411,["l",2,821]],
  [823,412,["l",2,823]],[827,414,["l",2,827]],
  [829,415,["l",2,829]],[829,828,["l",2,1657]],
  [832,831,["l",2,1663]],[834,833,["l",2,1667]],
  [835,834,["l",2,1669]],[839,420,["l",2,839]],
  [841,840,["l",2,1681]],[841,840,["s",4,41]],
  [847,846,["l",2,1693]],[849,848,["l",2,1697]],
  [850,849,["l",2,1699]],[853,427,["l",2,853]],
  [855,854,["l",2,1709]],[857,429,["l",2,857]],
  [859,430,["l",2,859]],[861,841,["s",4,41]],
  [861,860,["l",2,1721]],[862,861,["l",2,1723]],
  [863,432,["l",2,863]],[867,866,["l",2,1733]],
  [870,421,["s",4,29]],[870,435,["s",4,29]],
  [870,812,["s",4,29]],[871,840,["l",3,29]],
  [871,870,["l",2,1741]],[874,873,["l",2,1747]],
  [877,439,["l",2,877]],[877,876,["l",2,1753]],
  [880,879,["l",2,1759]],[881,441,["l",2,881]],
  [883,442,["l",2,883]],[887,444,["l",2,887]],
  [889,888,["l",2,1777]],[892,891,["l",2,1783]],
  [894,893,["l",2,1787]],[895,894,["l",2,1789]],
  [901,900,["l",2,1801]],[906,905,["l",2,1811]],
  [907,454,["l",2,907]],[911,456,["l",2,911]],
  [912,911,["l",2,1823]],[916,915,["l",2,1831]],
  [919,460,["l",2,919]],[924,923,["l",2,1847]],
  [925,924,["l",2,1849]],[925,924,["s",4,43]],
  [929,465,["l",2,929]],[930,481,["s",4,31]],
  [930,496,["s",4,31]],[931,930,["l",2,1861]],
  [934,933,["l",2,1867]],[936,919,["u",3,53]],
  [936,935,["l",2,1871]],[937,469,["l",2,937]],
  [937,936,["l",2,1873]],[939,938,["l",2,1877]],
  [940,939,["l",2,1879]],[941,471,["l",2,941]],
  [945,944,["l",2,1889]],[946,925,["s",4,43]],
  [947,474,["l",2,947]],[951,950,["l",2,1901]],
  [953,477,["l",2,953]],[954,936,["u",3,53]],
  [954,953,["l",2,1907]],[957,956,["l",2,1913]],
  [960,931,["u",3,31]],[966,965,["l",2,1931]],
  [967,484,["l",2,967]],[967,966,["l",2,1933]],
  [971,486,["l",2,971]],[975,974,["l",2,1949]],
  [976,975,["l",2,1951]],[977,489,["l",2,977]],
  [983,492,["l",2,983]],[987,986,["l",2,1973]],
  [990,989,["l",2,1979]],[991,496,["l",2,991]],
  [992,481,["s",4,31]],[992,496,["s",4,31]],
  [992,930,["s",4,31]],[992,960,["u",3,31]],
  [994,993,["l",2,1987]],[997,499,["l",2,997]],
  [997,996,["l",2,1993]],[999,998,["l",2,1997]],
  [1000,999,["l",2,1999]],[1002,1001,["l",2,2003]],
  [1006,1005,["l",2,2011]],[1009,505,["l",2,1009]],
  [1009,1008,["l",2,2017]],[1013,507,["l",2,1013]],
  [1014,1013,["l",2,2027]],[1015,1014,["l",2,2029]],
  [1019,510,["l",2,1019]],[1020,1019,["l",2,2039]],
  [1021,511,["l",2,1021]],[1023,765,["u",11,2]],
  [1023,765,["u",12,2]],[1023,774,["u",12,2]],
  [1023,889,["l",10,2]],[1023,1020,["u",11,2]],
  [1023,1020,["u",12,2]],[1025,819,["u",6,4]],
  [1025,1023,["l",2,1024]],[1026,1023,["u",11,2]],
  [1026,1023,["u",12,2]],[1027,1026,["l",2,2053]],
  [1031,516,["l",2,1031]],[1032,1031,["l",2,2063]],
  [1033,517,["l",2,1033]],[1035,1034,["l",2,2069]],
  [1039,520,["l",2,1039]],[1041,1040,["l",2,2081]],
  [1042,1041,["l",2,2083]],[1044,1043,["l",2,2087]],
  [1045,1044,["l",2,2089]],[1049,525,["l",2,1049]],
  [1050,1049,["l",2,2099]],[1051,526,["l",2,1051]],
  [1056,1055,["l",2,2111]],[1057,1023,["l",3,32]],
  [1057,1056,["l",2,2113]],[1061,531,["l",2,1061]],
  [1063,532,["l",2,1063]],[1065,1064,["l",2,2129]],
  [1066,1065,["l",2,2131]],[1069,535,["l",2,1069]],
  [1069,1068,["l",2,2137]],[1071,1070,["l",2,2141]],
  [1072,1071,["l",2,2143]],[1077,1076,["l",2,2153]],
  [1081,1080,["l",2,2161]],[1087,544,["l",2,1087]],
  [1090,1089,["l",2,2179]],[1091,546,["l",2,1091]],
  [1093,547,["l",2,1093]],[1093,1040,["l",7,3]],
  [1094,1093,["l",2,2187]],[1097,549,["l",2,1097]],
  [1099,1020,["u",4,13]],[1099,1092,["u",4,13]],
  [1099,1098,["l",2,2197]],[1102,1101,["l",2,2203]],
  [1103,552,["l",2,1103]],[1104,1103,["l",2,2207]],
  [1105,1104,["l",2,2209]],[1105,1104,["s",4,47]],
  [1107,1106,["l",2,2213]],[1109,555,["l",2,1109]],
  [1111,1110,["l",2,2221]],[1117,559,["l",2,1117]],
  [1119,1118,["l",2,2237]],[1120,1119,["l",2,2239]],
  [1122,1121,["l",2,2243]],[1123,562,["l",2,1123]],
  [1126,1125,["l",2,2251]],[1128,1105,["s",4,47]],
  [1129,565,["l",2,1129]],[1134,1133,["l",2,2267]],
  [1135,1134,["l",2,2269]],[1137,1136,["l",2,2273]],
  [1141,1140,["l",2,2281]],[1144,1143,["l",2,2287]],
  [1147,1146,["l",2,2293]],[1149,1148,["l",2,2297]],
  [1151,576,["l",2,1151]],[1153,577,["l",2,1153]],
  [1155,1154,["l",2,2309]],[1156,1155,["l",2,2311]],
  [1160,1141,["u",3,59]],[1163,582,["l",2,1163]],
  [1167,1166,["l",2,2333]],[1170,1169,["l",2,2339]],
  [1171,586,["l",2,1171]],[1171,1170,["l",2,2341]],
  [1174,1173,["l",2,2347]],[1176,1175,["l",2,2351]],
  [1179,1178,["l",2,2357]],[1180,1160,["u",3,59]],
  [1181,591,["l",2,1181]],[1186,1185,["l",2,2371]],
  [1187,594,["l",2,1187]],[1189,1188,["l",2,2377]],
  [1190,1099,["s",6,13]],[1190,1105,["s",6,13]],
  [1191,1190,["l",2,2381]],[1192,1191,["l",2,2383]],
  [1193,597,["l",2,1193]],[1195,1194,["l",2,2389]],
  [1197,1196,["l",2,2393]],[1200,1199,["l",2,2399]],
  [1201,601,["l",2,1201]],[1201,1200,["l",2,2401]],
  [1201,1200,["s",4,49]],[1206,1205,["l",2,2411]],
  [1209,1208,["l",2,2417]],[1212,1211,["l",2,2423]],
  [1213,607,["l",2,1213]],[1217,609,["l",2,1217]],
  [1219,1218,["l",2,2437]],[1221,1220,["l",2,2441]],
  [1223,612,["l",2,1223]],[1224,1223,["l",2,2447]],
  [1229,615,["l",2,1229]],[1230,1229,["l",2,2459]],
  [1231,616,["l",2,1231]],[1234,1233,["l",2,2467]],
  [1237,619,["l",2,1237]],[1237,1236,["l",2,2473]],
  [1239,1238,["l",2,2477]],[1249,625,["l",2,1249]],
  [1252,1251,["l",2,2503]],[1259,630,["l",2,1259]],
  [1261,1240,["l",3,61]],[1261,1260,["l",2,2521]],
  [1266,1265,["l",2,2531]],[1270,1269,["l",2,2539]],
  [1272,1271,["l",2,2543]],[1275,1274,["l",2,2549]],
  [1276,1275,["l",2,2551]],[1277,639,["l",2,1277]],
  [1279,640,["l",2,1279]],[1279,1278,["l",2,2557]],
  [1283,642,["l",2,1283]],[1289,645,["l",2,1289]],
  [1290,1289,["l",2,2579]],[1291,646,["l",2,1291]],
  [1296,1295,["l",2,2591]],[1297,649,["l",2,1297]],
  [1297,1296,["l",2,2593]],[1301,651,["l",2,1301]],
  [1303,652,["l",2,1303]],[1305,1304,["l",2,2609]],
  [1307,654,["l",2,1307]],[1309,1308,["l",2,2617]],
  [1311,1310,["l",2,2621]],[1317,1316,["l",2,2633]],
  [1319,660,["l",2,1319]],[1320,732,["s",6,11]],
  [1321,661,["l",2,1321]],[1324,1323,["l",2,2647]],
  [1327,664,["l",2,1327]],[1329,1328,["l",2,2657]],
  [1330,1329,["l",2,2659]],[1332,685,["s",4,37]],
  [1332,703,["s",4,37]],[1332,1331,["l",2,2663]],
  [1336,1335,["l",2,2671]],[1339,1338,["l",2,2677]],
  [1342,732,["s",6,11]],[1342,1320,["s",6,11]],
  [1342,1341,["l",2,2683]],[1344,1343,["l",2,2687]],
  [1345,1344,["l",2,2689]],[1347,1346,["l",2,2693]],
  [1350,1349,["l",2,2699]],[1354,1353,["l",2,2707]],
  [1356,1355,["l",2,2711]],[1357,1356,["l",2,2713]],
  [1360,1359,["l",2,2719]],[1361,681,["l",2,1361]],
  [1365,1364,["l",2,2729]],[1366,1365,["l",2,2731]],
  [1367,684,["l",2,1367]],[1368,1201,["s",8,7]],
  [1368,1204,["s",8,7]],[1368,1333,["u",3,37]],
  [1371,1370,["l",2,2741]],[1373,687,["l",2,1373]],
  [1375,1374,["l",2,2749]],[1377,1376,["l",2,2753]],
  [1381,691,["l",2,1381]],[1384,1383,["l",2,2767]],
  [1387,1365,["l",3,64]],[1389,1388,["l",2,2777]],
  [1395,1394,["l",2,2789]],[1396,1395,["l",2,2791]],
  [1399,700,["l",2,1399]],[1399,1398,["l",2,2797]],
  [1400,1368,["s",8,7]],[1401,1400,["l",2,2801]],
  [1402,1401,["l",2,2803]],[1405,1404,["l",2,2809]],
  [1405,1404,["s",4,53]],[1406,685,["s",4,37]],
  [1406,703,["s",4,37]],[1406,1332,["s",4,37]],
  [1406,1368,["u",3,37]],[1409,705,["l",2,1409]],
  [1410,1409,["l",2,2819]],[1417,1416,["l",2,2833]],
  [1419,1418,["l",2,2837]],[1422,1421,["l",2,2843]],
  [1423,712,["l",2,1423]],[1426,1425,["l",2,2851]],
  [1427,714,["l",2,1427]],[1429,715,["l",2,1429]],
  [1429,1428,["l",2,2857]],[1431,1405,["s",4,53]],
  [1431,1430,["l",2,2861]],[1433,717,["l",2,1433]],
  [1439,720,["l",2,1439]],[1440,1439,["l",2,2879]],
  [1444,1443,["l",2,2887]],[1447,724,["l",2,1447]],
  [1449,1448,["l",2,2897]],[1451,726,["l",2,1451]],
  [1452,1451,["l",2,2903]],[1453,727,["l",2,1453]],
  [1455,1454,["l",2,2909]],[1459,730,["l",2,1459]],
  [1459,1458,["l",2,2917]],[1460,1220,["s",14,3]],
  [1460,1230,["s",14,3]],[1460,1260,["s",14,3]],
  [1460,1452,["s",14,3]],[1464,1463,["l",2,2927]],
  [1470,1469,["l",2,2939]],[1471,736,["l",2,1471]],
  [1476,1460,["s",14,3]],[1477,1476,["l",2,2953]],
  [1479,1478,["l",2,2957]],[1481,741,["l",2,1481]],
  [1482,1481,["l",2,2963]],[1483,742,["l",2,1483]],
  [1485,1484,["l",2,2969]],[1486,1485,["l",2,2971]],
  [1487,744,["l",2,1487]],[1489,745,["l",2,1489]],
  [1493,747,["l",2,1493]],[1499,750,["l",2,1499]],
  [1500,1499,["l",2,2999]],[1501,1500,["l",2,3001]],
  [1506,1505,["l",2,3011]],[1510,1509,["l",2,3019]],
  [1511,756,["l",2,1511]],[1512,1511,["l",2,3023]],
  [1519,1496,["l",3,67]],[1519,1518,["l",2,3037]],
  [1521,1520,["l",2,3041]],[1523,762,["l",2,1523]],
  [1525,1524,["l",2,3049]],[1531,766,["l",2,1531]],
  [1531,1530,["l",2,3061]],[1534,1533,["l",2,3067]],
  [1540,1539,["l",2,3079]],[1542,1541,["l",2,3083]],
  [1543,772,["l",2,1543]],[1545,1544,["l",2,3089]],
  [1549,775,["l",2,1549]],[1553,777,["l",2,1553]],
  [1555,1554,["l",2,3109]],[1559,780,["l",2,1559]],
  [1560,1559,["l",2,3119]],[1561,1560,["l",2,3121]],
  [1563,1562,["l",2,3125]],[1567,784,["l",2,1567]],
  [1569,1568,["l",2,3137]],[1571,786,["l",2,1571]],
  [1579,790,["l",2,1579]],[1582,1581,["l",2,3163]],
  [1583,792,["l",2,1583]],[1584,1583,["l",2,3167]],
  [1585,1584,["l",2,3169]],[1591,1590,["l",2,3181]],
  [1594,1593,["l",2,3187]],[1596,1595,["l",2,3191]],
  [1597,799,["l",2,1597]],[1601,801,["l",2,1601]],
  [1602,1601,["l",2,3203]],[1605,1604,["l",2,3209]],
  [1607,804,["l",2,1607]],[1609,805,["l",2,1609]],
  [1609,1608,["l",2,3217]],[1611,1610,["l",2,3221]],
  [1613,807,["l",2,1613]],[1615,1614,["l",2,3229]],
  [1619,810,["l",2,1619]],[1621,811,["l",2,1621]],
  [1626,1625,["l",2,3251]],[1627,814,["l",2,1627]],
  [1627,1626,["l",2,3253]],[1629,1628,["l",2,3257]],
  [1630,1629,["l",2,3259]],[1636,1635,["l",2,3271]],
  [1637,819,["l",2,1637]],[1640,841,["s",4,41]],
  [1640,861,["s",4,41]],[1640,1573,["l",8,3]],
  [1650,1649,["l",2,3299]],[1651,1650,["l",2,3301]],
  [1654,1653,["l",2,3307]],[1657,829,["l",2,1657]],
  [1657,1656,["l",2,3313]],[1660,1659,["l",2,3319]],
  [1662,1661,["l",2,3323]],[1663,832,["l",2,1663]],
  [1665,1664,["l",2,3329]],[1666,1665,["l",2,3331]],
  [1667,834,["l",2,1667]],[1669,835,["l",2,1669]],
  [1672,1671,["l",2,3343]],[1674,1673,["l",2,3347]],
  [1680,1679,["l",2,3359]],[1681,1680,["l",2,3361]],
  [1686,1685,["l",2,3371]],[1687,1686,["l",2,3373]],
  [1693,847,["l",2,1693]],[1695,1694,["l",2,3389]],
  [1696,1695,["l",2,3391]],[1697,849,["l",2,1697]],
  [1699,850,["l",2,1699]],[1704,1703,["l",2,3407]],
  [1707,1706,["l",2,3413]],[1709,855,["l",2,1709]],
  [1717,1716,["l",2,3433]],[1721,861,["l",2,1721]],
  [1722,841,["s",4,41]],[1722,861,["s",4,41]],
  [1722,1640,["s",4,41]],[1723,862,["l",2,1723]],
  [1723,1680,["l",3,41]],[1725,1724,["l",2,3449]],
  [1729,1728,["l",2,3457]],[1731,1730,["l",2,3461]],
  [1732,1731,["l",2,3463]],[1733,867,["l",2,1733]],
  [1734,1733,["l",2,3467]],[1735,1734,["l",2,3469]],
  [1741,871,["l",2,1741]],[1741,1740,["l",2,3481]],
  [1741,1740,["s",4,59]],[1746,1745,["l",2,3491]],
  [1747,874,["l",2,1747]],[1750,1749,["l",2,3499]],
  [1753,877,["l",2,1753]],[1756,1755,["l",2,3511]],
  [1759,880,["l",2,1759]],[1759,1758,["l",2,3517]],
  [1764,1763,["l",2,3527]],[1765,1764,["l",2,3529]],
  [1767,1766,["l",2,3533]],[1770,1741,["s",4,59]],
  [1770,1769,["l",2,3539]],[1771,1770,["l",2,3541]],
  [1774,1773,["l",2,3547]],[1777,889,["l",2,1777]],
  [1779,1778,["l",2,3557]],[1780,1779,["l",2,3559]],
  [1783,892,["l",2,1783]],[1786,1785,["l",2,3571]],
  [1787,894,["l",2,1787]],[1789,895,["l",2,1789]],
  [1791,1790,["l",2,3581]],[1792,1791,["l",2,3583]],
  [1797,1796,["l",2,3593]],[1801,901,["l",2,1801]],
  [1804,1803,["l",2,3607]],[1806,925,["s",4,43]],
  [1806,946,["s",4,43]],[1807,1806,["l",2,3613]],
  [1809,1808,["l",2,3617]],[1811,906,["l",2,1811]],
  [1812,1811,["l",2,3623]],[1816,1815,["l",2,3631]],
  [1819,1818,["l",2,3637]],[1822,1821,["l",2,3643]],
  [1823,912,["l",2,1823]],[1830,1829,["l",2,3659]],
  [1831,916,["l",2,1831]],[1836,1835,["l",2,3671]],
  [1837,1836,["l",2,3673]],[1839,1838,["l",2,3677]],
  [1846,1845,["l",2,3691]],[1847,924,["l",2,1847]],
  [1848,1807,["u",3,43]],[1849,1848,["l",2,3697]],
  [1851,1850,["l",2,3701]],[1855,1854,["l",2,3709]],
  [1860,1859,["l",2,3719]],[1861,931,["l",2,1861]],
  [1861,1860,["l",2,3721]],[1861,1860,["s",4,61]],
  [1864,1863,["l",2,3727]],[1867,934,["l",2,1867]],
  [1867,1866,["l",2,3733]],[1870,1869,["l",2,3739]],
  [1871,936,["l",2,1871]],[1873,937,["l",2,1873]],
  [1877,939,["l",2,1877]],[1878,1638,["s",10,5]],
  [1878,1860,["s",10,5]],[1879,940,["l",2,1879]],
  [1881,1880,["l",2,3761]],[1884,1883,["l",2,3767]],
  [1885,1884,["l",2,3769]],[1889,945,["l",2,1889]],
  [1890,1889,["l",2,3779]],[1891,1861,["s",4,61]],
  [1892,925,["s",4,43]],[1892,946,["s",4,43]],
  [1892,1806,["s",4,43]],[1892,1848,["u",3,43]],
  [1897,1896,["l",2,3793]],[1899,1898,["l",2,3797]],
  [1901,951,["l",2,1901]],[1902,1901,["l",2,3803]],
  [1907,954,["l",2,1907]],[1911,1910,["l",2,3821]],
  [1912,1911,["l",2,3823]],[1913,957,["l",2,1913]],
  [1917,1916,["l",2,3833]],[1924,1923,["l",2,3847]],
  [1926,1925,["l",2,3851]],[1927,1926,["l",2,3853]],
  [1931,966,["l",2,1931]],[1932,1931,["l",2,3863]],
  [1933,967,["l",2,1933]],[1939,1938,["l",2,3877]],
  [1941,1940,["l",2,3881]],[1945,1944,["l",2,3889]],
  [1949,975,["l",2,1949]],[1951,976,["l",2,1951]],
  [1953,1562,["l",6,5]],[1954,1953,["l",2,3907]],
  [1956,1955,["l",2,3911]],[1959,1958,["l",2,3917]],
  [1960,1959,["l",2,3919]],[1962,1961,["l",2,3923]],
  [1965,1964,["l",2,3929]],[1966,1965,["l",2,3931]],
  [1972,1971,["l",2,3943]],[1973,987,["l",2,1973]],
  [1974,1973,["l",2,3947]],[1979,990,["l",2,1979]],
  [1984,1983,["l",2,3967]],[1987,994,["l",2,1987]],
  [1993,997,["l",2,1993]],[1995,1994,["l",2,3989]],
  [1997,999,["l",2,1997]],[1999,1000,["l",2,1999]],
  [2001,2000,["l",2,4001]],[2002,2001,["l",2,4003]],
  [2003,1002,["l",2,2003]],[2004,2003,["l",2,4007]],
  [2007,2006,["l",2,4013]],[2010,2009,["l",2,4019]],
  [2011,1006,["l",2,2011]],[2011,2010,["l",2,4021]],
  [2014,2013,["l",2,4027]],[2017,1009,["l",2,2017]],
  [2025,2024,["l",2,4049]],[2026,2025,["l",2,4051]],
  [2027,1014,["l",2,2027]],[2029,1015,["l",2,2029]],
  [2029,2028,["l",2,4057]],[2037,2036,["l",2,4073]],
  [2039,1020,["l",2,2039]],[2040,2039,["l",2,4079]],
  [2046,2045,["l",2,4091]],[2047,1953,["l",11,2]],
  [2047,2046,["l",2,4093]],[2049,2047,["l",2,2048]],
  [2050,2049,["l",2,4099]],[2053,1027,["l",2,2053]],
  [2056,2055,["l",2,4111]],[2063,1032,["l",2,2063]],
  [2064,2063,["l",2,4127]],[2065,2064,["l",2,4129]],
  [2067,2066,["l",2,4133]],[2069,1035,["l",2,2069]],
  [2070,2069,["l",2,4139]],[2077,2076,["l",2,4153]],
  [2079,2078,["l",2,4157]],[2080,2079,["l",2,4159]],
  [2081,1041,["l",2,2081]],[2083,1042,["l",2,2083]],
  [2087,1044,["l",2,2087]],[2089,1045,["l",2,2089]],
  [2089,2088,["l",2,4177]],[2099,1050,["l",2,2099]],
  [2101,2100,["l",2,4201]],[2106,2105,["l",2,4211]],
  [2109,2108,["l",2,4217]],[2110,2109,["l",2,4219]],
  [2111,1056,["l",2,2111]],[2113,1057,["l",2,2113]],
  [2115,2114,["l",2,4229]],[2116,2115,["l",2,4231]],
  [2121,2120,["l",2,4241]],[2122,2121,["l",2,4243]],
  [2127,2126,["l",2,4253]],[2129,1065,["l",2,2129]],
  [2130,2129,["l",2,4259]],[2131,1066,["l",2,2131]],
  [2131,2130,["l",2,4261]],[2136,2135,["l",2,4271]],
  [2137,1069,["l",2,2137]],[2137,2136,["l",2,4273]],
  [2141,1071,["l",2,2141]],[2142,2141,["l",2,4283]],
  [2143,1072,["l",2,2143]],[2145,2144,["l",2,4289]],
  [2149,2148,["l",2,4297]],[2153,1077,["l",2,2153]],
  [2161,1081,["l",2,2161]],[2162,1105,["s",4,47]],
  [2162,1128,["s",4,47]],[2164,2163,["l",2,4327]],
  [2169,2168,["l",2,4337]],[2170,2169,["l",2,4339]],
  [2175,2174,["l",2,4349]],[2178,1460,["s",14,3]],
  [2178,1476,["s",14,3]],[2179,1090,["l",2,2179]],
  [2179,2178,["l",2,4357]],[2182,2181,["l",2,4363]],
  [2184,1190,["s",6,13]],[2184,1460,["s",14,3]],
  [2184,1476,["s",14,3]],[2184,2178,["s",14,3]],
  [2187,2186,["l",2,4373]],[2190,1460,["s",14,3]],
  [2190,1476,["s",14,3]],[2190,2178,["s",14,3]],
  [2190,2184,["s",14,3]],[2196,1460,["s",14,3]],
  [2196,1476,["s",14,3]],[2196,2178,["s",14,3]],
  [2196,2184,["s",14,3]],[2196,2190,["s",14,3]],
  [2196,2195,["l",2,4391]],[2199,2198,["l",2,4397]],
  [2203,1102,["l",2,2203]],[2205,2204,["l",2,4409]],
  [2207,1104,["l",2,2207]],[2210,1190,["s",6,13]],
  [2210,2184,["s",6,13]],[2211,2210,["l",2,4421]],
  [2212,2211,["l",2,4423]],[2213,1107,["l",2,2213]],
  [2221,1111,["l",2,2221]],[2221,2220,["l",2,4441]],
  [2224,2223,["l",2,4447]],[2226,2225,["l",2,4451]],
  [2229,2228,["l",2,4457]],[2232,2231,["l",2,4463]],
  [2237,1119,["l",2,2237]],[2239,1120,["l",2,2239]],
  [2241,2240,["l",2,4481]],[2242,2241,["l",2,4483]],
  [2243,1122,["l",2,2243]],[2245,2244,["l",2,4489]],
  [2245,2244,["s",4,67]],[2247,2246,["l",2,4493]],
  [2251,1126,["l",2,2251]],[2254,2253,["l",2,4507]],
  [2256,1105,["s",4,47]],[2256,1128,["s",4,47]],
  [2256,2162,["s",4,47]],[2257,2208,["l",3,47]],
  [2257,2256,["l",2,4513]],[2259,2258,["l",2,4517]],
  [2260,2259,["l",2,4519]],[2262,2261,["l",2,4523]],
  [2267,1134,["l",2,2267]],[2269,1135,["l",2,2269]],
  [2273,1137,["l",2,2273]],[2274,2273,["l",2,4547]],
  [2275,2274,["l",2,4549]],[2278,2245,["s",4,67]],
  [2281,1141,["l",2,2281]],[2281,2280,["l",2,4561]],
  [2284,2283,["l",2,4567]],[2287,1144,["l",2,2287]],
  [2292,2291,["l",2,4583]],[2293,1147,["l",2,2293]],
  [2296,2295,["l",2,4591]],[2297,1149,["l",2,2297]],
  [2299,2298,["l",2,4597]],[2302,2301,["l",2,4603]],
  [2309,1155,["l",2,2309]],[2311,1156,["l",2,2311]],
  [2311,2310,["l",2,4621]],[2319,2318,["l",2,4637]],
  [2320,2319,["l",2,4639]],[2322,2321,["l",2,4643]],
  [2325,2324,["l",2,4649]],[2326,2325,["l",2,4651]],
  [2329,2328,["l",2,4657]],[2332,2331,["l",2,4663]],
  [2333,1167,["l",2,2333]],[2337,2336,["l",2,4673]],
  [2339,1170,["l",2,2339]],[2340,2339,["l",2,4679]],
  [2341,1171,["l",2,2341]],[2346,2345,["l",2,4691]],
  [2347,1174,["l",2,2347]],[2351,1176,["l",2,2351]],
  [2352,2351,["l",2,4703]],[2357,1179,["l",2,2357]],
  [2361,2360,["l",2,4721]],[2362,2361,["l",2,4723]],
  [2365,2364,["l",2,4729]],[2367,2366,["l",2,4733]],
  [2371,1186,["l",2,2371]],[2376,2375,["l",2,4751]],
  [2377,1189,["l",2,2377]],[2380,2379,["l",2,4759]],
  [2381,1191,["l",2,2381]],[2383,1192,["l",2,2383]],
  [2389,1195,["l",2,2389]],[2392,2391,["l",2,4783]],
  [2393,1197,["l",2,2393]],[2394,1368,["s",8,7]],
  [2394,1400,["s",8,7]],[2394,2393,["l",2,4787]],
  [2395,2394,["l",2,4789]],[2397,2396,["l",2,4793]],
  [2399,1200,["l",2,2399]],[2400,2101,["u",5,7]],
  [2400,2353,["u",3,49]],[2400,2399,["l",2,4799]],
  [2401,2400,["l",2,4801]],[2407,2406,["l",2,4813]],
  [2408,1368,["s",8,7]],[2408,1400,["s",8,7]],
  [2408,2394,["s",8,7]],[2408,2400,["u",5,7]],
  [2409,2408,["l",2,4817]],[2411,1206,["l",2,2411]],
  [2416,2415,["l",2,4831]],[2417,1209,["l",2,2417]],
  [2423,1212,["l",2,2423]],[2431,2430,["l",2,4861]],
  [2436,2435,["l",2,4871]],[2437,1219,["l",2,2437]],
  [2439,2438,["l",2,4877]],[2441,1221,["l",2,2441]],
  [2445,2444,["l",2,4889]],[2447,1224,["l",2,2447]],
  [2452,2451,["l",2,4903]],[2455,2454,["l",2,4909]],
  [2457,2320,["u",4,17]],[2457,2448,["u",4,17]],
  [2457,2456,["l",2,4913]],[2459,1230,["l",2,2459]],
  [2460,2459,["l",2,4919]],[2466,2465,["l",2,4931]],
  [2467,1234,["l",2,2467]],[2467,2466,["l",2,4933]],
  [2469,2468,["l",2,4937]],[2472,2471,["l",2,4943]],
  [2473,1237,["l",2,2473]],[2476,2475,["l",2,4951]],
  [2477,1239,["l",2,2477]],[2479,2478,["l",2,4957]],
  [2484,2483,["l",2,4967]],[2485,2484,["l",2,4969]],
  [2487,2486,["l",2,4973]],[2494,2493,["l",2,4987]],
  [2497,2496,["l",2,4993]],[2500,2499,["l",2,4999]],
  [2502,2501,["l",2,5003]],[2503,1252,["l",2,2503]],
  [2505,2504,["l",2,5009]],[2506,2505,["l",2,5011]],
  [2511,2510,["l",2,5021]],[2512,2511,["l",2,5023]],
  [2520,2519,["l",2,5039]],[2521,1261,["l",2,2521]],
  [2521,2520,["l",2,5041]],[2526,2525,["l",2,5051]],
  [2530,2529,["l",2,5059]],[2531,1266,["l",2,2531]],
  [2539,1270,["l",2,2539]],[2539,2538,["l",2,5077]],
  [2541,2540,["l",2,5081]],[2543,1272,["l",2,2543]],
  [2544,2543,["l",2,5087]],[2549,1275,["l",2,2549]],
  [2550,2549,["l",2,5099]],[2551,1276,["l",2,2551]],
  [2551,2550,["l",2,5101]],[2554,2553,["l",2,5107]],
  [2557,1279,["l",2,2557]],[2557,2556,["l",2,5113]],
  [2560,2559,["l",2,5119]],[2574,2573,["l",2,5147]],
  [2577,2576,["l",2,5153]],[2579,1290,["l",2,2579]],
  [2584,2583,["l",2,5167]],[2586,2585,["l",2,5171]],
  [2590,2589,["l",2,5179]],[2591,1296,["l",2,2591]],
  [2593,1297,["l",2,2593]],[2595,2594,["l",2,5189]],
  [2599,2598,["l",2,5197]],[2605,2604,["l",2,5209]],
  [2609,1305,["l",2,2609]],[2610,2457,["s",6,17]],
  [2610,2465,["s",6,17]],[2614,2613,["l",2,5227]],
  [2616,2615,["l",2,5231]],[2617,1309,["l",2,2617]],
  [2617,2616,["l",2,5233]],[2619,2618,["l",2,5237]],
  [2621,1311,["l",2,2621]],[2631,2630,["l",2,5261]],
  [2633,1317,["l",2,2633]],[2637,2636,["l",2,5273]],
  [2640,2639,["l",2,5279]],[2641,2640,["l",2,5281]],
  [2647,1324,["l",2,2647]],[2649,2648,["l",2,5297]],
  [2652,2651,["l",2,5303]],[2655,2654,["l",2,5309]],
  [2657,1329,["l",2,2657]],[2659,1330,["l",2,2659]],
  [2662,2661,["l",2,5323]],[2663,1332,["l",2,2663]],
  [2665,2664,["l",2,5329]],[2667,2666,["l",2,5333]],
  [2671,1336,["l",2,2671]],[2674,2673,["l",2,5347]],
  [2676,2675,["l",2,5351]],[2677,1339,["l",2,2677]],
  [2683,1342,["l",2,2683]],[2687,1344,["l",2,2687]],
  [2689,1345,["l",2,2689]],[2691,2690,["l",2,5381]],
  [2693,1347,["l",2,2693]],[2694,2693,["l",2,5387]],
  [2697,2696,["l",2,5393]],[2699,1350,["l",2,2699]],
  [2700,2699,["l",2,5399]],[2704,2703,["l",2,5407]],
  [2707,1354,["l",2,2707]],[2707,2706,["l",2,5413]],
  [2709,2708,["l",2,5417]],[2710,2709,["l",2,5419]],
  [2711,1356,["l",2,2711]],[2713,1357,["l",2,2713]],
  [2716,2715,["l",2,5431]],[2719,1360,["l",2,2719]],
  [2719,2718,["l",2,5437]],[2721,2720,["l",2,5441]],
  [2722,2721,["l",2,5443]],[2725,2724,["l",2,5449]],
  [2729,1365,["l",2,2729]],[2731,1366,["l",2,2731]],
  [2736,2735,["l",2,5471]],[2739,2738,["l",2,5477]],
  [2740,2739,["l",2,5479]],[2741,1371,["l",2,2741]],
  [2742,2741,["l",2,5483]],[2749,1375,["l",2,2749]],
  [2751,2750,["l",2,5501]],[2752,2751,["l",2,5503]],
  [2753,1377,["l",2,2753]],[2754,2753,["l",2,5507]],
  [2756,1405,["s",4,53]],[2756,1431,["s",4,53]],
  [2760,2759,["l",2,5519]],[2761,2760,["l",2,5521]],
  [2764,2763,["l",2,5527]],[2766,2765,["l",2,5531]],
  [2767,1384,["l",2,2767]],[2777,1389,["l",2,2777]],
  [2779,2778,["l",2,5557]],[2782,2781,["l",2,5563]],
  [2785,2784,["l",2,5569]],[2787,2786,["l",2,5573]],
  [2789,1395,["l",2,2789]],[2791,1396,["l",2,2791]],
  [2791,2790,["l",2,5581]],[2796,2795,["l",2,5591]],
  [2797,1399,["l",2,2797]],[2801,1401,["l",2,2801]],
  [2801,2736,["l",5,7]],[2803,1402,["l",2,2803]],
  [2812,2811,["l",2,5623]],[2819,1410,["l",2,2819]],
  [2820,2819,["l",2,5639]],[2821,2820,["l",2,5641]],
  [2824,2823,["l",2,5647]],[2826,2825,["l",2,5651]],
  [2827,2826,["l",2,5653]],[2829,2828,["l",2,5657]],
  [2830,2829,["l",2,5659]],[2833,1417,["l",2,2833]],
  [2835,2834,["l",2,5669]],[2837,1419,["l",2,2837]],
  [2842,2841,["l",2,5683]],[2843,1422,["l",2,2843]],
  [2845,2844,["l",2,5689]],[2847,2846,["l",2,5693]],
  [2851,1426,["l",2,2851]],[2851,2850,["l",2,5701]],
  [2856,2855,["l",2,5711]],[2857,1429,["l",2,2857]],
  [2859,2858,["l",2,5717]],[2861,1431,["l",2,2861]],
  [2862,1405,["s",4,53]],[2862,1431,["s",4,53]],
  [2862,2756,["s",4,53]],[2863,2808,["l",3,53]],
  [2869,2868,["l",2,5737]],[2871,2870,["l",2,5741]],
  [2872,2871,["l",2,5743]],[2875,2874,["l",2,5749]],
  [2879,1440,["l",2,2879]],[2887,1444,["l",2,2887]],
  [2890,2889,["l",2,5779]],[2892,2891,["l",2,5783]],
  [2896,2895,["l",2,5791]],[2897,1449,["l",2,2897]],
  [2901,2900,["l",2,5801]],[2903,1452,["l",2,2903]],
  [2904,2903,["l",2,5807]],[2907,2906,["l",2,5813]],
  [2909,1455,["l",2,2909]],[2911,2910,["l",2,5821]],
  [2914,2913,["l",2,5827]],[2917,1459,["l",2,2917]],
  [2920,2919,["l",2,5839]],[2922,2921,["l",2,5843]],
  [2925,2924,["l",2,5849]],[2926,2925,["l",2,5851]],
  [2927,1464,["l",2,2927]],[2929,2928,["l",2,5857]],
  [2931,2930,["l",2,5861]],[2934,2933,["l",2,5867]],
  [2935,2934,["l",2,5869]],[2939,1470,["l",2,2939]],
  [2940,2939,["l",2,5879]],[2941,2940,["l",2,5881]],
  [2949,2948,["l",2,5897]],[2952,2951,["l",2,5903]],
  [2953,1477,["l",2,2953]],[2957,1479,["l",2,2957]],
  [2962,2961,["l",2,5923]],[2963,1482,["l",2,2963]],
  [2964,2963,["l",2,5927]],[2969,1485,["l",2,2969]],
  [2970,2969,["l",2,5939]],[2971,1486,["l",2,2971]],
  [2977,2976,["l",2,5953]],[2991,2990,["l",2,5981]],
  [2994,2993,["l",2,5987]],[2999,1500,["l",2,2999]],
  [3001,1501,["l",2,3001]],[3004,3003,["l",2,6007]],
  [3006,3005,["l",2,6011]],[3011,1506,["l",2,3011]],
  [3015,3014,["l",2,6029]],[3019,1510,["l",2,3019]],
  [3019,3018,["l",2,6037]],[3022,3021,["l",2,6043]],
  [3023,1512,["l",2,3023]],[3024,3023,["l",2,6047]],
  [3027,3026,["l",2,6053]],[3034,3033,["l",2,6067]],
  [3037,1519,["l",2,3037]],[3037,3036,["l",2,6073]],
  [3040,3039,["l",2,6079]],[3041,1521,["l",2,3041]],
  [3045,3044,["l",2,6089]],[3046,3045,["l",2,6091]],
  [3049,1525,["l",2,3049]],[3051,3050,["l",2,6101]],
  [3057,3056,["l",2,6113]],[3061,1531,["l",2,3061]],
  [3061,3060,["l",2,6121]],[3066,3065,["l",2,6131]],
  [3067,1534,["l",2,3067]],[3067,3066,["l",2,6133]],
  [3072,3071,["l",2,6143]],[3076,3075,["l",2,6151]],
  [3079,1540,["l",2,3079]],[3082,3081,["l",2,6163]],
  [3083,1542,["l",2,3083]],[3087,3086,["l",2,6173]],
  [3089,1545,["l",2,3089]],[3099,3098,["l",2,6197]],
  [3100,3099,["l",2,6199]],[3102,3101,["l",2,6203]],
  [3106,3105,["l",2,6211]],[3109,1555,["l",2,3109]],
  [3109,3108,["l",2,6217]],[3111,3110,["l",2,6221]],
  [3115,3114,["l",2,6229]],[3119,1560,["l",2,3119]],
  [3120,1878,["s",10,5]],[3121,1561,["l",2,3121]],
  [3121,3120,["l",2,6241]],[3124,3123,["l",2,6247]],
  [3129,3128,["l",2,6257]],[3130,1878,["s",10,5]],
  [3130,3120,["s",10,5]],[3132,3131,["l",2,6263]],
  [3135,3134,["l",2,6269]],[3136,3135,["l",2,6271]],
  [3137,1569,["l",2,3137]],[3139,3138,["l",2,6277]],
  [3144,3143,["l",2,6287]],[3150,3149,["l",2,6299]],
  [3151,3150,["l",2,6301]],[3156,3155,["l",2,6311]],
  [3159,3158,["l",2,6317]],[3162,3161,["l",2,6323]],
  [3163,1582,["l",2,3163]],[3165,3164,["l",2,6329]],
  [3167,1584,["l",2,3167]],[3169,1585,["l",2,3169]],
  [3169,3168,["l",2,6337]],[3172,3171,["l",2,6343]],
  [3177,3176,["l",2,6353]],[3180,3179,["l",2,6359]],
  [3181,1591,["l",2,3181]],[3181,3180,["l",2,6361]],
  [3184,3183,["l",2,6367]],[3187,1594,["l",2,3187]],
  [3187,3186,["l",2,6373]],[3190,3189,["l",2,6379]],
  [3191,1596,["l",2,3191]],[3195,3194,["l",2,6389]],
  [3199,3198,["l",2,6397]],[3203,1602,["l",2,3203]],
  [3209,1605,["l",2,3209]],[3211,3210,["l",2,6421]],
  [3214,3213,["l",2,6427]],[3217,1609,["l",2,3217]],
  [3221,1611,["l",2,3221]],[3225,3224,["l",2,6449]],
  [3226,3225,["l",2,6451]],[3229,1615,["l",2,3229]],
  [3235,3234,["l",2,6469]],[3237,3236,["l",2,6473]],
  [3241,3240,["l",2,6481]],[3246,3245,["l",2,6491]],
  [3251,1626,["l",2,3251]],[3253,1627,["l",2,3253]],
  [3257,1629,["l",2,3257]],[3259,1630,["l",2,3259]],
  [3261,3260,["l",2,6521]],[3265,3264,["l",2,6529]],
  [3271,1636,["l",2,3271]],[3274,3273,["l",2,6547]],
  [3276,3275,["l",2,6551]],[3277,3276,["l",2,6553]],
  [3281,3280,["l",2,6561]],[3282,3281,["l",2,6563]],
  [3285,3284,["l",2,6569]],[3286,3285,["l",2,6571]],
  [3289,3288,["l",2,6577]],[3291,3290,["l",2,6581]],
  [3299,1650,["l",2,3299]],[3300,3299,["l",2,6599]],
  [3301,1651,["l",2,3301]],[3304,3303,["l",2,6607]],
  [3307,1654,["l",2,3307]],[3310,3309,["l",2,6619]],
  [3313,1657,["l",2,3313]],[3319,1660,["l",2,3319]],
  [3319,3318,["l",2,6637]],[3323,1662,["l",2,3323]],
  [3327,3326,["l",2,6653]],[3329,1665,["l",2,3329]],
  [3330,3329,["l",2,6659]],[3331,1666,["l",2,3331]],
  [3331,3330,["l",2,6661]],[3337,3336,["l",2,6673]],
  [3340,3339,["l",2,6679]],[3343,1672,["l",2,3343]],
  [3345,3344,["l",2,6689]],[3346,3345,["l",2,6691]],
  [3347,1674,["l",2,3347]],[3351,3350,["l",2,6701]],
  [3352,3351,["l",2,6703]],[3355,3354,["l",2,6709]],
  [3359,1680,["l",2,3359]],[3360,3359,["l",2,6719]],
  [3361,1681,["l",2,3361]],[3367,3366,["l",2,6733]],
  [3369,3368,["l",2,6737]],[3371,1686,["l",2,3371]],
  [3373,1687,["l",2,3373]],[3381,3380,["l",2,6761]],
  [3382,3381,["l",2,6763]],[3389,1695,["l",2,3389]],
  [3390,3389,["l",2,6779]],[3391,1696,["l",2,3391]],
  [3391,3390,["l",2,6781]],[3396,3395,["l",2,6791]],
  [3397,3396,["l",2,6793]],[3402,3401,["l",2,6803]],
  [3407,1704,["l",2,3407]],[3412,3411,["l",2,6823]],
  [3413,1707,["l",2,3413]],[3414,3413,["l",2,6827]],
  [3415,3414,["l",2,6829]],[3417,3416,["l",2,6833]],
  [3421,3420,["l",2,6841]],[3422,1741,["s",4,59]],
  [3422,1770,["s",4,59]],[3429,3428,["l",2,6857]],
  [3430,3429,["l",2,6859]],[3432,3431,["l",2,6863]],
  [3433,1717,["l",2,3433]],[3435,3434,["l",2,6869]],
  [3436,3435,["l",2,6871]],[3442,3441,["l",2,6883]],
  [3445,3444,["l",2,6889]],[3449,1725,["l",2,3449]],
  [3450,3449,["l",2,6899]],[3454,3453,["l",2,6907]],
  [3456,3455,["l",2,6911]],[3457,1729,["l",2,3457]],
  [3459,3458,["l",2,6917]],[3461,1731,["l",2,3461]],
  [3463,1732,["l",2,3463]],[3467,1734,["l",2,3467]],
  [3469,1735,["l",2,3469]],[3474,3473,["l",2,6947]],
  [3475,3474,["l",2,6949]],[3480,3479,["l",2,6959]],
  [3481,3480,["l",2,6961]],[3484,3483,["l",2,6967]],
  [3486,3485,["l",2,6971]],[3489,3488,["l",2,6977]],
  [3491,1746,["l",2,3491]],[3492,3491,["l",2,6983]],
  [3496,3495,["l",2,6991]],[3499,1750,["l",2,3499]],
  [3499,3498,["l",2,6997]],[3501,3500,["l",2,7001]],
  [3507,3506,["l",2,7013]],[3510,3509,["l",2,7019]],
  [3511,1756,["l",2,3511]],[3514,3513,["l",2,7027]],
  [3517,1759,["l",2,3517]],[3520,3519,["l",2,7039]],
  [3522,3521,["l",2,7043]],[3527,1764,["l",2,3527]],
  [3529,1765,["l",2,3529]],[3529,3528,["l",2,7057]],
  [3533,1767,["l",2,3533]],[3535,3534,["l",2,7069]],
  [3539,1770,["l",2,3539]],[3540,1741,["s",4,59]],
  [3540,1770,["s",4,59]],[3540,3422,["s",4,59]],
  [3540,3539,["l",2,7079]],[3541,1771,["l",2,3541]],
  [3541,3480,["l",3,59]],[3547,1774,["l",2,3547]],
  [3552,3551,["l",2,7103]],[3555,3554,["l",2,7109]],
  [3557,1779,["l",2,3557]],[3559,1780,["l",2,3559]],
  [3561,3560,["l",2,7121]],[3564,3563,["l",2,7127]],
  [3565,3564,["l",2,7129]],[3571,1786,["l",2,3571]],
  [3576,3575,["l",2,7151]],[3580,3579,["l",2,7159]],
  [3581,1791,["l",2,3581]],[3583,1792,["l",2,3583]],
  [3589,3588,["l",2,7177]],[3593,1797,["l",2,3593]],
  [3594,3593,["l",2,7187]],[3597,3596,["l",2,7193]],
  [3604,3603,["l",2,7207]],[3606,3605,["l",2,7211]],
  [3607,1804,["l",2,3607]],[3607,3606,["l",2,7213]],
  [3610,3609,["l",2,7219]],[3613,1807,["l",2,3613]],
  [3615,3614,["l",2,7229]],[3617,1809,["l",2,3617]],
  [3619,3618,["l",2,7237]],[3620,3430,["s",6,19]],
  [3620,3439,["s",6,19]],[3622,3621,["l",2,7243]],
  [3623,1812,["l",2,3623]],[3624,3623,["l",2,7247]],
  [3627,3626,["l",2,7253]],[3631,1816,["l",2,3631]],
  [3637,1819,["l",2,3637]],[3640,3281,["s",8,9]],
  [3642,3641,["l",2,7283]],[3643,1822,["l",2,3643]],
  [3649,3648,["l",2,7297]],[3654,3653,["l",2,7307]],
  [3655,3654,["l",2,7309]],[3659,1830,["l",2,3659]],
  [3660,1861,["s",4,61]],[3660,1891,["s",4,61]],
  [3661,3660,["l",2,7321]],[3666,3665,["l",2,7331]],
  [3667,3666,["l",2,7333]],[3671,1836,["l",2,3671]],
  [3673,1837,["l",2,3673]],[3675,3674,["l",2,7349]],
  [3676,3675,["l",2,7351]],[3677,1839,["l",2,3677]],
  [3685,3684,["l",2,7369]],[3691,1846,["l",2,3691]],
  [3697,1849,["l",2,3697]],[3697,3696,["l",2,7393]],
  [3701,1851,["l",2,3701]],[3706,3705,["l",2,7411]],
  [3709,1855,["l",2,3709]],[3709,3708,["l",2,7417]],
  [3717,3716,["l",2,7433]],[3719,1860,["l",2,3719]],
  [3720,3661,["u",3,61]],[3726,3725,["l",2,7451]],
  [3727,1864,["l",2,3727]],[3729,3728,["l",2,7457]],
  [3730,3729,["l",2,7459]],[3733,1867,["l",2,3733]],
  [3739,1870,["l",2,3739]],[3739,3738,["l",2,7477]],
  [3741,3740,["l",2,7481]],[3744,3743,["l",2,7487]],
  [3745,3744,["l",2,7489]],[3750,3749,["l",2,7499]],
  [3754,3753,["l",2,7507]],[3759,3758,["l",2,7517]],
  [3761,1881,["l",2,3761]],[3762,3761,["l",2,7523]],
  [3765,3764,["l",2,7529]],[3767,1884,["l",2,3767]],
  [3769,1885,["l",2,3769]],[3769,3768,["l",2,7537]],
  [3771,3770,["l",2,7541]],[3774,3773,["l",2,7547]],
  [3775,3774,["l",2,7549]],[3779,1890,["l",2,3779]],
  [3780,3779,["l",2,7559]],[3781,3780,["l",2,7561]],
  [3782,1861,["s",4,61]],[3782,1891,["s",4,61]],
  [3782,3660,["s",4,61]],[3782,3720,["u",3,61]],
  [3787,3786,["l",2,7573]],[3789,3788,["l",2,7577]],
  [3792,3791,["l",2,7583]],[3793,1897,["l",2,3793]],
  [3795,3794,["l",2,7589]],[3796,3795,["l",2,7591]],
  [3797,1899,["l",2,3797]],[3802,3801,["l",2,7603]],
  [3803,1902,["l",2,3803]],[3804,3803,["l",2,7607]],
  [3811,3810,["l",2,7621]],[3820,3819,["l",2,7639]],
  [3821,1911,["l",2,3821]],[3822,3821,["l",2,7643]],
  [3823,1912,["l",2,3823]],[3825,3824,["l",2,7649]],
  [3833,1917,["l",2,3833]],[3835,3834,["l",2,7669]],
  [3837,3836,["l",2,7673]],[3841,3840,["l",2,7681]],
  [3844,3843,["l",2,7687]],[3846,3845,["l",2,7691]],
  [3847,1924,["l",2,3847]],[3850,3849,["l",2,7699]],
  [3851,1926,["l",2,3851]],[3852,3851,["l",2,7703]],
  [3853,1927,["l",2,3853]],[3859,3858,["l",2,7717]],
  [3862,3861,["l",2,7723]],[3863,1932,["l",2,3863]],
  [3864,3863,["l",2,7727]],[3871,3870,["l",2,7741]],
  [3877,1939,["l",2,3877]],[3877,3876,["l",2,7753]],
  [3879,3878,["l",2,7757]],[3880,3879,["l",2,7759]],
  [3881,1941,["l",2,3881]],[3889,1945,["l",2,3889]],
  [3895,3894,["l",2,7789]],[3897,3896,["l",2,7793]],
  [3907,1954,["l",2,3907]],[3909,3908,["l",2,7817]],
  [3911,1956,["l",2,3911]],[3912,3911,["l",2,7823]],
  [3915,3914,["l",2,7829]],[3917,1959,["l",2,3917]],
  [3919,1960,["l",2,3919]],[3921,3920,["l",2,7841]],
  [3923,1962,["l",2,3923]],[3927,3926,["l",2,7853]],
  [3929,1965,["l",2,3929]],[3931,1966,["l",2,3931]],
  [3934,3933,["l",2,7867]],[3937,3936,["l",2,7873]],
  [3939,3938,["l",2,7877]],[3940,3939,["l",2,7879]],
  [3942,3941,["l",2,7883]],[3943,1972,["l",2,3943]],
  [3947,1974,["l",2,3947]],[3951,3950,["l",2,7901]],
  [3954,3953,["l",2,7907]],[3960,3959,["l",2,7919]],
  [3961,3960,["l",2,7921]],[3964,3963,["l",2,7927]],
  [3967,1984,["l",2,3967]],[3967,3966,["l",2,7933]],
  [3969,3968,["l",2,7937]],[3975,3974,["l",2,7949]],
  [3976,3975,["l",2,7951]],[3982,3981,["l",2,7963]],
  [3989,1995,["l",2,3989]],[3997,3996,["l",2,7993]],
  [4001,2001,["l",2,4001]],[4003,2002,["l",2,4003]],
  [4005,4004,["l",2,8009]],[4006,4005,["l",2,8011]],
  [4007,2004,["l",2,4007]],[4009,4008,["l",2,8017]],
  [4013,2007,["l",2,4013]],[4019,2010,["l",2,4019]],
  [4020,4019,["l",2,8039]],[4021,2011,["l",2,4021]],
  [4027,2014,["l",2,4027]],[4027,4026,["l",2,8053]],
  [4030,4029,["l",2,8059]],[4035,4034,["l",2,8069]],
  [4041,4040,["l",2,8081]],[4044,4043,["l",2,8087]],
  [4045,4044,["l",2,8089]],[4047,4046,["l",2,8093]],
  [4049,2025,["l",2,4049]],[4051,2026,["l",2,4051]],
  [4051,4050,["l",2,8101]],[4056,4055,["l",2,8111]],
  [4057,2029,["l",2,4057]],[4059,4058,["l",2,8117]],
  [4062,4061,["l",2,8123]],[4073,2037,["l",2,4073]],
  [4074,4073,["l",2,8147]],[4079,2040,["l",2,4079]],
  [4081,4080,["l",2,8161]],[4084,4083,["l",2,8167]],
  [4086,4085,["l",2,8171]],[4090,4089,["l",2,8179]],
  [4091,2046,["l",2,4091]],[4093,2047,["l",2,4093]],
  [4095,3315,["u",7,4]],[4095,3641,["u",5,8]],
  [4095,3937,["l",12,2]],[4095,4033,["u",3,64]],
  [4096,4095,["l",2,8191]],[4097,3855,["u",4,16]],
  [4097,4095,["l",2,4096]],[4099,2050,["l",2,4099]],
  [4105,4104,["l",2,8209]],[4110,4109,["l",2,8219]],
  [4111,2056,["l",2,4111]],[4111,4110,["l",2,8221]],
  [4116,4115,["l",2,8231]],[4117,4116,["l",2,8233]],
  [4119,4118,["l",2,8237]],[4122,4121,["l",2,8243]],
  [4127,2064,["l",2,4127]],[4129,2065,["l",2,4129]],
  [4132,4131,["l",2,8263]],[4133,2067,["l",2,4133]],
  [4135,4134,["l",2,8269]],[4137,4136,["l",2,8273]],
  [4139,2070,["l",2,4139]],[4144,4143,["l",2,8287]],
  [4146,4145,["l",2,8291]],[4147,4146,["l",2,8293]],
  [4149,4148,["l",2,8297]],[4153,2077,["l",2,4153]],
  [4156,4155,["l",2,8311]],[4157,2079,["l",2,4157]],
  [4159,2080,["l",2,4159]],[4159,4158,["l",2,8317]],
  [4165,4164,["l",2,8329]],[4177,2089,["l",2,4177]],
  [4177,4176,["l",2,8353]],[4182,4181,["l",2,8363]],
  [4185,4184,["l",2,8369]],[4189,4188,["l",2,8377]],
  [4194,4193,["l",2,8387]],[4195,4194,["l",2,8389]],
  [4201,2101,["l",2,4201]],[4210,4209,["l",2,8419]],
  [4211,2106,["l",2,4211]],[4212,4211,["l",2,8423]],
  [4215,4214,["l",2,8429]],[4216,4215,["l",2,8431]],
  [4217,2109,["l",2,4217]],[4219,2110,["l",2,4219]],
  [4222,4221,["l",2,8443]],[4224,4223,["l",2,8447]],
  [4229,2115,["l",2,4229]],[4231,2116,["l",2,4231]],
  [4231,4230,["l",2,8461]],[4234,4233,["l",2,8467]],
  [4241,2121,["l",2,4241]],[4243,2122,["l",2,4243]],
  [4251,4250,["l",2,8501]],[4253,2127,["l",2,4253]],
  [4257,4256,["l",2,8513]],[4259,2130,["l",2,4259]],
  [4261,2131,["l",2,4261]],[4261,4260,["l",2,8521]],
  [4264,4263,["l",2,8527]],[4269,4268,["l",2,8537]],
  [4270,4269,["l",2,8539]],[4271,2136,["l",2,4271]],
  [4272,4271,["l",2,8543]],[4273,2137,["l",2,4273]],
  [4282,4281,["l",2,8563]],[4283,2142,["l",2,4283]],
  [4287,4286,["l",2,8573]],[4289,2145,["l",2,4289]],
  [4291,4290,["l",2,8581]],[4297,2149,["l",2,4297]],
  [4299,4298,["l",2,8597]],[4300,4299,["l",2,8599]],
  [4305,4304,["l",2,8609]],[4312,4311,["l",2,8623]],
  [4314,4313,["l",2,8627]],[4315,4314,["l",2,8629]],
  [4321,4320,["l",2,8641]],[4324,4323,["l",2,8647]],
  [4327,2164,["l",2,4327]],[4332,4331,["l",2,8663]],
  [4335,4334,["l",2,8669]],[4337,2169,["l",2,4337]],
  [4339,2170,["l",2,4339]],[4339,4338,["l",2,8677]],
  [4341,4340,["l",2,8681]],[4345,4344,["l",2,8689]],
  [4347,4346,["l",2,8693]],[4349,2175,["l",2,4349]],
  [4350,4349,["l",2,8699]],[4354,4353,["l",2,8707]],
  [4357,2179,["l",2,4357]],[4357,4356,["l",2,8713]],
  [4360,4359,["l",2,8719]],[4363,2182,["l",2,4363]],
  [4366,4365,["l",2,8731]],[4369,4095,["l",4,16]],
  [4369,4368,["l",2,8737]],[4371,4370,["l",2,8741]],
  [4372,3640,["s",16,3]],[4372,3660,["s",16,3]],
  [4372,3690,["s",16,3]],[4372,4356,["s",16,3]],
  [4373,2187,["l",2,4373]],[4374,4373,["l",2,8747]],
  [4377,4376,["l",2,8753]],[4380,4372,["s",16,3]],
  [4381,4380,["l",2,8761]],[4390,4389,["l",2,8779]],
  [4391,2196,["l",2,4391]],[4392,4391,["l",2,8783]],
  [4397,2199,["l",2,4397]],[4402,4401,["l",2,8803]],
  [4404,4403,["l",2,8807]],[4409,2205,["l",2,4409]],
  [4410,4409,["l",2,8819]],[4411,4410,["l",2,8821]],
  [4416,4415,["l",2,8831]],[4419,4418,["l",2,8837]],
  [4420,4419,["l",2,8839]],[4421,2211,["l",2,4421]],
  [4422,2245,["s",4,67]],[4422,2278,["s",4,67]],
  [4423,2212,["l",2,4423]],[4425,4424,["l",2,8849]],
  [4431,4430,["l",2,8861]],[4432,4431,["l",2,8863]],
  [4434,4433,["l",2,8867]],[4441,2221,["l",2,4441]],
  [4444,4443,["l",2,8887]],[4447,2224,["l",2,4447]],
  [4447,4446,["l",2,8893]],[4451,2226,["l",2,4451]],
  [4457,2229,["l",2,4457]],[4462,4461,["l",2,8923]],
  [4463,2232,["l",2,4463]],[4465,4464,["l",2,8929]],
  [4467,4466,["l",2,8933]],[4471,4470,["l",2,8941]],
  [4476,4475,["l",2,8951]],[4481,2241,["l",2,4481]],
  [4482,4481,["l",2,8963]],[4483,2242,["l",2,4483]],
  [4485,4484,["l",2,8969]],[4486,4485,["l",2,8971]],
  [4488,4423,["u",3,67]],[4493,2247,["l",2,4493]],
  [4500,4499,["l",2,8999]],[4501,4500,["l",2,9001]],
  [4504,4503,["l",2,9007]],[4506,4505,["l",2,9011]],
  [4507,2254,["l",2,4507]],[4507,4506,["l",2,9013]],
  [4513,2257,["l",2,4513]],[4515,4514,["l",2,9029]],
  [4517,2259,["l",2,4517]],[4519,2260,["l",2,4519]],
  [4521,4520,["l",2,9041]],[4522,4521,["l",2,9043]],
  [4523,2262,["l",2,4523]],[4525,4524,["l",2,9049]],
  [4530,4529,["l",2,9059]],[4534,4533,["l",2,9067]],
  [4546,4545,["l",2,9091]],[4547,2274,["l",2,4547]],
  [4549,2275,["l",2,4549]],[4552,4551,["l",2,9103]],
  [4555,4554,["l",2,9109]],[4556,2245,["s",4,67]],
  [4556,2278,["s",4,67]],[4556,4422,["s",4,67]],
  [4556,4488,["u",3,67]],[4561,2281,["l",2,4561]],
  [4564,4563,["l",2,9127]],[4567,2284,["l",2,4567]],
  [4567,4566,["l",2,9133]],[4569,4568,["l",2,9137]],
  [4576,4575,["l",2,9151]],[4579,4578,["l",2,9157]],
  [4581,4580,["l",2,9161]],[4583,2292,["l",2,4583]],
  [4587,4586,["l",2,9173]],[4591,2296,["l",2,4591]],
  [4591,4590,["l",2,9181]],[4594,4593,["l",2,9187]],
  [4597,2299,["l",2,4597]],[4600,4599,["l",2,9199]],
  [4602,4601,["l",2,9203]],[4603,2302,["l",2,4603]],
  [4605,4604,["l",2,9209]],[4611,4610,["l",2,9221]],
  [4614,4613,["l",2,9227]],[4620,4619,["l",2,9239]],
  [4621,2311,["l",2,4621]],[4621,4620,["l",2,9241]],
  [4629,4628,["l",2,9257]],[4637,2319,["l",2,4637]],
  [4639,2320,["l",2,4639]],[4639,4638,["l",2,9277]],
  [4641,4640,["l",2,9281]],[4642,4641,["l",2,9283]],
  [4643,2322,["l",2,4643]],[4647,4646,["l",2,9293]],
  [4649,2325,["l",2,4649]],[4651,2326,["l",2,4651]],
  [4656,4655,["l",2,9311]],[4657,2329,["l",2,4657]],
  [4660,4659,["l",2,9319]],[4662,4661,["l",2,9323]],
  [4663,2332,["l",2,4663]],[4669,4668,["l",2,9337]],
  [4671,4670,["l",2,9341]],[4672,4671,["l",2,9343]],
  [4673,2337,["l",2,4673]],[4675,4674,["l",2,9349]],
  [4679,2340,["l",2,4679]],[4681,4599,["l",5,8]],
  [4686,4685,["l",2,9371]],[4689,4688,["l",2,9377]],
  [4691,2346,["l",2,4691]],[4696,4695,["l",2,9391]],
  [4699,4698,["l",2,9397]],[4702,4701,["l",2,9403]],
  [4703,2352,["l",2,4703]],[4705,4704,["l",2,9409]],
  [4707,4706,["l",2,9413]],[4710,4709,["l",2,9419]],
  [4711,4710,["l",2,9421]],[4716,4715,["l",2,9431]],
  [4717,4716,["l",2,9433]],[4719,4718,["l",2,9437]],
  [4720,4719,["l",2,9439]],[4721,2361,["l",2,4721]],
  [4723,2362,["l",2,4723]],[4729,2365,["l",2,4729]],
  [4731,4730,["l",2,9461]],[4732,4731,["l",2,9463]],
  [4733,2367,["l",2,4733]],[4734,4733,["l",2,9467]],
  [4737,4736,["l",2,9473]],[4740,4739,["l",2,9479]],
  [4746,4745,["l",2,9491]],[4749,4748,["l",2,9497]],
  [4751,2376,["l",2,4751]],[4756,4755,["l",2,9511]],
  [4759,2380,["l",2,4759]],[4761,4760,["l",2,9521]],
  [4767,4766,["l",2,9533]],[4770,4769,["l",2,9539]],
  [4774,4773,["l",2,9547]],[4776,4775,["l",2,9551]],
  [4783,2392,["l",2,4783]],[4787,2394,["l",2,4787]],
  [4789,2395,["l",2,4789]],[4793,2397,["l",2,4793]],
  [4794,4793,["l",2,9587]],[4799,2400,["l",2,4799]],
  [4801,2401,["l",2,4801]],[4801,4800,["l",2,9601]],
  [4807,4806,["l",2,9613]],[4810,4809,["l",2,9619]],
  [4812,4811,["l",2,9623]],[4813,2407,["l",2,4813]],
  [4815,4814,["l",2,9629]],[4816,4815,["l",2,9631]],
  [4817,2409,["l",2,4817]],[4822,4821,["l",2,9643]],
  [4825,4824,["l",2,9649]],[4831,2416,["l",2,4831]],
  [4831,4830,["l",2,9661]],[4839,4838,["l",2,9677]],
  [4840,4839,["l",2,9679]],[4845,4844,["l",2,9689]],
  [4849,4848,["l",2,9697]],[4860,4859,["l",2,9719]],
  [4861,2431,["l",2,4861]],[4861,4860,["l",2,9721]],
  [4867,4866,["l",2,9733]],[4870,4869,["l",2,9739]],
  [4871,2436,["l",2,4871]],[4872,4871,["l",2,9743]],
  [4875,4874,["l",2,9749]],[4877,2439,["l",2,4877]],
  [4884,4883,["l",2,9767]],[4885,4884,["l",2,9769]],
  [4889,2445,["l",2,4889]],[4891,4890,["l",2,9781]],
  [4894,4893,["l",2,9787]],[4896,2610,["s",6,17]],
  [4896,4895,["l",2,9791]],[4902,4901,["l",2,9803]],
  [4903,2452,["l",2,4903]],[4906,4905,["l",2,9811]],
  [4909,2455,["l",2,4909]],[4909,4908,["l",2,9817]],
  [4915,4914,["l",2,9829]],[4917,4916,["l",2,9833]],
  [4919,2460,["l",2,4919]],[4920,4919,["l",2,9839]],
  [4926,4925,["l",2,9851]],[4929,4928,["l",2,9857]],
  [4930,2610,["s",6,17]],[4930,4896,["s",6,17]],
  [4930,4929,["l",2,9859]],[4931,2466,["l",2,4931]],
  [4933,2467,["l",2,4933]],[4936,4935,["l",2,9871]],
  [4937,2469,["l",2,4937]],[4942,4941,["l",2,9883]],
  [4943,2472,["l",2,4943]],[4944,4943,["l",2,9887]],
  [4951,2476,["l",2,4951]],[4951,4950,["l",2,9901]],
  [4954,4953,["l",2,9907]],[4957,2479,["l",2,4957]],
  [4962,4961,["l",2,9923]],[4965,4964,["l",2,9929]],
  [4966,4965,["l",2,9931]],[4967,2484,["l",2,4967]],
  [4969,2485,["l",2,4969]],[4971,4970,["l",2,9941]],
  [4973,2487,["l",2,4973]],[4975,4974,["l",2,9949]],
  [4984,4983,["l",2,9967]],[4987,2494,["l",2,4987]],
  [4987,4986,["l",2,9973]],[4993,2497,["l",2,4993]],
  [4999,2500,["l",2,4999]],[5003,2502,["l",2,5003]],
  [5009,2505,["l",2,5009]],[5011,2506,["l",2,5011]],
  [5021,2511,["l",2,5021]],[5023,2512,["l",2,5023]],
  [5039,2520,["l",2,5039]],[5051,2526,["l",2,5051]],
  [5059,2530,["l",2,5059]],[5077,2539,["l",2,5077]],
  [5081,2541,["l",2,5081]],[5087,2544,["l",2,5087]],
  [5099,2550,["l",2,5099]],[5101,2551,["l",2,5101]],
  [5107,2554,["l",2,5107]],[5113,2557,["l",2,5113]],
  [5119,2560,["l",2,5119]],[5147,2574,["l",2,5147]],
  [5153,2577,["l",2,5153]],[5167,2584,["l",2,5167]],
  [5171,2586,["l",2,5171]],[5179,2590,["l",2,5179]],
  [5189,2595,["l",2,5189]],[5197,2599,["l",2,5197]],
  [5209,2605,["l",2,5209]],[5227,2614,["l",2,5227]],
  [5231,2616,["l",2,5231]],[5233,2617,["l",2,5233]],
  [5237,2619,["l",2,5237]],[5261,2631,["l",2,5261]],
  [5273,2637,["l",2,5273]],[5279,2640,["l",2,5279]],
  [5281,2641,["l",2,5281]],[5297,2649,["l",2,5297]],
  [5303,2652,["l",2,5303]],[5309,2655,["l",2,5309]],
  [5323,2662,["l",2,5323]],[5333,2667,["l",2,5333]],
  [5347,2674,["l",2,5347]],[5351,2676,["l",2,5351]],
  [5381,2691,["l",2,5381]],[5387,2694,["l",2,5387]],
  [5393,2697,["l",2,5393]],[5399,2700,["l",2,5399]],
  [5407,2704,["l",2,5407]],[5413,2707,["l",2,5413]],
  [5417,2709,["l",2,5417]],[5419,2710,["l",2,5419]],
  [5431,2716,["l",2,5431]],[5437,2719,["l",2,5437]],
  [5441,2721,["l",2,5441]],[5443,2722,["l",2,5443]],
  [5449,2725,["l",2,5449]],[5471,2736,["l",2,5471]],
  [5477,2739,["l",2,5477]],[5479,2740,["l",2,5479]],
  [5483,2742,["l",2,5483]],[5501,2751,["l",2,5501]],
  [5503,2752,["l",2,5503]],[5507,2754,["l",2,5507]],
  [5519,2760,["l",2,5519]],[5521,2761,["l",2,5521]],
  [5527,2764,["l",2,5527]],[5531,2766,["l",2,5531]],
  [5557,2779,["l",2,5557]],[5563,2782,["l",2,5563]],
  [5569,2785,["l",2,5569]],[5573,2787,["l",2,5573]],
  [5581,2791,["l",2,5581]],[5591,2796,["l",2,5591]],
  [5623,2812,["l",2,5623]],[5639,2820,["l",2,5639]],
  [5641,2821,["l",2,5641]],[5647,2824,["l",2,5647]],
  [5651,2826,["l",2,5651]],[5653,2827,["l",2,5653]],
  [5657,2829,["l",2,5657]],[5659,2830,["l",2,5659]],
  [5669,2835,["l",2,5669]],[5683,2842,["l",2,5683]],
  [5689,2845,["l",2,5689]],[5693,2847,["l",2,5693]],
  [5701,2851,["l",2,5701]],[5711,2856,["l",2,5711]],
  [5717,2859,["l",2,5717]],[5737,2869,["l",2,5737]],
  [5741,2871,["l",2,5741]],[5743,2872,["l",2,5743]],
  [5749,2875,["l",2,5749]],[5779,2890,["l",2,5779]],
  [5783,2892,["l",2,5783]],[5791,2896,["l",2,5791]],
  [5801,2901,["l",2,5801]],[5807,2904,["l",2,5807]],
  [5813,2907,["l",2,5813]],[5821,2911,["l",2,5821]],
  [5827,2914,["l",2,5827]],[5839,2920,["l",2,5839]],
  [5843,2922,["l",2,5843]],[5849,2925,["l",2,5849]],
  [5851,2926,["l",2,5851]],[5857,2929,["l",2,5857]],
  [5861,2931,["l",2,5861]],[5867,2934,["l",2,5867]],
  [5869,2935,["l",2,5869]],[5879,2940,["l",2,5879]],
  [5881,2941,["l",2,5881]],[5897,2949,["l",2,5897]],
  [5903,2952,["l",2,5903]],[5923,2962,["l",2,5923]],
  [5927,2964,["l",2,5927]],[5939,2970,["l",2,5939]],
  [5953,2977,["l",2,5953]],[5981,2991,["l",2,5981]],
  [5987,2994,["l",2,5987]],[6007,3004,["l",2,6007]],
  [6011,3006,["l",2,6011]],[6029,3015,["l",2,6029]],
  [6037,3019,["l",2,6037]],[6043,3022,["l",2,6043]],
  [6047,3024,["l",2,6047]],[6053,3027,["l",2,6053]],
  [6067,3034,["l",2,6067]],[6073,3037,["l",2,6073]],
  [6079,3040,["l",2,6079]],[6089,3045,["l",2,6089]],
  [6091,3046,["l",2,6091]],[6101,3051,["l",2,6101]],
  [6113,3057,["l",2,6113]],[6121,3061,["l",2,6121]],
  [6131,3066,["l",2,6131]],[6133,3067,["l",2,6133]],
  [6143,3072,["l",2,6143]],[6151,3076,["l",2,6151]],
  [6163,3082,["l",2,6163]],[6173,3087,["l",2,6173]],
  [6197,3099,["l",2,6197]],[6199,3100,["l",2,6199]],
  [6203,3102,["l",2,6203]],[6211,3106,["l",2,6211]],
  [6217,3109,["l",2,6217]],[6221,3111,["l",2,6221]],
  [6229,3115,["l",2,6229]],[6247,3124,["l",2,6247]],
  [6257,3129,["l",2,6257]],[6263,3132,["l",2,6263]],
  [6269,3135,["l",2,6269]],[6271,3136,["l",2,6271]],
  [6277,3139,["l",2,6277]],[6287,3144,["l",2,6287]],
  [6299,3150,["l",2,6299]],[6301,3151,["l",2,6301]],
  [6311,3156,["l",2,6311]],[6317,3159,["l",2,6317]],
  [6323,3162,["l",2,6323]],[6329,3165,["l",2,6329]],
  [6337,3169,["l",2,6337]],[6343,3172,["l",2,6343]],
  [6353,3177,["l",2,6353]],[6359,3180,["l",2,6359]],
  [6361,3181,["l",2,6361]],[6367,3184,["l",2,6367]],
  [6373,3187,["l",2,6373]],[6379,3190,["l",2,6379]],
  [6389,3195,["l",2,6389]],[6397,3199,["l",2,6397]],
  [6421,3211,["l",2,6421]],[6427,3214,["l",2,6427]],
  [6449,3225,["l",2,6449]],[6451,3226,["l",2,6451]],
  [6469,3235,["l",2,6469]],[6473,3237,["l",2,6473]],
  [6481,3241,["l",2,6481]],[6491,3246,["l",2,6491]],
  [6521,3261,["l",2,6521]],[6529,3265,["l",2,6529]],
  [6547,3274,["l",2,6547]],[6551,3276,["l",2,6551]],
  [6552,4372,["s",16,3]],[6552,4380,["s",16,3]],
  [6553,3277,["l",2,6553]],[6558,4372,["s",16,3]],
  [6558,4380,["s",16,3]],[6558,6552,["s",16,3]],
  [6560,4921,["u",9,3]],[6563,3282,["l",2,6563]],
  [6564,4372,["s",16,3]],[6564,4380,["s",16,3]],
  [6564,6552,["s",16,3]],[6564,6558,["s",16,3]],
  [6564,6560,["u",9,3]],[6569,3285,["l",2,6569]],
  [6570,4372,["s",16,3]],[6570,4380,["s",16,3]],
  [6570,6552,["s",16,3]],[6570,6558,["s",16,3]],
  [6570,6564,["s",16,3]],[6571,3286,["l",2,6571]],
  [6577,3289,["l",2,6577]],[6581,3291,["l",2,6581]],
  [6599,3300,["l",2,6599]],[6607,3304,["l",2,6607]],
  [6619,3310,["l",2,6619]],[6637,3319,["l",2,6637]],
  [6653,3327,["l",2,6653]],[6659,3330,["l",2,6659]],
  [6661,3331,["l",2,6661]],[6673,3337,["l",2,6673]],
  [6679,3340,["l",2,6679]],[6689,3345,["l",2,6689]],
  [6691,3346,["l",2,6691]],[6701,3351,["l",2,6701]],
  [6703,3352,["l",2,6703]],[6709,3355,["l",2,6709]],
  [6719,3360,["l",2,6719]],[6733,3367,["l",2,6733]],
  [6737,3369,["l",2,6737]],[6761,3381,["l",2,6761]],
  [6763,3382,["l",2,6763]],[6779,3390,["l",2,6779]],
  [6781,3391,["l",2,6781]],[6791,3396,["l",2,6791]],
  [6793,3397,["l",2,6793]],[6803,3402,["l",2,6803]],
  [6823,3412,["l",2,6823]],[6827,3414,["l",2,6827]],
  [6829,3415,["l",2,6829]],[6833,3417,["l",2,6833]],
  [6840,3620,["s",6,19]],[6841,3421,["l",2,6841]],
  [6857,3429,["l",2,6857]],[6863,3432,["l",2,6863]],
  [6869,3435,["l",2,6869]],[6871,3436,["l",2,6871]],
  [6878,3620,["s",6,19]],[6878,6840,["s",6,19]],
  [6883,3442,["l",2,6883]],[6899,3450,["l",2,6899]],
  [6907,3454,["l",2,6907]],[6911,3456,["l",2,6911]],
  [6917,3459,["l",2,6917]],[6947,3474,["l",2,6947]],
  [6949,3475,["l",2,6949]],[6959,3480,["l",2,6959]],
  [6961,3481,["l",2,6961]],[6967,3484,["l",2,6967]],
  [6971,3486,["l",2,6971]],[6977,3489,["l",2,6977]],
  [6983,3492,["l",2,6983]],[6991,3496,["l",2,6991]],
  [6997,3499,["l",2,6997]],[7001,3501,["l",2,7001]],
  [7013,3507,["l",2,7013]],[7019,3510,["l",2,7019]],
  [7027,3514,["l",2,7027]],[7039,3520,["l",2,7039]],
  [7043,3522,["l",2,7043]],[7057,3529,["l",2,7057]],
  [7069,3535,["l",2,7069]],[7079,3540,["l",2,7079]],
  [7103,3552,["l",2,7103]],[7109,3555,["l",2,7109]],
  [7121,3561,["l",2,7121]],[7127,3564,["l",2,7127]],
  [7129,3565,["l",2,7129]],[7151,3576,["l",2,7151]],
  [7159,3580,["l",2,7159]],[7177,3589,["l",2,7177]],
  [7187,3594,["l",2,7187]],[7193,3597,["l",2,7193]],
  [7207,3604,["l",2,7207]],[7211,3606,["l",2,7211]],
  [7213,3607,["l",2,7213]],[7219,3610,["l",2,7219]],
  [7229,3615,["l",2,7229]],[7237,3619,["l",2,7237]],
  [7243,3622,["l",2,7243]],[7247,3624,["l",2,7247]],
  [7253,3627,["l",2,7253]],[7283,3642,["l",2,7283]],
  [7297,3649,["l",2,7297]],[7307,3654,["l",2,7307]],
  [7309,3655,["l",2,7309]],[7321,3661,["l",2,7321]],
  [7331,3666,["l",2,7331]],[7333,3667,["l",2,7333]],
  [7349,3675,["l",2,7349]],[7351,3676,["l",2,7351]],
  [7369,3685,["l",2,7369]],[7393,3697,["l",2,7393]],
  [7411,3706,["l",2,7411]],[7417,3709,["l",2,7417]],
  [7433,3717,["l",2,7433]],[7451,3726,["l",2,7451]],
  [7457,3729,["l",2,7457]],[7459,3730,["l",2,7459]],
  [7477,3739,["l",2,7477]],[7481,3741,["l",2,7481]],
  [7487,3744,["l",2,7487]],[7489,3745,["l",2,7489]],
  [7499,3750,["l",2,7499]],[7507,3754,["l",2,7507]],
  [7517,3759,["l",2,7517]],[7523,3762,["l",2,7523]],
  [7529,3765,["l",2,7529]],[7537,3769,["l",2,7537]],
  [7541,3771,["l",2,7541]],[7547,3774,["l",2,7547]],
  [7549,3775,["l",2,7549]],[7559,3780,["l",2,7559]],
  [7561,3781,["l",2,7561]],[7573,3787,["l",2,7573]],
  [7577,3789,["l",2,7577]],[7583,3792,["l",2,7583]],
  [7589,3795,["l",2,7589]],[7591,3796,["l",2,7591]],
  [7603,3802,["l",2,7603]],[7607,3804,["l",2,7607]],
  [7621,3811,["l",2,7621]],[7639,3820,["l",2,7639]],
  [7643,3822,["l",2,7643]],[7649,3825,["l",2,7649]],
  [7669,3835,["l",2,7669]],[7673,3837,["l",2,7673]],
  [7681,3841,["l",2,7681]],[7687,3844,["l",2,7687]],
  [7691,3846,["l",2,7691]],[7699,3850,["l",2,7699]],
  [7703,3852,["l",2,7703]],[7717,3859,["l",2,7717]],
  [7723,3862,["l",2,7723]],[7727,3864,["l",2,7727]],
  [7741,3871,["l",2,7741]],[7753,3877,["l",2,7753]],
  [7757,3879,["l",2,7757]],[7759,3880,["l",2,7759]],
  [7789,3895,["l",2,7789]],[7793,3897,["l",2,7793]],
  [7817,3909,["l",2,7817]],[7823,3912,["l",2,7823]],
  [7829,3915,["l",2,7829]],[7841,3921,["l",2,7841]],
  [7853,3927,["l",2,7853]],[7867,3934,["l",2,7867]],
  [7873,3937,["l",2,7873]],[7877,3939,["l",2,7877]],
  [7879,3940,["l",2,7879]],[7883,3942,["l",2,7883]],
  [7901,3951,["l",2,7901]],[7907,3954,["l",2,7907]],
  [7919,3960,["l",2,7919]],[7927,3964,["l",2,7927]],
  [7933,3967,["l",2,7933]],[7937,3969,["l",2,7937]],
  [7949,3975,["l",2,7949]],[7951,3976,["l",2,7951]],
  [7963,3982,["l",2,7963]],[7993,3997,["l",2,7993]],
  [8009,4005,["l",2,8009]],[8011,4006,["l",2,8011]],
  [8017,4009,["l",2,8017]],[8039,4020,["l",2,8039]],
  [8053,4027,["l",2,8053]],[8059,4030,["l",2,8059]],
  [8069,4035,["l",2,8069]],[8081,4041,["l",2,8081]],
  [8087,4044,["l",2,8087]],[8089,4045,["l",2,8089]],
  [8093,4047,["l",2,8093]],[8101,4051,["l",2,8101]],
  [8111,4056,["l",2,8111]],[8117,4059,["l",2,8117]],
  [8123,4062,["l",2,8123]],[8147,4074,["l",2,8147]],
  [8161,4081,["l",2,8161]],[8167,4084,["l",2,8167]],
  [8171,4086,["l",2,8171]],[8179,4090,["l",2,8179]],
  [8191,4096,["l",2,8191]],[8193,8191,["l",2,8192]],
  [8209,4105,["l",2,8209]],[8219,4110,["l",2,8219]],
  [8221,4111,["l",2,8221]],[8231,4116,["l",2,8231]],
  [8233,4117,["l",2,8233]],[8237,4119,["l",2,8237]],
  [8243,4122,["l",2,8243]],[8263,4132,["l",2,8263]],
  [8269,4135,["l",2,8269]],[8273,4137,["l",2,8273]],
  [8287,4144,["l",2,8287]],[8291,4146,["l",2,8291]],
  [8293,4147,["l",2,8293]],[8297,4149,["l",2,8297]],
  [8311,4156,["l",2,8311]],[8317,4159,["l",2,8317]],
  [8329,4165,["l",2,8329]],[8353,4177,["l",2,8353]],
  [8363,4182,["l",2,8363]],[8369,4185,["l",2,8369]],
  [8377,4189,["l",2,8377]],[8387,4194,["l",2,8387]],
  [8389,4195,["l",2,8389]],[8419,4210,["l",2,8419]],
  [8423,4212,["l",2,8423]],[8429,4215,["l",2,8429]],
  [8431,4216,["l",2,8431]],[8443,4222,["l",2,8443]],
  [8447,4224,["l",2,8447]],[8461,4231,["l",2,8461]],
  [8467,4234,["l",2,8467]],[8501,4251,["l",2,8501]],
  [8513,4257,["l",2,8513]],[8521,4261,["l",2,8521]],
  [8527,4264,["l",2,8527]],[8537,4269,["l",2,8537]],
  [8539,4270,["l",2,8539]],[8543,4272,["l",2,8543]],
  [8563,4282,["l",2,8563]],[8573,4287,["l",2,8573]],
  [8581,4291,["l",2,8581]],[8597,4299,["l",2,8597]],
  [8599,4300,["l",2,8599]],[8609,4305,["l",2,8609]],
  [8623,4312,["l",2,8623]],[8627,4314,["l",2,8627]],
  [8629,4315,["l",2,8629]],[8641,4321,["l",2,8641]],
  [8647,4324,["l",2,8647]],[8663,4332,["l",2,8663]],
  [8669,4335,["l",2,8669]],[8677,4339,["l",2,8677]],
  [8681,4341,["l",2,8681]],[8689,4345,["l",2,8689]],
  [8693,4347,["l",2,8693]],[8699,4350,["l",2,8699]],
  [8707,4354,["l",2,8707]],[8713,4357,["l",2,8713]],
  [8719,4360,["l",2,8719]],[8731,4366,["l",2,8731]],
  [8737,4369,["l",2,8737]],[8741,4371,["l",2,8741]],
  [8747,4374,["l",2,8747]],[8753,4377,["l",2,8753]],
  [8761,4381,["l",2,8761]],[8779,4390,["l",2,8779]],
  [8783,4392,["l",2,8783]],[8803,4402,["l",2,8803]],
  [8807,4404,["l",2,8807]],[8819,4410,["l",2,8819]],
  [8821,4411,["l",2,8821]],[8831,4416,["l",2,8831]],
  [8837,4419,["l",2,8837]],[8839,4420,["l",2,8839]],
  [8849,4425,["l",2,8849]],[8861,4431,["l",2,8861]],
  [8863,4432,["l",2,8863]],[8867,4434,["l",2,8867]],
  [8887,4444,["l",2,8887]],[8893,4447,["l",2,8893]],
  [8923,4462,["l",2,8923]],[8929,4465,["l",2,8929]],
  [8933,4467,["l",2,8933]],[8941,4471,["l",2,8941]],
  [8951,4476,["l",2,8951]],[8963,4482,["l",2,8963]],
  [8969,4485,["l",2,8969]],[8971,4486,["l",2,8971]],
  [8999,4500,["l",2,8999]],[9001,4501,["l",2,9001]],
  [9007,4504,["l",2,9007]],[9011,4506,["l",2,9011]],
  [9013,4507,["l",2,9013]],[9029,4515,["l",2,9029]],
  [9041,4521,["l",2,9041]],[9043,4522,["l",2,9043]],
  [9049,4525,["l",2,9049]],[9059,4530,["l",2,9059]],
  [9067,4534,["l",2,9067]],[9091,4546,["l",2,9091]],
  [9103,4552,["l",2,9103]],[9109,4555,["l",2,9109]],
  [9127,4564,["l",2,9127]],[9133,4567,["l",2,9133]],
  [9137,4569,["l",2,9137]],[9151,4576,["l",2,9151]],
  [9157,4579,["l",2,9157]],[9161,4581,["l",2,9161]],
  [9173,4587,["l",2,9173]],[9181,4591,["l",2,9181]],
  [9187,4594,["l",2,9187]],[9199,4600,["l",2,9199]],
  [9203,4602,["l",2,9203]],[9209,4605,["l",2,9209]],
  [9221,4611,["l",2,9221]],[9227,4614,["l",2,9227]],
  [9239,4620,["l",2,9239]],[9241,4621,["l",2,9241]],
  [9257,4629,["l",2,9257]],[9277,4639,["l",2,9277]],
  [9281,4641,["l",2,9281]],[9283,4642,["l",2,9283]],
  [9293,4647,["l",2,9293]],[9311,4656,["l",2,9311]],
  [9319,4660,["l",2,9319]],[9323,4662,["l",2,9323]],
  [9337,4669,["l",2,9337]],[9341,4671,["l",2,9341]],
  [9343,4672,["l",2,9343]],[9349,4675,["l",2,9349]],
  [9371,4686,["l",2,9371]],[9377,4689,["l",2,9377]],
  [9391,4696,["l",2,9391]],[9397,4699,["l",2,9397]],
  [9403,4702,["l",2,9403]],[9413,4707,["l",2,9413]],
  [9419,4710,["l",2,9419]],[9421,4711,["l",2,9421]],
  [9431,4716,["l",2,9431]],[9433,4717,["l",2,9433]],
  [9437,4719,["l",2,9437]],[9439,4720,["l",2,9439]],
  [9461,4731,["l",2,9461]],[9463,4732,["l",2,9463]],
  [9467,4734,["l",2,9467]],[9473,4737,["l",2,9473]],
  [9479,4740,["l",2,9479]],[9491,4746,["l",2,9491]],
  [9497,4749,["l",2,9497]],[9511,4756,["l",2,9511]],
  [9521,4761,["l",2,9521]],[9533,4767,["l",2,9533]],
  [9539,4770,["l",2,9539]],[9547,4774,["l",2,9547]],
  [9551,4776,["l",2,9551]],[9587,4794,["l",2,9587]],
  [9601,4801,["l",2,9601]],[9613,4807,["l",2,9613]],
  [9619,4810,["l",2,9619]],[9623,4812,["l",2,9623]],
  [9629,4815,["l",2,9629]],[9631,4816,["l",2,9631]],
  [9643,4822,["l",2,9643]],[9649,4825,["l",2,9649]],
  [9661,4831,["l",2,9661]],[9677,4839,["l",2,9677]],
  [9679,4840,["l",2,9679]],[9689,4845,["l",2,9689]],
  [9697,4849,["l",2,9697]],[9719,4860,["l",2,9719]],
  [9721,4861,["l",2,9721]],[9733,4867,["l",2,9733]],
  [9739,4870,["l",2,9739]],[9743,4872,["l",2,9743]],
  [9749,4875,["l",2,9749]],[9767,4884,["l",2,9767]],
  [9769,4885,["l",2,9769]],[9781,4891,["l",2,9781]],
  [9787,4894,["l",2,9787]],[9791,4896,["l",2,9791]],
  [9803,4902,["l",2,9803]],[9811,4906,["l",2,9811]],
  [9817,4909,["l",2,9817]],[9829,4915,["l",2,9829]],
  [9833,4917,["l",2,9833]],[9839,4920,["l",2,9839]],
  [9851,4926,["l",2,9851]],[9857,4929,["l",2,9857]],
  [9859,4930,["l",2,9859]],[9871,4936,["l",2,9871]],
  [9883,4942,["l",2,9883]],[9887,4944,["l",2,9887]],
  [9901,4951,["l",2,9901]],[9907,4954,["l",2,9907]],
  [9923,4962,["l",2,9923]],[9929,4965,["l",2,9929]],
  [9931,4966,["l",2,9931]],[9941,4971,["l",2,9941]],
  [9949,4975,["l",2,9949]],[9967,4984,["l",2,9967]],
  [9973,4987,["l",2,9973]] ];

# This is the same as above, however the fourth position now holds
# the minimum degree this group occurs:

RECOG.TableOfTwoLargestGroupElementOrdersWithDegreeBound :=
[ [3,2,["l",2,5],2],[4,3,["l",2,7],3],
  [5,3,["l",2,4],2],[5,3,["l",2,5],2],
  [5,3,["s",4,2],4],[5,4,["l",2,9],4],
  [5,4,["s",4,2],4],[5,4,["s",4,3],9],
  [6,5,["l",2,11],5],[6,5,["s",4,2],4],
  [6,5,["s",4,3],9],[7,4,["l",2,7],3],
  [7,4,["l",3,2],8],[7,5,["l",3,4],19],
  [7,6,["l",2,13],6],[8,7,["u",3,3],9],
  [8,7,["u",3,5],25],[8,7,["u",4,3],20],
  [9,5,["s",4,3],9],[9,5,["u",4,2],16],
  [9,6,["s",4,3],9],[9,6,["u",4,2],16],
  [9,7,["l",2,8],4],[9,8,["l",2,17],8],
  [9,8,["u",4,3],20],[10,8,["u",3,5],25],
  [10,9,["l",2,19],9],[11,6,["l",2,11],5],
  [11,8,["M11"],5],[11,8,["M22"],6],[11,10,["M12"],6],
  [12,5,["s",4,3],9],[12,6,["s",4,3],9],
  [12,8,["u",3,3],9],[12,8,["u",4,3],20],
  [12,9,["s",4,3],9],[12,9,["u",4,2],16],
  [12,9,["u",4,3],20],[12,11,["l",2,23],11],
  [13,5,["2F4",2],16],[13,6,["2F4",2],16],
  [13,7,["2B2",8],14],[13,7,["l",2,13],6],
  [13,8,["2F4",2],16],[13,8,["G2",3],90],
  [13,8,["l",3,3],11],[13,9,["G2",3],90],
  [13,10,["2F4",2],16],[13,12,["2F4",2],16],
  [13,12,["G2",3],90],[13,12,["l",2,25],12],
  [13,12,["s",4,5],25],[14,13,["l",2,27],13],
  [15,7,["l",4,2],16],[15,9,["o+",8,2],35],
  [15,9,["s",6,2],8],[15,10,["o+",8,2],35],
  [15,10,["s",6,2],8],[15,11,["u",5,2],32],
  [15,11,["u",6,2],64],[15,12,["o+",8,2],35],
  [15,12,["s",6,2],8],[15,12,["u",5,2],32],
  [15,12,["u",6,2],64],[15,12,["J2"],6],
  [15,13,["s",4,5],25],[15,13,["u",3,4],16],
  [15,14,["l",2,29],14],[16,13,["2F4",2],16],
  [16,15,["l",2,31],15],[17,9,["l",2,17],8],
  [17,15,["l",2,16],8],[17,15,["s",4,4],18],
  [18,15,["u",5,2],32],[18,15,["u",6,2],64],
  [19,10,["l",2,19],9],[19,15,["J1"],7],
  [19,16,["l",3,7],55],[19,17,["J3"],9],
  [19,18,["l",2,37],18],[20,13,["l",4,3],38],
  [20,13,["s",4,5],25],[20,14,["o",7,3],78],
  [20,14,["o+",8,3],260],[20,14,["s",6,3],13],
  [20,15,["o",7,3],78],[20,15,["o+",8,3],260],
  [20,15,["s",4,5],25],[20,15,["s",6,3],13],
  [20,15,["HS"],20],[20,18,["o",7,3],78],
  [20,18,["o+",8,3],260],[20,18,["s",6,3],13],
  [21,13,["3D4",2],25],[21,14,["3D4",2],25],
  [21,15,["G2",4],63],[21,17,["F4",2],448],
  [21,17,["o-",8,2],33],[21,17,["s",8,2],35],
  [21,18,["3D4",2],25],[21,18,["F4",2],448],
  [21,18,["s",8,2],35],[21,19,["u",3,8],64],
  [21,20,["F4",2],448],[21,20,["l",2,41],20],
  [21,20,["s",8,2],35],[22,21,["l",2,43],21],
  [23,12,["l",2,23],11],[23,15,["M23"],11],
  [23,21,["M24"],11],[24,20,["s",6,3],13],
  [24,21,["F4",2],448],[24,21,["s",8,2],35],
  [24,21,["Suz"],12],[24,23,["l",2,47],23],
  [25,24,["l",2,49],24],[25,24,["s",4,7],49],
  [27,26,["l",2,53],26],[28,21,["3D4",2],25],
  [28,21,["F4",2],448],[28,21,["He"],50],
  [28,24,["F4",2],448],[28,25,["s",4,7],49],
  [29,15,["l",2,29],14],[29,26,["Ru"],28],
  [30,13,["s",4,5],25],[30,15,["s",4,5],25],
  [30,15,["McL"],21],[30,20,["s",4,5],25],
  [30,20,["s",6,3],13],[30,21,["F4",2],448],
  [30,21,["o-",8,2],33],[30,21,["s",8,2],35],
  [30,24,["F4",2],448],[30,24,["s",6,3],13],
  [30,24,["s",8,2],35],[30,24,["Co3"],22],
  [30,24,["Fi22"],27],[30,28,["F4",2],448],
  [30,28,["Co2"],22],[30,29,["l",2,59],29],
  [31,16,["l",2,31],15],[31,21,["l",5,2],32],
  [31,24,["G2",5],124],[31,24,["l",3,5],29],
  [31,25,["G2",5],124],[31,28,["O'N"],45],
  [31,30,["G2",5],124],[31,30,["l",2,61],30],
  [33,31,["l",2,32],16],[34,33,["l",2,67],33],
  [35,33,["2E6",2],1536],[35,33,["o-",10,2],153],
  [36,20,["s",6,3],13],[36,24,["s",6,3],13],
  [36,30,["s",6,3],13],[36,35,["l",2,71],35],
  [37,19,["l",2,37],18],[37,26,["2G2",27],702],
  [37,36,["l",2,73],36],[39,31,["l",4,5],154],
  [39,36,["Th"],248],[40,35,["HN"],132],
  [40,37,["u",3,11],121],[40,39,["l",2,79],39],
  [41,21,["l",2,41],20],[41,31,["2B2",32],124],
  [41,40,["l",2,81],40],[41,40,["s",4,9],81],
  [42,25,["s",4,7],49],[42,28,["s",4,7],49],
  [42,41,["l",2,83],41],[43,22,["l",2,43],21],
  [44,40,["u",3,11],121],[45,44,["l",2,89],44],
  [47,24,["l",2,47],23],[48,43,["u",3,7],49],
  [49,48,["l",2,97],48],[51,45,["o+",10,2],155],
  [51,45,["s",10,2],155],[51,50,["l",2,101],50],
  [52,41,["o",9,3],780],[52,41,["o-",8,3],245],
  [52,41,["s",8,3],40],[52,42,["o",9,3],780],
  [52,42,["s",8,3],40],[52,45,["o",9,3],780],
  [52,45,["s",8,3],40],[52,51,["l",2,103],51],
  [53,27,["l",2,53],26],[54,53,["l",2,107],53],
  [55,54,["l",2,109],54],[56,25,["s",4,7],49],
  [56,28,["s",4,7],49],[56,42,["s",4,7],49],
  [56,48,["u",3,7],49],[57,48,["G2",7],342],
  [57,49,["G2",7],342],[57,56,["G2",7],342],
  [57,56,["l",2,113],56],[59,30,["l",2,59],29],
  [60,42,["Co1"],24],[60,42,["Fi23"],253],
  [60,45,["Fi24'"],781],[60,51,["o+",10,2],155],
  [60,51,["s",10,2],155],[60,52,["o",9,3],780],
  [60,52,["o-",8,3],245],[60,52,["s",8,3],40],
  [61,31,["l",2,61],30],[61,56,["l",3,13],181],
  [61,60,["l",2,121],60],[61,60,["s",4,11],121],
  [63,31,["l",6,2],64],[63,45,["u",7,2],128],
  [63,52,["u",4,5],104],[63,60,["u",4,5],104],
  [63,60,["u",7,2],128],[63,62,["l",2,125],62],
  [64,63,["l",2,127],63],[65,51,["u",4,4],51],
  [65,51,["u",5,4],205],[65,63,["l",2,64],32],
  [65,63,["s",4,8],196],[66,44,["J4"],112],
  [66,61,["s",4,11],121],[66,63,["u",7,2],128],
  [66,65,["l",2,131],65],[67,34,["l",2,67],33],
  [67,42,["Ly"],111],[69,68,["l",2,137],68],
  [70,66,["B"],4370],[70,69,["l",2,139],69],
  [71,36,["l",2,71],35],[72,52,["s",8,3],40],
  [72,60,["s",8,3],40],[73,37,["l",2,73],36],
  [73,63,["G2",8],511],[73,63,["l",3,8],71],
  [75,74,["l",2,149],74],[76,75,["l",2,151],75],
  [78,52,["s",8,3],40],[78,60,["s",8,3],40],
  [78,63,["o",7,5],649],[78,63,["s",6,5],62],
  [78,65,["o",7,5],649],[78,65,["s",6,5],62],
  [78,72,["s",8,3],40],[79,40,["l",2,79],39],
  [79,78,["l",2,157],78],[80,61,["u",5,3],61],
  [80,65,["o-",10,3],2378],[80,73,["u",3,9],81],
  [82,81,["l",2,163],81],[83,42,["l",2,83],41],
  [84,52,["s",8,3],40],[84,60,["s",8,3],40],
  [84,72,["s",8,3],40],[84,78,["s",8,3],40],
  [84,80,["o-",10,3],2378],[84,80,["u",5,3],61],
  [84,83,["l",2,167],83],[85,63,["l",4,4],83],
  [85,65,["s",6,4],378],[85,84,["l",2,169],84],
  [85,84,["s",4,13],169],[86,75,["u",4,7],300],
  [86,84,["u",4,7],300],[87,86,["l",2,173],86],
  [89,45,["l",2,89],44],[90,52,["s",8,3],40],
  [90,60,["s",8,3],40],[90,72,["s",8,3],40],
  [90,78,["s",8,3],40],[90,84,["s",8,3],40],
  [90,89,["l",2,179],89],[91,80,["l",3,9],89],
  [91,85,["l",3,16],271],[91,85,["s",4,13],169],
  [91,90,["l",2,181],90],[93,65,["o-",12,2],649],
  [93,84,["o-",12,2],649],[93,91,["E6",2],1536],
  [96,91,["u",3,17],289],[96,95,["l",2,191],95],
  [97,49,["l",2,97],48],[97,96,["l",2,193],96],
  [99,98,["l",2,197],98],[100,99,["l",2,199],99],
  [101,51,["l",2,101],50],[102,93,["o-",12,2],649],
  [102,96,["u",3,17],289],[103,52,["l",2,103],51],
  [104,73,["3D4",3],218],[104,78,["3D4",3],218],
  [104,84,["3D4",3],218],[105,85,["o+",12,2],651],
  [105,93,["s",12,2],651],[105,102,["s",12,2],651],
  [106,105,["l",2,211],105],[107,54,["l",2,107],53],
  [109,55,["l",2,109],54],[110,61,["s",4,11],121],
  [110,66,["s",4,11],121],[112,111,["l",2,223],111],
  [113,57,["l",2,113],56],[114,113,["l",2,227],113],
  [115,114,["l",2,229],114],[117,116,["l",2,233],116],
  [119,110,["M"]],[120,78,["s",6,5],62],
  [120,105,["s",12,2],651],[120,119,["l",2,239],119],
  [121,104,["l",5,3],119],[121,120,["l",2,241],120],
  [122,91,["u",6,3],182],[122,120,["u",6,3],182],
  [122,121,["l",2,243],121],[126,93,["E6",2],1536],
  [126,125,["l",2,251],125],[127,64,["l",2,127],63],
  [127,105,["l",7,2],128],[127,120,["l",3,19],379],
  [129,99,["u",8,2],256],[129,99,["u",9,2],512],
  [129,126,["u",8,2],256],[129,126,["u",9,2],512],
  [129,127,["l",2,128],64],[129,128,["l",2,257],128],
  [130,78,["s",6,5],62],[130,120,["s",6,5],62],
  [131,66,["l",2,131],65],[132,61,["s",4,11],121],
  [132,66,["s",4,11],121],[132,110,["s",4,11],121],
  [132,129,["u",8,2],256],[132,129,["u",9,2],512],
  [132,131,["l",2,263],131],[133,120,["G2",11],1330],
  [133,120,["l",3,11],131],[133,121,["G2",11],1330],
  [133,132,["G2",11],1330],[135,134,["l",2,269],134],
  [136,135,["l",2,271],135],[137,69,["l",2,137],68],
  [139,70,["l",2,139],69],[139,138,["l",2,277],138],
  [141,140,["l",2,281],140],[142,141,["l",2,283],141],
  [145,127,["2B2",128],1016],[145,144,["l",2,289],144],
  [145,144,["s",4,17],289],[147,146,["l",2,293],146],
  [149,75,["l",2,149],74],[151,76,["l",2,151],75],
  [153,145,["s",4,17],289],[154,153,["l",2,307],153],
  [156,78,["o+",8,5],3380],[156,85,["s",4,13],169],
  [156,91,["s",4,13],169],[156,155,["l",2,311],155],
  [157,79,["l",2,157],78],[157,156,["l",2,313],156],
  [159,158,["l",2,317],158],[163,82,["l",2,163],81],
  [164,140,["o+",10,3],2420],[164,140,["s",10,3],121],
  [164,156,["o+",10,3],2420],[164,156,["s",10,3],121],
  [166,165,["l",2,331],165],[167,84,["l",2,167],83],
  [168,157,["u",3,13],169],[169,168,["l",2,337],168],
  [172,171,["l",2,343],171],[173,87,["l",2,173],86],
  [174,173,["l",2,347],173],[175,174,["l",2,349],174],
  [176,169,["u",3,23],529],[177,176,["l",2,353],176],
  [179,90,["l",2,179],89],[180,164,["s",10,3],121],
  [180,179,["l",2,359],179],[181,91,["l",2,181],90],
  [181,180,["l",2,361],180],[181,180,["s",4,19],361],
  [182,85,["s",4,13],169],[182,91,["s",4,13],169],
  [182,121,["l",6,3],362],[182,156,["s",4,13],169],
  [182,168,["u",3,13],169],[183,168,["G2",13],2196],
  [183,169,["G2",13],2196],[183,182,["G2",13],2196],
  [184,176,["u",3,23],529],[184,183,["l",2,367],183],
  [187,186,["l",2,373],186],[190,181,["s",4,19],361],
  [190,189,["l",2,379],189],[191,96,["l",2,191],95],
  [192,191,["l",2,383],191],[193,97,["l",2,193],96],
  [195,165,["o+",14,2],2667],[195,186,["o+",14,2],2667],
  [195,194,["l",2,389],194],[197,99,["l",2,197],98],
  [199,100,["l",2,199],99],[199,198,["l",2,397],198],
  [200,171,["l",4,7],398],[200,172,["o",7,7],2449],
  [200,172,["s",6,7],171],[200,175,["o",7,7],2449],
  [200,175,["s",6,7],171],[201,200,["l",2,401],200],
  [204,195,["o+",14,2],2667],[205,182,["l",4,9],818],
  [205,204,["l",2,409],204],[210,209,["l",2,419],209],
  [211,106,["l",2,211],105],[211,210,["l",2,421],210],
  [216,215,["l",2,431],215],[217,208,["l",3,25],649],
  [217,216,["l",2,433],216],[220,219,["l",2,439],219],
  [222,221,["l",2,443],221],[223,112,["l",2,223],111],
  [225,224,["l",2,449],224],[227,114,["l",2,227],113],
  [229,115,["l",2,229],114],[229,228,["l",2,457],228],
  [231,230,["l",2,461],230],[232,231,["l",2,463],231],
  [233,117,["l",2,233],116],[234,164,["s",10,3],121],
  [234,180,["s",10,3],121],[234,233,["l",2,467],233],
  [239,120,["l",2,239],119],[240,164,["s",10,3],121],
  [240,180,["s",10,3],121],[240,234,["s",10,3],121],
  [240,239,["l",2,479],239],[241,121,["l",2,241],120],
  [244,243,["l",2,487],243],[246,164,["s",10,3],121],
  [246,180,["s",10,3],121],[246,234,["s",10,3],121],
  [246,240,["s",10,3],121],[246,245,["l",2,491],245],
  [250,249,["l",2,499],249],[251,126,["l",2,251],125],
  [252,164,["s",10,3],121],[252,180,["s",10,3],121],
  [252,234,["s",10,3],121],[252,240,["s",10,3],121],
  [252,246,["s",10,3],121],[252,251,["l",2,503],251],
  [255,85,["o+",8,4],1156],[255,155,["o-",14,2],2665],
  [255,195,["s",14,2],2667],[255,204,["s",14,2],2667],
  [255,210,["o-",14,2],2665],[255,210,["s",14,2],2667],
  [255,217,["l",8,2],256],[255,241,["u",3,16],256],
  [255,254,["l",2,509],254],[257,129,["l",2,257],128],
  [257,255,["l",2,256],128],[257,255,["s",4,16],1800],
  [261,260,["l",2,521],260],[262,261,["l",2,523],261],
  [263,132,["l",2,263],131],[265,264,["l",2,529],264],
  [265,264,["s",4,23],529],[269,135,["l",2,269],134],
  [271,136,["l",2,271],135],[271,270,["l",2,541],270],
  [272,145,["s",4,17],289],[272,153,["s",4,17],289],
  [273,255,["G2",16],4095],[274,273,["l",2,547],273],
  [276,265,["s",4,23],529],[277,139,["l",2,277],138],
  [279,278,["l",2,557],278],[280,271,["u",3,29],841],
  [281,141,["l",2,281],140],[282,281,["l",2,563],281],
  [283,142,["l",2,283],141],[285,284,["l",2,569],284],
  [286,285,["l",2,571],285],[289,288,["l",2,577],288],
  [290,280,["u",3,29],841],[293,147,["l",2,293],146],
  [294,293,["l",2,587],293],[297,296,["l",2,593],296],
  [300,299,["l",2,599],299],[301,300,["l",2,601],300],
  [304,303,["l",2,607],303],[306,145,["s",4,17],289],
  [306,153,["s",4,17],289],[306,272,["s",4,17],289],
  [307,154,["l",2,307],153],[307,288,["G2",17],4912],
  [307,288,["l",3,17],305],[307,289,["G2",17],4912],
  [307,306,["G2",17],4912],[307,306,["l",2,613],306],
  [309,308,["l",2,617],308],[310,309,["l",2,619],309],
  [311,156,["l",2,311],155],[313,157,["l",2,313],156],
  [313,312,["l",2,625],312],[313,312,["s",4,25],625],
  [315,241,["3D4",4],963],[315,257,["o-",8,4],1027],
  [316,315,["l",2,631],315],[317,159,["l",2,317],158],
  [321,320,["l",2,641],320],[322,321,["l",2,643],321],
  [324,323,["l",2,647],323],[327,326,["l",2,653],326],
  [330,329,["l",2,659],329],[331,166,["l",2,331],165],
  [331,320,["l",3,31],991],[331,330,["l",2,661],330],
  [333,305,["u",4,11],1220],[333,330,["u",4,11],1220],
  [336,200,["s",6,7],171],[337,169,["l",2,337],168],
  [337,336,["l",2,673],336],[339,338,["l",2,677],338],
  [341,315,["l",5,4],339],[341,331,["u",3,32],1024],
  [342,181,["s",4,19],361],[342,190,["s",4,19],361],
  [342,341,["l",2,683],341],[346,345,["l",2,691],345],
  [347,174,["l",2,347],173],[349,175,["l",2,349],174],
  [350,200,["s",6,7],171],[350,336,["s",6,7],171],
  [351,350,["l",2,701],350],[353,177,["l",2,353],176],
  [355,354,["l",2,709],354],[359,180,["l",2,359],179],
  [360,343,["u",3,19],361],[360,359,["l",2,719],359],
  [364,363,["l",2,727],363],[365,328,["u",4,9],656],
  [365,364,["l",2,729],364],[365,364,["s",4,27],729],
  [367,184,["l",2,367],183],[367,366,["l",2,733],366],
  [370,369,["l",2,739],369],[372,313,["o-",8,5],3129],
  [372,313,["s",8,5],312],[372,315,["s",8,5],312],
  [372,371,["l",2,743],371],[373,187,["l",2,373],186],
  [376,375,["l",2,751],375],[379,190,["l",2,379],189],
  [379,378,["l",2,757],378],[380,181,["s",4,19],361],
  [380,190,["s",4,19],361],[380,342,["s",4,19],361],
  [380,360,["u",3,19],361],[381,380,["l",2,761],380],
  [383,192,["l",2,383],191],[385,384,["l",2,769],384],
  [387,386,["l",2,773],386],[389,195,["l",2,389],194],
  [390,372,["o-",8,5],3129],[390,372,["s",8,5],312],
  [394,393,["l",2,787],393],[397,199,["l",2,397],198],
  [399,398,["l",2,797],398],[401,201,["l",2,401],200],
  [405,404,["l",2,809],404],[406,405,["l",2,811],405],
  [409,205,["l",2,409],204],[410,365,["s",6,9],364],
  [411,410,["l",2,821],410],[412,411,["l",2,823],411],
  [414,413,["l",2,827],413],[415,414,["l",2,829],414],
  [419,210,["l",2,419],209],[420,419,["l",2,839],419],
  [421,211,["l",2,421],210],[421,420,["l",2,841],420],
  [421,420,["s",4,29],841],[427,426,["l",2,853],426],
  [429,428,["l",2,857],428],[430,429,["l",2,859],429],
  [431,216,["l",2,431],215],[432,431,["l",2,863],431],
  [433,217,["l",2,433],216],[435,421,["s",4,29],841],
  [439,220,["l",2,439],219],[439,438,["l",2,877],438],
  [441,440,["l",2,881],440],[442,441,["l",2,883],441],
  [443,222,["l",2,443],221],[444,443,["l",2,887],443],
  [449,225,["l",2,449],224],[454,453,["l",2,907],453],
  [455,341,["l",6,4],1363],[456,455,["l",2,911],455],
  [457,229,["l",2,457],228],[460,459,["l",2,919],459],
  [461,231,["l",2,461],230],[463,232,["l",2,463],231],
  [465,464,["l",2,929],464],[467,234,["l",2,467],233],
  [469,456,["l",3,37],1405],[469,468,["l",2,937],468],
  [471,470,["l",2,941],470],[474,473,["l",2,947],473],
  [477,476,["l",2,953],476],[479,240,["l",2,479],239],
  [481,480,["l",2,961],480],[481,480,["s",4,31],961],
  [484,410,["s",12,3],364],[484,420,["s",12,3],364],
  [484,468,["s",12,3],364],[484,483,["l",2,967],483],
  [486,485,["l",2,971],485],[487,244,["l",2,487],243],
  [489,488,["l",2,977],488],[491,246,["l",2,491],245],
  [492,484,["s",12,3],364],[492,491,["l",2,983],491],
  [496,481,["s",4,31],961],[496,495,["l",2,991],495],
  [499,250,["l",2,499],249],[499,498,["l",2,997],498],
  [503,252,["l",2,503],251],[505,504,["l",2,1009],504],
  [506,265,["s",4,23],529],[506,276,["s",4,23],529],
  [507,506,["l",2,1013],506],[509,255,["l",2,509],254],
  [510,509,["l",2,1019],509],[511,465,["l",9,2],512],
  [511,510,["l",2,1021],510],[513,387,["u",10,2],1024],
  [513,455,["u",4,8],455],[513,510,["u",10,2],1024],
  [513,511,["l",2,512],256],[516,513,["u",10,2],1024],
  [516,515,["l",2,1031],515],[517,516,["l",2,1033],516],
  [520,519,["l",2,1039],519],[521,261,["l",2,521],260],
  [523,262,["l",2,523],261],[525,524,["l",2,1049],524],
  [526,525,["l",2,1051],525],[531,530,["l",2,1061],530],
  [532,531,["l",2,1063],531],[535,534,["l",2,1069],534],
  [541,271,["l",2,541],270],[544,543,["l",2,1087],543],
  [546,545,["l",2,1091],545],[547,274,["l",2,547],273],
  [547,546,["l",2,1093],546],[549,548,["l",2,1097],548],
  [552,265,["s",4,23],529],[552,276,["s",4,23],529],
  [552,506,["s",4,23],529],[552,551,["l",2,1103],551],
  [553,528,["l",3,23],551],[555,554,["l",2,1109],554],
  [557,279,["l",2,557],278],[559,558,["l",2,1117],558],
  [560,547,["u",3,41],1681],[562,561,["l",2,1123],561],
  [563,282,["l",2,563],281],[565,564,["l",2,1129],564],
  [569,285,["l",2,569],284],[571,286,["l",2,571],285],
  [574,560,["u",3,41],1681],[576,575,["l",2,1151],575],
  [577,289,["l",2,577],288],[577,576,["l",2,1153],576],
  [582,581,["l",2,1163],581],[585,511,["l",4,8],583],
  [586,585,["l",2,1171],585],[587,294,["l",2,587],293],
  [591,590,["l",2,1181],590],[593,297,["l",2,593],296],
  [594,593,["l",2,1187],593],[595,549,["l",4,13],2378],
  [597,596,["l",2,1193],596],[599,300,["l",2,599],299],
  [601,301,["l",2,601],300],[601,600,["l",2,1201],600],
  [607,304,["l",2,607],303],[607,606,["l",2,1213],606],
  [609,608,["l",2,1217],608],[612,611,["l",2,1223],611],
  [613,307,["l",2,613],306],[615,614,["l",2,1229],614],
  [616,615,["l",2,1231],615],[617,309,["l",2,617],308],
  [619,310,["l",2,619],309],[619,618,["l",2,1237],618],
  [620,372,["s",8,5],312],[620,390,["s",8,5],312],
  [624,521,["u",5,5],521],[624,521,["u",6,5],2604],
  [624,601,["u",3,25],625],[625,624,["l",2,1249],624],
  [630,372,["s",8,5],312],[630,390,["s",8,5],312],
  [630,620,["s",8,5],312],[630,624,["u",5,5],521],
  [630,624,["u",6,5],2604],[630,629,["l",2,1259],629],
  [631,316,["l",2,631],315],[631,616,["l",3,43],1891],
  [639,638,["l",2,1277],638],[640,639,["l",2,1279],639],
  [641,321,["l",2,641],320],[642,641,["l",2,1283],641],
  [643,322,["l",2,643],321],[645,644,["l",2,1289],644],
  [646,645,["l",2,1291],645],[647,324,["l",2,647],323],
  [649,648,["l",2,1297],648],[651,650,["l",2,1301],650],
  [652,651,["l",2,1303],651],[653,327,["l",2,653],326],
  [654,653,["l",2,1307],653],[659,330,["l",2,659],329],
  [660,659,["l",2,1319],659],[661,331,["l",2,661],330],
  [661,660,["l",2,1321],660],[664,663,["l",2,1327],663],
  [666,665,["l",2,1331],665],[673,337,["l",2,673],336],
  [677,339,["l",2,677],338],[681,680,["l",2,1361],680],
  [683,342,["l",2,683],341],[684,683,["l",2,1367],683],
  [685,684,["l",2,1369],684],[685,684,["s",4,37],1369],
  [687,686,["l",2,1373],686],[691,346,["l",2,691],345],
  [691,690,["l",2,1381],690],[700,699,["l",2,1399],699],
  [701,351,["l",2,701],350],[703,685,["s",4,37],1369],
  [705,704,["l",2,1409],704],[709,355,["l",2,709],354],
  [712,711,["l",2,1423],711],[714,713,["l",2,1427],713],
  [715,714,["l",2,1429],714],[717,716,["l",2,1433],716],
  [719,360,["l",2,719],359],[720,484,["s",12,3],364],
  [720,492,["s",12,3],364],[720,719,["l",2,1439],719],
  [724,723,["l",2,1447],723],[726,484,["s",12,3],364],
  [726,492,["s",12,3],364],[726,720,["s",12,3],364],
  [726,725,["l",2,1451],725],[727,364,["l",2,727],363],
  [727,726,["l",2,1453],726],[728,560,["u",7,3],547],
  [728,560,["u",8,3],1640],[728,703,["u",3,27],729],
  [730,729,["l",2,1459],729],[732,484,["s",12,3],364],
  [732,492,["s",12,3],364],[732,665,["l",4,11],1462],
  [732,666,["s",6,11],665],[732,671,["s",6,11],665],
  [732,720,["s",12,3],364],[732,726,["s",12,3],364],
  [732,728,["u",7,3],547],[732,728,["u",8,3],1640],
  [733,367,["l",2,733],366],[736,721,["u",3,47],2209],
  [736,735,["l",2,1471],735],[738,484,["s",12,3],364],
  [738,492,["s",12,3],364],[738,720,["s",12,3],364],
  [738,726,["s",12,3],364],[738,732,["s",12,3],364],
  [739,370,["l",2,739],369],[741,740,["l",2,1481],740],
  [742,741,["l",2,1483],741],[743,372,["l",2,743],371],
  [744,601,["3D4",5],3004],[744,620,["3D4",5],3004],
  [744,630,["3D4",5],3004],[744,743,["l",2,1487],743],
  [745,744,["l",2,1489],744],[747,746,["l",2,1493],746],
  [750,749,["l",2,1499],749],[751,376,["l",2,751],375],
  [752,736,["u",3,47],2209],[756,755,["l",2,1511],755],
  [757,379,["l",2,757],378],[757,728,["l",3,27],755],
  [761,381,["l",2,761],380],[762,761,["l",2,1523],761],
  [766,765,["l",2,1531],765],[769,385,["l",2,769],384],
  [772,771,["l",2,1543],771],[773,387,["l",2,773],386],
  [775,774,["l",2,1549],774],[777,776,["l",2,1553],776],
  [780,779,["l",2,1559],779],[781,744,["l",5,5],779],
  [784,783,["l",2,1567],783],[786,785,["l",2,1571],785],
  [787,394,["l",2,787],393],[790,789,["l",2,1579],789],
  [792,791,["l",2,1583],791],[797,399,["l",2,797],398],
  [799,798,["l",2,1597],798],[801,800,["l",2,1601],800],
  [804,803,["l",2,1607],803],[805,804,["l",2,1609],804],
  [807,806,["l",2,1613],806],[809,405,["l",2,809],404],
  [810,809,["l",2,1619],809],[811,406,["l",2,811],405],
  [811,810,["l",2,1621],810],[812,421,["s",4,29],841],
  [812,435,["s",4,29],841],[814,813,["l",2,1627],813],
  [817,800,["l",3,49],2449],[819,818,["l",2,1637],818],
  [821,411,["l",2,821],410],[823,412,["l",2,823],411],
  [827,414,["l",2,827],413],[829,415,["l",2,829],414],
  [829,828,["l",2,1657],828],[832,831,["l",2,1663],831],
  [834,833,["l",2,1667],833],[835,834,["l",2,1669],834],
  [839,420,["l",2,839],419],[841,840,["l",2,1681],840],
  [841,840,["s",4,41],1681],[847,846,["l",2,1693],846],
  [849,848,["l",2,1697],848],[850,849,["l",2,1699],849],
  [853,427,["l",2,853],426],[855,854,["l",2,1709],854],
  [857,429,["l",2,857],428],[859,430,["l",2,859],429],
  [861,841,["s",4,41],1681],[861,860,["l",2,1721],860],
  [862,861,["l",2,1723],861],[863,432,["l",2,863],431],
  [867,866,["l",2,1733],866],[870,421,["s",4,29],841],
  [870,435,["s",4,29],841],[870,812,["s",4,29],841],
  [871,840,["l",3,29],869],[871,870,["l",2,1741],870],
  [874,873,["l",2,1747],873],[877,439,["l",2,877],438],
  [877,876,["l",2,1753],876],[880,879,["l",2,1759],879],
  [881,441,["l",2,881],440],[883,442,["l",2,883],441],
  [887,444,["l",2,887],443],[889,888,["l",2,1777],888],
  [892,891,["l",2,1783],891],[894,893,["l",2,1787],893],
  [895,894,["l",2,1789],894],[901,900,["l",2,1801],900],
  [906,905,["l",2,1811],905],[907,454,["l",2,907],453],
  [911,456,["l",2,911],455],[912,911,["l",2,1823],911],
  [916,915,["l",2,1831],915],[919,460,["l",2,919],459],
  [924,923,["l",2,1847],923],[925,924,["l",2,1849],924],
  [925,924,["s",4,43],1849],[929,465,["l",2,929],464],
  [930,481,["s",4,31],961],[930,496,["s",4,31],961],
  [931,930,["l",2,1861],930],[934,933,["l",2,1867],933],
  [936,919,["u",3,53],2809],[936,935,["l",2,1871],935],
  [937,469,["l",2,937],468],[937,936,["l",2,1873],936],
  [939,938,["l",2,1877],938],[940,939,["l",2,1879],939],
  [941,471,["l",2,941],470],[945,944,["l",2,1889],944],
  [946,925,["s",4,43],1849],[947,474,["l",2,947],473],
  [951,950,["l",2,1901],950],[953,477,["l",2,953],476],
  [954,936,["u",3,53],2809],[954,953,["l",2,1907],953],
  [957,956,["l",2,1913],956],[960,931,["u",3,31],961],
  [966,965,["l",2,1931],965],[967,484,["l",2,967],483],
  [967,966,["l",2,1933],966],[971,486,["l",2,971],485],
  [975,974,["l",2,1949],974],[976,975,["l",2,1951],975],
  [977,489,["l",2,977],488],[983,492,["l",2,983],491],
  [987,986,["l",2,1973],986],[990,989,["l",2,1979],989],
  [991,496,["l",2,991],495],[992,481,["s",4,31],961],
  [992,496,["s",4,31],961],[992,930,["s",4,31],961],
  [992,960,["u",3,31],961],[994,993,["l",2,1987],993],
  [997,499,["l",2,997],498],[997,996,["l",2,1993],996],
  [999,998,["l",2,1997],998],[1000,999,["l",2,1999],999],
  [1002,1001,["l",2,2003],1001],
  [1006,1005,["l",2,2011],1005],
  [1009,505,["l",2,1009],504],
  [1009,1008,["l",2,2017],1008],
  [1013,507,["l",2,1013],506],
  [1014,1013,["l",2,2027],1013],
  [1015,1014,["l",2,2029],1014],
  [1019,510,["l",2,1019],509],
  [1020,1019,["l",2,2039],1019],
  [1021,511,["l",2,1021],510],[1023,765,["u",11,2],2048],
  [1023,765,["u",12,2],4096],[1023,774,["u",12,2],4096],
  [1023,889,["l",10,2],1024],[1023,1020,["u",11,2],2048],
  [1023,1020,["u",12,2],4096],[1025,819,["u",6,4],819],
  [1025,1023,["l",2,1024],512],[1026,1023,["u",11,2],2048]
    ,[1026,1023,["u",12,2],4096],
  [1027,1026,["l",2,2053],1026],
  [1031,516,["l",2,1031],515],
  [1032,1031,["l",2,2063],1031],
  [1033,517,["l",2,1033],516],
  [1035,1034,["l",2,2069],1034],
  [1039,520,["l",2,1039],519],
  [1041,1040,["l",2,2081],1040],
  [1042,1041,["l",2,2083],1041],
  [1044,1043,["l",2,2087],1043],
  [1045,1044,["l",2,2089],1044],
  [1049,525,["l",2,1049],524],
  [1050,1049,["l",2,2099],1049],
  [1051,526,["l",2,1051],525],
  [1056,1055,["l",2,2111],1055],
  [1057,1023,["l",3,32],1055],
  [1057,1056,["l",2,2113],1056],
  [1061,531,["l",2,1061],530],[1063,532,["l",2,1063],531],
  [1065,1064,["l",2,2129],1064],
  [1066,1065,["l",2,2131],1065],
  [1069,535,["l",2,1069],534],
  [1069,1068,["l",2,2137],1068],
  [1071,1070,["l",2,2141],1070],
  [1072,1071,["l",2,2143],1071],
  [1077,1076,["l",2,2153],1076],
  [1081,1080,["l",2,2161],1080],
  [1087,544,["l",2,1087],543],
  [1090,1089,["l",2,2179],1089],
  [1091,546,["l",2,1091],545],[1093,547,["l",2,1093],546],
  [1093,1040,["l",7,3],1091],[1094,1093,["l",2,2187],1093]
    ,[1097,549,["l",2,1097],548],
  [1099,1020,["u",4,13],2040],[1099,1092,["u",4,13],2040],
  [1099,1098,["l",2,2197],1098],
  [1102,1101,["l",2,2203],1101],
  [1103,552,["l",2,1103],551],
  [1104,1103,["l",2,2207],1103],
  [1105,1104,["l",2,2209],1104],
  [1105,1104,["s",4,47],2209],
  [1107,1106,["l",2,2213],1106],
  [1109,555,["l",2,1109],554],
  [1111,1110,["l",2,2221],1110],
  [1117,559,["l",2,1117],558],
  [1119,1118,["l",2,2237],1118],
  [1120,1119,["l",2,2239],1119],
  [1122,1121,["l",2,2243],1121],
  [1123,562,["l",2,1123],561],
  [1126,1125,["l",2,2251],1125],
  [1128,1105,["s",4,47],2209],[1129,565,["l",2,1129],564],
  [1134,1133,["l",2,2267],1133],
  [1135,1134,["l",2,2269],1134],
  [1137,1136,["l",2,2273],1136],
  [1141,1140,["l",2,2281],1140],
  [1144,1143,["l",2,2287],1143],
  [1147,1146,["l",2,2293],1146],
  [1149,1148,["l",2,2297],1148],
  [1151,576,["l",2,1151],575],[1153,577,["l",2,1153],576],
  [1155,1154,["l",2,2309],1154],
  [1156,1155,["l",2,2311],1155],
  [1160,1141,["u",3,59],3481],[1163,582,["l",2,1163],581],
  [1167,1166,["l",2,2333],1166],
  [1170,1169,["l",2,2339],1169],
  [1171,586,["l",2,1171],585],
  [1171,1170,["l",2,2341],1170],
  [1174,1173,["l",2,2347],1173],
  [1176,1175,["l",2,2351],1175],
  [1179,1178,["l",2,2357],1178],
  [1180,1160,["u",3,59],3481],[1181,591,["l",2,1181],590],
  [1186,1185,["l",2,2371],1185],
  [1187,594,["l",2,1187],593],
  [1189,1188,["l",2,2377],1188],
  [1190,1099,["s",6,13],1098],[1190,1105,["s",6,13],1098],
  [1191,1190,["l",2,2381],1190],
  [1192,1191,["l",2,2383],1191],
  [1193,597,["l",2,1193],596],
  [1195,1194,["l",2,2389],1194],
  [1197,1196,["l",2,2393],1196],
  [1200,1199,["l",2,2399],1199],
  [1201,601,["l",2,1201],600],
  [1201,1200,["l",2,2401],1200],
  [1201,1200,["s",4,49],2401],
  [1206,1205,["l",2,2411],1205],
  [1209,1208,["l",2,2417],1208],
  [1212,1211,["l",2,2423],1211],
  [1213,607,["l",2,1213],606],[1217,609,["l",2,1217],608],
  [1219,1218,["l",2,2437],1218],
  [1221,1220,["l",2,2441],1220],
  [1223,612,["l",2,1223],611],
  [1224,1223,["l",2,2447],1223],
  [1229,615,["l",2,1229],614],
  [1230,1229,["l",2,2459],1229],
  [1231,616,["l",2,1231],615],
  [1234,1233,["l",2,2467],1233],
  [1237,619,["l",2,1237],618],
  [1237,1236,["l",2,2473],1236],
  [1239,1238,["l",2,2477],1238],
  [1249,625,["l",2,1249],624],
  [1252,1251,["l",2,2503],1251],
  [1259,630,["l",2,1259],629],[1261,1240,["l",3,61],3781],
  [1261,1260,["l",2,2521],1260],
  [1266,1265,["l",2,2531],1265],
  [1270,1269,["l",2,2539],1269],
  [1272,1271,["l",2,2543],1271],
  [1275,1274,["l",2,2549],1274],
  [1276,1275,["l",2,2551],1275],
  [1277,639,["l",2,1277],638],[1279,640,["l",2,1279],639],
  [1279,1278,["l",2,2557],1278],
  [1283,642,["l",2,1283],641],[1289,645,["l",2,1289],644],
  [1290,1289,["l",2,2579],1289],
  [1291,646,["l",2,1291],645],
  [1296,1295,["l",2,2591],1295],
  [1297,649,["l",2,1297],648],
  [1297,1296,["l",2,2593],1296],
  [1301,651,["l",2,1301],650],[1303,652,["l",2,1303],651],
  [1305,1304,["l",2,2609],1304],
  [1307,654,["l",2,1307],653],
  [1309,1308,["l",2,2617],1308],
  [1311,1310,["l",2,2621],1310],
  [1317,1316,["l",2,2633],1316],
  [1319,660,["l",2,1319],659],[1320,732,["s",6,11],665],
  [1321,661,["l",2,1321],660],
  [1324,1323,["l",2,2647],1323],
  [1327,664,["l",2,1327],663],
  [1329,1328,["l",2,2657],1328],
  [1330,1329,["l",2,2659],1329],[1332,685,["s",4,37],1369]
    ,[1332,703,["s",4,37],1369],
  [1332,1331,["l",2,2663],1331],
  [1336,1335,["l",2,2671],1335],
  [1339,1338,["l",2,2677],1338],[1342,732,["s",6,11],665],
  [1342,1320,["s",6,11],665],[1342,1341,["l",2,2683],1341]
    ,[1344,1343,["l",2,2687],1343],
  [1345,1344,["l",2,2689],1344],
  [1347,1346,["l",2,2693],1346],
  [1350,1349,["l",2,2699],1349],
  [1354,1353,["l",2,2707],1353],
  [1356,1355,["l",2,2711],1355],
  [1357,1356,["l",2,2713],1356],
  [1360,1359,["l",2,2719],1359],
  [1361,681,["l",2,1361],680],
  [1365,1364,["l",2,2729],1364],
  [1366,1365,["l",2,2731],1365],
  [1367,684,["l",2,1367],683],[1368,1201,["s",8,7],1200],
  [1368,1204,["s",8,7],1200],[1368,1333,["u",3,37],1369],
  [1371,1370,["l",2,2741],1370],
  [1373,687,["l",2,1373],686],
  [1375,1374,["l",2,2749],1374],
  [1377,1376,["l",2,2753],1376],
  [1381,691,["l",2,1381],690],
  [1384,1383,["l",2,2767],1383],
  [1387,1365,["l",3,64],4159],
  [1389,1388,["l",2,2777],1388],
  [1395,1394,["l",2,2789],1394],
  [1396,1395,["l",2,2791],1395],
  [1399,700,["l",2,1399],699],
  [1399,1398,["l",2,2797],1398],[1400,1368,["s",8,7],1200]
    ,[1401,1400,["l",2,2801],1400],
  [1402,1401,["l",2,2803],1401],
  [1405,1404,["l",2,2809],1404],
  [1405,1404,["s",4,53],2809],[1406,685,["s",4,37],1369],
  [1406,703,["s",4,37],1369],[1406,1332,["s",4,37],1369],
  [1406,1368,["u",3,37],1369],[1409,705,["l",2,1409],704],
  [1410,1409,["l",2,2819],1409],
  [1417,1416,["l",2,2833],1416],
  [1419,1418,["l",2,2837],1418],
  [1422,1421,["l",2,2843],1421],
  [1423,712,["l",2,1423],711],
  [1426,1425,["l",2,2851],1425],
  [1427,714,["l",2,1427],713],[1429,715,["l",2,1429],714],
  [1429,1428,["l",2,2857],1428],
  [1431,1405,["s",4,53],2809],
  [1431,1430,["l",2,2861],1430],
  [1433,717,["l",2,1433],716],[1439,720,["l",2,1439],719],
  [1440,1439,["l",2,2879],1439],
  [1444,1443,["l",2,2887],1443],
  [1447,724,["l",2,1447],723],
  [1449,1448,["l",2,2897],1448],
  [1451,726,["l",2,1451],725],
  [1452,1451,["l",2,2903],1451],
  [1453,727,["l",2,1453],726],
  [1455,1454,["l",2,2909],1454],
  [1459,730,["l",2,1459],729],
  [1459,1458,["l",2,2917],1458],
  [1460,1220,["s",14,3],1093],[1460,1230,["s",14,3],1093],
  [1460,1260,["s",14,3],1093],[1460,1452,["s",14,3],1093],
  [1464,1463,["l",2,2927],1463],
  [1470,1469,["l",2,2939],1469],
  [1471,736,["l",2,1471],735],[1476,1460,["s",14,3],1093],
  [1477,1476,["l",2,2953],1476],
  [1479,1478,["l",2,2957],1478],
  [1481,741,["l",2,1481],740],
  [1482,1481,["l",2,2963],1481],
  [1483,742,["l",2,1483],741],
  [1485,1484,["l",2,2969],1484],
  [1486,1485,["l",2,2971],1485],
  [1487,744,["l",2,1487],743],[1489,745,["l",2,1489],744],
  [1493,747,["l",2,1493],746],[1499,750,["l",2,1499],749],
  [1500,1499,["l",2,2999],1499],
  [1501,1500,["l",2,3001],1500],
  [1506,1505,["l",2,3011],1505],
  [1510,1509,["l",2,3019],1509],
  [1511,756,["l",2,1511],755],
  [1512,1511,["l",2,3023],1511],
  [1519,1496,["l",3,67],4555],
  [1519,1518,["l",2,3037],1518],
  [1521,1520,["l",2,3041],1520],
  [1523,762,["l",2,1523],761],
  [1525,1524,["l",2,3049],1524],
  [1531,766,["l",2,1531],765],
  [1531,1530,["l",2,3061],1530],
  [1534,1533,["l",2,3067],1533],
  [1540,1539,["l",2,3079],1539],
  [1542,1541,["l",2,3083],1541],
  [1543,772,["l",2,1543],771],
  [1545,1544,["l",2,3089],1544],
  [1549,775,["l",2,1549],774],[1553,777,["l",2,1553],776],
  [1555,1554,["l",2,3109],1554],
  [1559,780,["l",2,1559],779],
  [1560,1559,["l",2,3119],1559],
  [1561,1560,["l",2,3121],1560],
  [1563,1562,["l",2,3125],1562],
  [1567,784,["l",2,1567],783],
  [1569,1568,["l",2,3137],1568],
  [1571,786,["l",2,1571],785],[1579,790,["l",2,1579],789],
  [1582,1581,["l",2,3163],1581],
  [1583,792,["l",2,1583],791],
  [1584,1583,["l",2,3167],1583],
  [1585,1584,["l",2,3169],1584],
  [1591,1590,["l",2,3181],1590],
  [1594,1593,["l",2,3187],1593],
  [1596,1595,["l",2,3191],1595],
  [1597,799,["l",2,1597],798],[1601,801,["l",2,1601],800],
  [1602,1601,["l",2,3203],1601],
  [1605,1604,["l",2,3209],1604],
  [1607,804,["l",2,1607],803],[1609,805,["l",2,1609],804],
  [1609,1608,["l",2,3217],1608],
  [1611,1610,["l",2,3221],1610],
  [1613,807,["l",2,1613],806],
  [1615,1614,["l",2,3229],1614],
  [1619,810,["l",2,1619],809],[1621,811,["l",2,1621],810],
  [1626,1625,["l",2,3251],1625],
  [1627,814,["l",2,1627],813],
  [1627,1626,["l",2,3253],1626],
  [1629,1628,["l",2,3257],1628],
  [1630,1629,["l",2,3259],1629],
  [1636,1635,["l",2,3271],1635],
  [1637,819,["l",2,1637],818],[1640,841,["s",4,41],1681],
  [1640,861,["s",4,41],1681],[1640,1573,["l",8,3],3278],
  [1650,1649,["l",2,3299],1649],
  [1651,1650,["l",2,3301],1650],
  [1654,1653,["l",2,3307],1653],
  [1657,829,["l",2,1657],828],
  [1657,1656,["l",2,3313],1656],
  [1660,1659,["l",2,3319],1659],
  [1662,1661,["l",2,3323],1661],
  [1663,832,["l",2,1663],831],
  [1665,1664,["l",2,3329],1664],
  [1666,1665,["l",2,3331],1665],
  [1667,834,["l",2,1667],833],[1669,835,["l",2,1669],834],
  [1672,1671,["l",2,3343],1671],
  [1674,1673,["l",2,3347],1673],
  [1680,1679,["l",2,3359],1679],
  [1681,1680,["l",2,3361],1680],
  [1686,1685,["l",2,3371],1685],
  [1687,1686,["l",2,3373],1686],
  [1693,847,["l",2,1693],846],
  [1695,1694,["l",2,3389],1694],
  [1696,1695,["l",2,3391],1695],
  [1697,849,["l",2,1697],848],[1699,850,["l",2,1699],849],
  [1704,1703,["l",2,3407],1703],
  [1707,1706,["l",2,3413],1706],
  [1709,855,["l",2,1709],854],
  [1717,1716,["l",2,3433],1716],
  [1721,861,["l",2,1721],860],[1722,841,["s",4,41],1681],
  [1722,861,["s",4,41],1681],[1722,1640,["s",4,41],1681],
  [1723,862,["l",2,1723],861],[1723,1680,["l",3,41],1721],
  [1725,1724,["l",2,3449],1724],
  [1729,1728,["l",2,3457],1728],
  [1731,1730,["l",2,3461],1730],
  [1732,1731,["l",2,3463],1731],
  [1733,867,["l",2,1733],866],
  [1734,1733,["l",2,3467],1733],
  [1735,1734,["l",2,3469],1734],
  [1741,871,["l",2,1741],870],
  [1741,1740,["l",2,3481],1740],
  [1741,1740,["s",4,59],3481],
  [1746,1745,["l",2,3491],1745],
  [1747,874,["l",2,1747],873],
  [1750,1749,["l",2,3499],1749],
  [1753,877,["l",2,1753],876],
  [1756,1755,["l",2,3511],1755],
  [1759,880,["l",2,1759],879],
  [1759,1758,["l",2,3517],1758],
  [1764,1763,["l",2,3527],1763],
  [1765,1764,["l",2,3529],1764],
  [1767,1766,["l",2,3533],1766],
  [1770,1741,["s",4,59],3481],
  [1770,1769,["l",2,3539],1769],
  [1771,1770,["l",2,3541],1770],
  [1774,1773,["l",2,3547],1773],
  [1777,889,["l",2,1777],888],
  [1779,1778,["l",2,3557],1778],
  [1780,1779,["l",2,3559],1779],
  [1783,892,["l",2,1783],891],
  [1786,1785,["l",2,3571],1785],
  [1787,894,["l",2,1787],893],[1789,895,["l",2,1789],894],
  [1791,1790,["l",2,3581],1790],
  [1792,1791,["l",2,3583],1791],
  [1797,1796,["l",2,3593],1796],
  [1801,901,["l",2,1801],900],
  [1804,1803,["l",2,3607],1803],[1806,925,["s",4,43],1849]
    ,[1806,946,["s",4,43],1849],
  [1807,1806,["l",2,3613],1806],
  [1809,1808,["l",2,3617],1808],
  [1811,906,["l",2,1811],905],
  [1812,1811,["l",2,3623],1811],
  [1816,1815,["l",2,3631],1815],
  [1819,1818,["l",2,3637],1818],
  [1822,1821,["l",2,3643],1821],
  [1823,912,["l",2,1823],911],
  [1830,1829,["l",2,3659],1829],
  [1831,916,["l",2,1831],915],
  [1836,1835,["l",2,3671],1835],
  [1837,1836,["l",2,3673],1836],
  [1839,1838,["l",2,3677],1838],
  [1846,1845,["l",2,3691],1845],
  [1847,924,["l",2,1847],923],[1848,1807,["u",3,43],1849],
  [1849,1848,["l",2,3697],1848],
  [1851,1850,["l",2,3701],1850],
  [1855,1854,["l",2,3709],1854],
  [1860,1859,["l",2,3719],1859],
  [1861,931,["l",2,1861],930],
  [1861,1860,["l",2,3721],1860],
  [1861,1860,["s",4,61],3721],
  [1864,1863,["l",2,3727],1863],
  [1867,934,["l",2,1867],933],
  [1867,1866,["l",2,3733],1866],
  [1870,1869,["l",2,3739],1869],
  [1871,936,["l",2,1871],935],[1873,937,["l",2,1873],936],
  [1877,939,["l",2,1877],938],[1878,1638,["s",10,5],1562],
  [1878,1860,["s",10,5],1562],[1879,940,["l",2,1879],939],
  [1881,1880,["l",2,3761],1880],
  [1884,1883,["l",2,3767],1883],
  [1885,1884,["l",2,3769],1884],
  [1889,945,["l",2,1889],944],
  [1890,1889,["l",2,3779],1889],
  [1891,1861,["s",4,61],3721],[1892,925,["s",4,43],1849],
  [1892,946,["s",4,43],1849],[1892,1806,["s",4,43],1849],
  [1892,1848,["u",3,43],1849],
  [1897,1896,["l",2,3793],1896],
  [1899,1898,["l",2,3797],1898],
  [1901,951,["l",2,1901],950],
  [1902,1901,["l",2,3803],1901],
  [1907,954,["l",2,1907],953],
  [1911,1910,["l",2,3821],1910],
  [1912,1911,["l",2,3823],1911],
  [1913,957,["l",2,1913],956],
  [1917,1916,["l",2,3833],1916],
  [1924,1923,["l",2,3847],1923],
  [1926,1925,["l",2,3851],1925],
  [1927,1926,["l",2,3853],1926],
  [1931,966,["l",2,1931],965],
  [1932,1931,["l",2,3863],1931],
  [1933,967,["l",2,1933],966],
  [1939,1938,["l",2,3877],1938],
  [1941,1940,["l",2,3881],1940],
  [1945,1944,["l",2,3889],1944],
  [1949,975,["l",2,1949],974],[1951,976,["l",2,1951],975],
  [1953,1562,["l",6,5],3904],[1954,1953,["l",2,3907],1953]
    ,[1956,1955,["l",2,3911],1955],
  [1959,1958,["l",2,3917],1958],
  [1960,1959,["l",2,3919],1959],
  [1962,1961,["l",2,3923],1961],
  [1965,1964,["l",2,3929],1964],
  [1966,1965,["l",2,3931],1965],
  [1972,1971,["l",2,3943],1971],
  [1973,987,["l",2,1973],986],
  [1974,1973,["l",2,3947],1973],
  [1979,990,["l",2,1979],989],
  [1984,1983,["l",2,3967],1983],
  [1987,994,["l",2,1987],993],[1993,997,["l",2,1993],996],
  [1995,1994,["l",2,3989],1994],
  [1997,999,["l",2,1997],998],[1999,1000,["l",2,1999],999]
    ,[2001,2000,["l",2,4001],2000],
  [2002,2001,["l",2,4003],2001],
  [2003,1002,["l",2,2003],1001],
  [2004,2003,["l",2,4007],2003],
  [2007,2006,["l",2,4013],2006],
  [2010,2009,["l",2,4019],2009],
  [2011,1006,["l",2,2011],1005],
  [2011,2010,["l",2,4021],2010],
  [2014,2013,["l",2,4027],2013],
  [2017,1009,["l",2,2017],1008],
  [2025,2024,["l",2,4049],2024],
  [2026,2025,["l",2,4051],2025],
  [2027,1014,["l",2,2027],1013],
  [2029,1015,["l",2,2029],1014],
  [2029,2028,["l",2,4057],2028],
  [2037,2036,["l",2,4073],2036],
  [2039,1020,["l",2,2039],1019],
  [2040,2039,["l",2,4079],2039],
  [2046,2045,["l",2,4091],2045],
  [2047,1953,["l",11,2],2048],
  [2047,2046,["l",2,4093],2046],
  [2049,2047,["l",2,2048],1024],
  [2050,2049,["l",2,4099],2049],
  [2053,1027,["l",2,2053],1026],
  [2056,2055,["l",2,4111],2055],
  [2063,1032,["l",2,2063],1031],
  [2064,2063,["l",2,4127],2063],
  [2065,2064,["l",2,4129],2064],
  [2067,2066,["l",2,4133],2066],
  [2069,1035,["l",2,2069],1034],
  [2070,2069,["l",2,4139],2069],
  [2077,2076,["l",2,4153],2076],
  [2079,2078,["l",2,4157],2078],
  [2080,2079,["l",2,4159],2079],
  [2081,1041,["l",2,2081],1040],
  [2083,1042,["l",2,2083],1041],
  [2087,1044,["l",2,2087],1043],
  [2089,1045,["l",2,2089],1044],
  [2089,2088,["l",2,4177],2088],
  [2099,1050,["l",2,2099],1049],
  [2101,2100,["l",2,4201],2100],
  [2106,2105,["l",2,4211],2105],
  [2109,2108,["l",2,4217],2108],
  [2110,2109,["l",2,4219],2109],
  [2111,1056,["l",2,2111],1055],
  [2113,1057,["l",2,2113],1056],
  [2115,2114,["l",2,4229],2114],
  [2116,2115,["l",2,4231],2115],
  [2121,2120,["l",2,4241],2120],
  [2122,2121,["l",2,4243],2121],
  [2127,2126,["l",2,4253],2126],
  [2129,1065,["l",2,2129],1064],
  [2130,2129,["l",2,4259],2129],
  [2131,1066,["l",2,2131],1065],
  [2131,2130,["l",2,4261],2130],
  [2136,2135,["l",2,4271],2135],
  [2137,1069,["l",2,2137],1068],
  [2137,2136,["l",2,4273],2136],
  [2141,1071,["l",2,2141],1070],
  [2142,2141,["l",2,4283],2141],
  [2143,1072,["l",2,2143],1071],
  [2145,2144,["l",2,4289],2144],
  [2149,2148,["l",2,4297],2148],
  [2153,1077,["l",2,2153],1076],
  [2161,1081,["l",2,2161],1080],
  [2162,1105,["s",4,47],2209],[2162,1128,["s",4,47],2209],
  [2164,2163,["l",2,4327],2163],
  [2169,2168,["l",2,4337],2168],
  [2170,2169,["l",2,4339],2169],
  [2175,2174,["l",2,4349],2174],
  [2178,1460,["s",14,3],1093],[2178,1476,["s",14,3],1093],
  [2179,1090,["l",2,2179],1089],
  [2179,2178,["l",2,4357],2178],
  [2182,2181,["l",2,4363],2181],
  [2184,1190,["s",6,13],1098],[2184,1460,["s",14,3],1093],
  [2184,1476,["s",14,3],1093],[2184,2178,["s",14,3],1093],
  [2187,2186,["l",2,4373],2186],
  [2190,1460,["s",14,3],1093],[2190,1476,["s",14,3],1093],
  [2190,2178,["s",14,3],1093],[2190,2184,["s",14,3],1093],
  [2196,1460,["s",14,3],1093],[2196,1476,["s",14,3],1093],
  [2196,2178,["s",14,3],1093],[2196,2184,["s",14,3],1093],
  [2196,2190,["s",14,3],1093],
  [2196,2195,["l",2,4391],2195],
  [2199,2198,["l",2,4397],2198],
  [2203,1102,["l",2,2203],1101],
  [2205,2204,["l",2,4409],2204],
  [2207,1104,["l",2,2207],1103],
  [2210,1190,["s",6,13],1098],[2210,2184,["s",6,13],1098],
  [2211,2210,["l",2,4421],2210],
  [2212,2211,["l",2,4423],2211],
  [2213,1107,["l",2,2213],1106],
  [2221,1111,["l",2,2221],1110],
  [2221,2220,["l",2,4441],2220],
  [2224,2223,["l",2,4447],2223],
  [2226,2225,["l",2,4451],2225],
  [2229,2228,["l",2,4457],2228],
  [2232,2231,["l",2,4463],2231],
  [2237,1119,["l",2,2237],1118],
  [2239,1120,["l",2,2239],1119],
  [2241,2240,["l",2,4481],2240],
  [2242,2241,["l",2,4483],2241],
  [2243,1122,["l",2,2243],1121],
  [2245,2244,["l",2,4489],2244],
  [2245,2244,["s",4,67],4489],
  [2247,2246,["l",2,4493],2246],
  [2251,1126,["l",2,2251],1125],
  [2254,2253,["l",2,4507],2253],
  [2256,1105,["s",4,47],2209],[2256,1128,["s",4,47],2209],
  [2256,2162,["s",4,47],2209],[2257,2208,["l",3,47],2255],
  [2257,2256,["l",2,4513],2256],
  [2259,2258,["l",2,4517],2258],
  [2260,2259,["l",2,4519],2259],
  [2262,2261,["l",2,4523],2261],
  [2267,1134,["l",2,2267],1133],
  [2269,1135,["l",2,2269],1134],
  [2273,1137,["l",2,2273],1136],
  [2274,2273,["l",2,4547],2273],
  [2275,2274,["l",2,4549],2274],
  [2278,2245,["s",4,67],4489],
  [2281,1141,["l",2,2281],1140],
  [2281,2280,["l",2,4561],2280],
  [2284,2283,["l",2,4567],2283],
  [2287,1144,["l",2,2287],1143],
  [2292,2291,["l",2,4583],2291],
  [2293,1147,["l",2,2293],1146],
  [2296,2295,["l",2,4591],2295],
  [2297,1149,["l",2,2297],1148],
  [2299,2298,["l",2,4597],2298],
  [2302,2301,["l",2,4603],2301],
  [2309,1155,["l",2,2309],1154],
  [2311,1156,["l",2,2311],1155],
  [2311,2310,["l",2,4621],2310],
  [2319,2318,["l",2,4637],2318],
  [2320,2319,["l",2,4639],2319],
  [2322,2321,["l",2,4643],2321],
  [2325,2324,["l",2,4649],2324],
  [2326,2325,["l",2,4651],2325],
  [2329,2328,["l",2,4657],2328],
  [2332,2331,["l",2,4663],2331],
  [2333,1167,["l",2,2333],1166],
  [2337,2336,["l",2,4673],2336],
  [2339,1170,["l",2,2339],1169],
  [2340,2339,["l",2,4679],2339],
  [2341,1171,["l",2,2341],1170],
  [2346,2345,["l",2,4691],2345],
  [2347,1174,["l",2,2347],1173],
  [2351,1176,["l",2,2351],1175],
  [2352,2351,["l",2,4703],2351],
  [2357,1179,["l",2,2357],1178],
  [2361,2360,["l",2,4721],2360],
  [2362,2361,["l",2,4723],2361],
  [2365,2364,["l",2,4729],2364],
  [2367,2366,["l",2,4733],2366],
  [2371,1186,["l",2,2371],1185],
  [2376,2375,["l",2,4751],2375],
  [2377,1189,["l",2,2377],1188],
  [2380,2379,["l",2,4759],2379],
  [2381,1191,["l",2,2381],1190],
  [2383,1192,["l",2,2383],1191],
  [2389,1195,["l",2,2389],1194],
  [2392,2391,["l",2,4783],2391],
  [2393,1197,["l",2,2393],1196],[2394,1368,["s",8,7],1200]
    ,[2394,1400,["s",8,7],1200],
  [2394,2393,["l",2,4787],2393],
  [2395,2394,["l",2,4789],2394],
  [2397,2396,["l",2,4793],2396],
  [2399,1200,["l",2,2399],1199],[2400,2101,["u",5,7],2101]
    ,[2400,2353,["u",3,49],2401],
  [2400,2399,["l",2,4799],2399],
  [2401,2400,["l",2,4801],2400],
  [2407,2406,["l",2,4813],2406],[2408,1368,["s",8,7],1200]
    ,[2408,1400,["s",8,7],1200],
  [2408,2394,["s",8,7],1200],[2408,2400,["u",5,7],2101],
  [2409,2408,["l",2,4817],2408],
  [2411,1206,["l",2,2411],1205],
  [2416,2415,["l",2,4831],2415],
  [2417,1209,["l",2,2417],1208],
  [2423,1212,["l",2,2423],1211],
  [2431,2430,["l",2,4861],2430],
  [2436,2435,["l",2,4871],2435],
  [2437,1219,["l",2,2437],1218],
  [2439,2438,["l",2,4877],2438],
  [2441,1221,["l",2,2441],1220],
  [2445,2444,["l",2,4889],2444],
  [2447,1224,["l",2,2447],1223],
  [2452,2451,["l",2,4903],2451],
  [2455,2454,["l",2,4909],2454],
  [2457,2320,["u",4,17],4640],[2457,2448,["u",4,17],4640],
  [2457,2456,["l",2,4913],2456],
  [2459,1230,["l",2,2459],1229],
  [2460,2459,["l",2,4919],2459],
  [2466,2465,["l",2,4931],2465],
  [2467,1234,["l",2,2467],1233],
  [2467,2466,["l",2,4933],2466],
  [2469,2468,["l",2,4937],2468],
  [2472,2471,["l",2,4943],2471],
  [2473,1237,["l",2,2473],1236],
  [2476,2475,["l",2,4951],2475],
  [2477,1239,["l",2,2477],1238],
  [2479,2478,["l",2,4957],2478],
  [2484,2483,["l",2,4967],2483],
  [2485,2484,["l",2,4969],2484],
  [2487,2486,["l",2,4973],2486],
  [2494,2493,["l",2,4987],2493],
  [2497,2496,["l",2,4993],2496],
  [2500,2499,["l",2,4999],2499],
  [2502,2501,["l",2,5003],2501],
  [2503,1252,["l",2,2503],1251],
  [2505,2504,["l",2,5009],2504],
  [2506,2505,["l",2,5011],2505],
  [2511,2510,["l",2,5021],2510],
  [2512,2511,["l",2,5023],2511],
  [2520,2519,["l",2,5039],2519],
  [2521,1261,["l",2,2521],1260],
  [2521,2520,["l",2,5041],2520],
  [2526,2525,["l",2,5051],2525],
  [2530,2529,["l",2,5059],2529],
  [2531,1266,["l",2,2531],1265],
  [2539,1270,["l",2,2539],1269],
  [2539,2538,["l",2,5077],2538],
  [2541,2540,["l",2,5081],2540],
  [2543,1272,["l",2,2543],1271],
  [2544,2543,["l",2,5087],2543],
  [2549,1275,["l",2,2549],1274],
  [2550,2549,["l",2,5099],2549],
  [2551,1276,["l",2,2551],1275],
  [2551,2550,["l",2,5101],2550],
  [2554,2553,["l",2,5107],2553],
  [2557,1279,["l",2,2557],1278],
  [2557,2556,["l",2,5113],2556],
  [2560,2559,["l",2,5119],2559],
  [2574,2573,["l",2,5147],2573],
  [2577,2576,["l",2,5153],2576],
  [2579,1290,["l",2,2579],1289],
  [2584,2583,["l",2,5167],2583],
  [2586,2585,["l",2,5171],2585],
  [2590,2589,["l",2,5179],2589],
  [2591,1296,["l",2,2591],1295],
  [2593,1297,["l",2,2593],1296],
  [2595,2594,["l",2,5189],2594],
  [2599,2598,["l",2,5197],2598],
  [2605,2604,["l",2,5209],2604],
  [2609,1305,["l",2,2609],1304],
  [2610,2457,["s",6,17],2456],[2610,2465,["s",6,17],2456],
  [2614,2613,["l",2,5227],2613],
  [2616,2615,["l",2,5231],2615],
  [2617,1309,["l",2,2617],1308],
  [2617,2616,["l",2,5233],2616],
  [2619,2618,["l",2,5237],2618],
  [2621,1311,["l",2,2621],1310],
  [2631,2630,["l",2,5261],2630],
  [2633,1317,["l",2,2633],1316],
  [2637,2636,["l",2,5273],2636],
  [2640,2639,["l",2,5279],2639],
  [2641,2640,["l",2,5281],2640],
  [2647,1324,["l",2,2647],1323],
  [2649,2648,["l",2,5297],2648],
  [2652,2651,["l",2,5303],2651],
  [2655,2654,["l",2,5309],2654],
  [2657,1329,["l",2,2657],1328],
  [2659,1330,["l",2,2659],1329],
  [2662,2661,["l",2,5323],2661],
  [2663,1332,["l",2,2663],1331],
  [2665,2664,["l",2,5329],2664],
  [2667,2666,["l",2,5333],2666],
  [2671,1336,["l",2,2671],1335],
  [2674,2673,["l",2,5347],2673],
  [2676,2675,["l",2,5351],2675],
  [2677,1339,["l",2,2677],1338],
  [2683,1342,["l",2,2683],1341],
  [2687,1344,["l",2,2687],1343],
  [2689,1345,["l",2,2689],1344],
  [2691,2690,["l",2,5381],2690],
  [2693,1347,["l",2,2693],1346],
  [2694,2693,["l",2,5387],2693],
  [2697,2696,["l",2,5393],2696],
  [2699,1350,["l",2,2699],1349],
  [2700,2699,["l",2,5399],2699],
  [2704,2703,["l",2,5407],2703],
  [2707,1354,["l",2,2707],1353],
  [2707,2706,["l",2,5413],2706],
  [2709,2708,["l",2,5417],2708],
  [2710,2709,["l",2,5419],2709],
  [2711,1356,["l",2,2711],1355],
  [2713,1357,["l",2,2713],1356],
  [2716,2715,["l",2,5431],2715],
  [2719,1360,["l",2,2719],1359],
  [2719,2718,["l",2,5437],2718],
  [2721,2720,["l",2,5441],2720],
  [2722,2721,["l",2,5443],2721],
  [2725,2724,["l",2,5449],2724],
  [2729,1365,["l",2,2729],1364],
  [2731,1366,["l",2,2731],1365],
  [2736,2735,["l",2,5471],2735],
  [2739,2738,["l",2,5477],2738],
  [2740,2739,["l",2,5479],2739],
  [2741,1371,["l",2,2741],1370],
  [2742,2741,["l",2,5483],2741],
  [2749,1375,["l",2,2749],1374],
  [2751,2750,["l",2,5501],2750],
  [2752,2751,["l",2,5503],2751],
  [2753,1377,["l",2,2753],1376],
  [2754,2753,["l",2,5507],2753],
  [2756,1405,["s",4,53],2809],[2756,1431,["s",4,53],2809],
  [2760,2759,["l",2,5519],2759],
  [2761,2760,["l",2,5521],2760],
  [2764,2763,["l",2,5527],2763],
  [2766,2765,["l",2,5531],2765],
  [2767,1384,["l",2,2767],1383],
  [2777,1389,["l",2,2777],1388],
  [2779,2778,["l",2,5557],2778],
  [2782,2781,["l",2,5563],2781],
  [2785,2784,["l",2,5569],2784],
  [2787,2786,["l",2,5573],2786],
  [2789,1395,["l",2,2789],1394],
  [2791,1396,["l",2,2791],1395],
  [2791,2790,["l",2,5581],2790],
  [2796,2795,["l",2,5591],2795],
  [2797,1399,["l",2,2797],1398],
  [2801,1401,["l",2,2801],1400],[2801,2736,["l",5,7],2799]
    ,[2803,1402,["l",2,2803],1401],
  [2812,2811,["l",2,5623],2811],
  [2819,1410,["l",2,2819],1409],
  [2820,2819,["l",2,5639],2819],
  [2821,2820,["l",2,5641],2820],
  [2824,2823,["l",2,5647],2823],
  [2826,2825,["l",2,5651],2825],
  [2827,2826,["l",2,5653],2826],
  [2829,2828,["l",2,5657],2828],
  [2830,2829,["l",2,5659],2829],
  [2833,1417,["l",2,2833],1416],
  [2835,2834,["l",2,5669],2834],
  [2837,1419,["l",2,2837],1418],
  [2842,2841,["l",2,5683],2841],
  [2843,1422,["l",2,2843],1421],
  [2845,2844,["l",2,5689],2844],
  [2847,2846,["l",2,5693],2846],
  [2851,1426,["l",2,2851],1425],
  [2851,2850,["l",2,5701],2850],
  [2856,2855,["l",2,5711],2855],
  [2857,1429,["l",2,2857],1428],
  [2859,2858,["l",2,5717],2858],
  [2861,1431,["l",2,2861],1430],
  [2862,1405,["s",4,53],2809],[2862,1431,["s",4,53],2809],
  [2862,2756,["s",4,53],2809],[2863,2808,["l",3,53],2861],
  [2869,2868,["l",2,5737],2868],
  [2871,2870,["l",2,5741],2870],
  [2872,2871,["l",2,5743],2871],
  [2875,2874,["l",2,5749],2874],
  [2879,1440,["l",2,2879],1439],
  [2887,1444,["l",2,2887],1443],
  [2890,2889,["l",2,5779],2889],
  [2892,2891,["l",2,5783],2891],
  [2896,2895,["l",2,5791],2895],
  [2897,1449,["l",2,2897],1448],
  [2901,2900,["l",2,5801],2900],
  [2903,1452,["l",2,2903],1451],
  [2904,2903,["l",2,5807],2903],
  [2907,2906,["l",2,5813],2906],
  [2909,1455,["l",2,2909],1454],
  [2911,2910,["l",2,5821],2910],
  [2914,2913,["l",2,5827],2913],
  [2917,1459,["l",2,2917],1458],
  [2920,2919,["l",2,5839],2919],
  [2922,2921,["l",2,5843],2921],
  [2925,2924,["l",2,5849],2924],
  [2926,2925,["l",2,5851],2925],
  [2927,1464,["l",2,2927],1463],
  [2929,2928,["l",2,5857],2928],
  [2931,2930,["l",2,5861],2930],
  [2934,2933,["l",2,5867],2933],
  [2935,2934,["l",2,5869],2934],
  [2939,1470,["l",2,2939],1469],
  [2940,2939,["l",2,5879],2939],
  [2941,2940,["l",2,5881],2940],
  [2949,2948,["l",2,5897],2948],
  [2952,2951,["l",2,5903],2951],
  [2953,1477,["l",2,2953],1476],
  [2957,1479,["l",2,2957],1478],
  [2962,2961,["l",2,5923],2961],
  [2963,1482,["l",2,2963],1481],
  [2964,2963,["l",2,5927],2963],
  [2969,1485,["l",2,2969],1484],
  [2970,2969,["l",2,5939],2969],
  [2971,1486,["l",2,2971],1485],
  [2977,2976,["l",2,5953],2976],
  [2991,2990,["l",2,5981],2990],
  [2994,2993,["l",2,5987],2993],
  [2999,1500,["l",2,2999],1499],
  [3001,1501,["l",2,3001],1500],
  [3004,3003,["l",2,6007],3003],
  [3006,3005,["l",2,6011],3005],
  [3011,1506,["l",2,3011],1505],
  [3015,3014,["l",2,6029],3014],
  [3019,1510,["l",2,3019],1509],
  [3019,3018,["l",2,6037],3018],
  [3022,3021,["l",2,6043],3021],
  [3023,1512,["l",2,3023],1511],
  [3024,3023,["l",2,6047],3023],
  [3027,3026,["l",2,6053],3026],
  [3034,3033,["l",2,6067],3033],
  [3037,1519,["l",2,3037],1518],
  [3037,3036,["l",2,6073],3036],
  [3040,3039,["l",2,6079],3039],
  [3041,1521,["l",2,3041],1520],
  [3045,3044,["l",2,6089],3044],
  [3046,3045,["l",2,6091],3045],
  [3049,1525,["l",2,3049],1524],
  [3051,3050,["l",2,6101],3050],
  [3057,3056,["l",2,6113],3056],
  [3061,1531,["l",2,3061],1530],
  [3061,3060,["l",2,6121],3060],
  [3066,3065,["l",2,6131],3065],
  [3067,1534,["l",2,3067],1533],
  [3067,3066,["l",2,6133],3066],
  [3072,3071,["l",2,6143],3071],
  [3076,3075,["l",2,6151],3075],
  [3079,1540,["l",2,3079],1539],
  [3082,3081,["l",2,6163],3081],
  [3083,1542,["l",2,3083],1541],
  [3087,3086,["l",2,6173],3086],
  [3089,1545,["l",2,3089],1544],
  [3099,3098,["l",2,6197],3098],
  [3100,3099,["l",2,6199],3099],
  [3102,3101,["l",2,6203],3101],
  [3106,3105,["l",2,6211],3105],
  [3109,1555,["l",2,3109],1554],
  [3109,3108,["l",2,6217],3108],
  [3111,3110,["l",2,6221],3110],
  [3115,3114,["l",2,6229],3114],
  [3119,1560,["l",2,3119],1559],
  [3120,1878,["s",10,5],1562],
  [3121,1561,["l",2,3121],1560],
  [3121,3120,["l",2,6241],3120],
  [3124,3123,["l",2,6247],3123],
  [3129,3128,["l",2,6257],3128],
  [3130,1878,["s",10,5],1562],[3130,3120,["s",10,5],1562],
  [3132,3131,["l",2,6263],3131],
  [3135,3134,["l",2,6269],3134],
  [3136,3135,["l",2,6271],3135],
  [3137,1569,["l",2,3137],1568],
  [3139,3138,["l",2,6277],3138],
  [3144,3143,["l",2,6287],3143],
  [3150,3149,["l",2,6299],3149],
  [3151,3150,["l",2,6301],3150],
  [3156,3155,["l",2,6311],3155],
  [3159,3158,["l",2,6317],3158],
  [3162,3161,["l",2,6323],3161],
  [3163,1582,["l",2,3163],1581],
  [3165,3164,["l",2,6329],3164],
  [3167,1584,["l",2,3167],1583],
  [3169,1585,["l",2,3169],1584],
  [3169,3168,["l",2,6337],3168],
  [3172,3171,["l",2,6343],3171],
  [3177,3176,["l",2,6353],3176],
  [3180,3179,["l",2,6359],3179],
  [3181,1591,["l",2,3181],1590],
  [3181,3180,["l",2,6361],3180],
  [3184,3183,["l",2,6367],3183],
  [3187,1594,["l",2,3187],1593],
  [3187,3186,["l",2,6373],3186],
  [3190,3189,["l",2,6379],3189],
  [3191,1596,["l",2,3191],1595],
  [3195,3194,["l",2,6389],3194],
  [3199,3198,["l",2,6397],3198],
  [3203,1602,["l",2,3203],1601],
  [3209,1605,["l",2,3209],1604],
  [3211,3210,["l",2,6421],3210],
  [3214,3213,["l",2,6427],3213],
  [3217,1609,["l",2,3217],1608],
  [3221,1611,["l",2,3221],1610],
  [3225,3224,["l",2,6449],3224],
  [3226,3225,["l",2,6451],3225],
  [3229,1615,["l",2,3229],1614],
  [3235,3234,["l",2,6469],3234],
  [3237,3236,["l",2,6473],3236],
  [3241,3240,["l",2,6481],3240],
  [3246,3245,["l",2,6491],3245],
  [3251,1626,["l",2,3251],1625],
  [3253,1627,["l",2,3253],1626],
  [3257,1629,["l",2,3257],1628],
  [3259,1630,["l",2,3259],1629],
  [3261,3260,["l",2,6521],3260],
  [3265,3264,["l",2,6529],3264],
  [3271,1636,["l",2,3271],1635],
  [3274,3273,["l",2,6547],3273],
  [3276,3275,["l",2,6551],3275],
  [3277,3276,["l",2,6553],3276],
  [3281,3280,["l",2,6561],3280],
  [3282,3281,["l",2,6563],3281],
  [3285,3284,["l",2,6569],3284],
  [3286,3285,["l",2,6571],3285],
  [3289,3288,["l",2,6577],3288],
  [3291,3290,["l",2,6581],3290],
  [3299,1650,["l",2,3299],1649],
  [3300,3299,["l",2,6599],3299],
  [3301,1651,["l",2,3301],1650],
  [3304,3303,["l",2,6607],3303],
  [3307,1654,["l",2,3307],1653],
  [3310,3309,["l",2,6619],3309],
  [3313,1657,["l",2,3313],1656],
  [3319,1660,["l",2,3319],1659],
  [3319,3318,["l",2,6637],3318],
  [3323,1662,["l",2,3323],1661],
  [3327,3326,["l",2,6653],3326],
  [3329,1665,["l",2,3329],1664],
  [3330,3329,["l",2,6659],3329],
  [3331,1666,["l",2,3331],1665],
  [3331,3330,["l",2,6661],3330],
  [3337,3336,["l",2,6673],3336],
  [3340,3339,["l",2,6679],3339],
  [3343,1672,["l",2,3343],1671],
  [3345,3344,["l",2,6689],3344],
  [3346,3345,["l",2,6691],3345],
  [3347,1674,["l",2,3347],1673],
  [3351,3350,["l",2,6701],3350],
  [3352,3351,["l",2,6703],3351],
  [3355,3354,["l",2,6709],3354],
  [3359,1680,["l",2,3359],1679],
  [3360,3359,["l",2,6719],3359],
  [3361,1681,["l",2,3361],1680],
  [3367,3366,["l",2,6733],3366],
  [3369,3368,["l",2,6737],3368],
  [3371,1686,["l",2,3371],1685],
  [3373,1687,["l",2,3373],1686],
  [3381,3380,["l",2,6761],3380],
  [3382,3381,["l",2,6763],3381],
  [3389,1695,["l",2,3389],1694],
  [3390,3389,["l",2,6779],3389],
  [3391,1696,["l",2,3391],1695],
  [3391,3390,["l",2,6781],3390],
  [3396,3395,["l",2,6791],3395],
  [3397,3396,["l",2,6793],3396],
  [3402,3401,["l",2,6803],3401],
  [3407,1704,["l",2,3407],1703],
  [3412,3411,["l",2,6823],3411],
  [3413,1707,["l",2,3413],1706],
  [3414,3413,["l",2,6827],3413],
  [3415,3414,["l",2,6829],3414],
  [3417,3416,["l",2,6833],3416],
  [3421,3420,["l",2,6841],3420],
  [3422,1741,["s",4,59],3481],[3422,1770,["s",4,59],3481],
  [3429,3428,["l",2,6857],3428],
  [3430,3429,["l",2,6859],3429],
  [3432,3431,["l",2,6863],3431],
  [3433,1717,["l",2,3433],1716],
  [3435,3434,["l",2,6869],3434],
  [3436,3435,["l",2,6871],3435],
  [3442,3441,["l",2,6883],3441],
  [3445,3444,["l",2,6889],3444],
  [3449,1725,["l",2,3449],1724],
  [3450,3449,["l",2,6899],3449],
  [3454,3453,["l",2,6907],3453],
  [3456,3455,["l",2,6911],3455],
  [3457,1729,["l",2,3457],1728],
  [3459,3458,["l",2,6917],3458],
  [3461,1731,["l",2,3461],1730],
  [3463,1732,["l",2,3463],1731],
  [3467,1734,["l",2,3467],1733],
  [3469,1735,["l",2,3469],1734],
  [3474,3473,["l",2,6947],3473],
  [3475,3474,["l",2,6949],3474],
  [3480,3479,["l",2,6959],3479],
  [3481,3480,["l",2,6961],3480],
  [3484,3483,["l",2,6967],3483],
  [3486,3485,["l",2,6971],3485],
  [3489,3488,["l",2,6977],3488],
  [3491,1746,["l",2,3491],1745],
  [3492,3491,["l",2,6983],3491],
  [3496,3495,["l",2,6991],3495],
  [3499,1750,["l",2,3499],1749],
  [3499,3498,["l",2,6997],3498],
  [3501,3500,["l",2,7001],3500],
  [3507,3506,["l",2,7013],3506],
  [3510,3509,["l",2,7019],3509],
  [3511,1756,["l",2,3511],1755],
  [3514,3513,["l",2,7027],3513],
  [3517,1759,["l",2,3517],1758],
  [3520,3519,["l",2,7039],3519],
  [3522,3521,["l",2,7043],3521],
  [3527,1764,["l",2,3527],1763],
  [3529,1765,["l",2,3529],1764],
  [3529,3528,["l",2,7057],3528],
  [3533,1767,["l",2,3533],1766],
  [3535,3534,["l",2,7069],3534],
  [3539,1770,["l",2,3539],1769],
  [3540,1741,["s",4,59],3481],[3540,1770,["s",4,59],3481],
  [3540,3422,["s",4,59],3481],
  [3540,3539,["l",2,7079],3539],
  [3541,1771,["l",2,3541],1770],
  [3541,3480,["l",3,59],3539],
  [3547,1774,["l",2,3547],1773],
  [3552,3551,["l",2,7103],3551],
  [3555,3554,["l",2,7109],3554],
  [3557,1779,["l",2,3557],1778],
  [3559,1780,["l",2,3559],1779],
  [3561,3560,["l",2,7121],3560],
  [3564,3563,["l",2,7127],3563],
  [3565,3564,["l",2,7129],3564],
  [3571,1786,["l",2,3571],1785],
  [3576,3575,["l",2,7151],3575],
  [3580,3579,["l",2,7159],3579],
  [3581,1791,["l",2,3581],1790],
  [3583,1792,["l",2,3583],1791],
  [3589,3588,["l",2,7177],3588],
  [3593,1797,["l",2,3593],1796],
  [3594,3593,["l",2,7187],3593],
  [3597,3596,["l",2,7193],3596],
  [3604,3603,["l",2,7207],3603],
  [3606,3605,["l",2,7211],3605],
  [3607,1804,["l",2,3607],1803],
  [3607,3606,["l",2,7213],3606],
  [3610,3609,["l",2,7219],3609],
  [3613,1807,["l",2,3613],1806],
  [3615,3614,["l",2,7229],3614],
  [3617,1809,["l",2,3617],1808],
  [3619,3618,["l",2,7237],3618],
  [3620,3430,["s",6,19],3429],[3620,3439,["s",6,19],3429],
  [3622,3621,["l",2,7243],3621],
  [3623,1812,["l",2,3623],1811],
  [3624,3623,["l",2,7247],3623],
  [3627,3626,["l",2,7253],3626],
  [3631,1816,["l",2,3631],1815],
  [3637,1819,["l",2,3637],1818],[3640,3281,["s",8,9],3280]
    ,[3642,3641,["l",2,7283],3641],
  [3643,1822,["l",2,3643],1821],
  [3649,3648,["l",2,7297],3648],
  [3654,3653,["l",2,7307],3653],
  [3655,3654,["l",2,7309],3654],
  [3659,1830,["l",2,3659],1829],
  [3660,1861,["s",4,61],3721],[3660,1891,["s",4,61],3721],
  [3661,3660,["l",2,7321],3660],
  [3666,3665,["l",2,7331],3665],
  [3667,3666,["l",2,7333],3666],
  [3671,1836,["l",2,3671],1835],
  [3673,1837,["l",2,3673],1836],
  [3675,3674,["l",2,7349],3674],
  [3676,3675,["l",2,7351],3675],
  [3677,1839,["l",2,3677],1838],
  [3685,3684,["l",2,7369],3684],
  [3691,1846,["l",2,3691],1845],
  [3697,1849,["l",2,3697],1848],
  [3697,3696,["l",2,7393],3696],
  [3701,1851,["l",2,3701],1850],
  [3706,3705,["l",2,7411],3705],
  [3709,1855,["l",2,3709],1854],
  [3709,3708,["l",2,7417],3708],
  [3717,3716,["l",2,7433],3716],
  [3719,1860,["l",2,3719],1859],
  [3720,3661,["u",3,61],3721],
  [3726,3725,["l",2,7451],3725],
  [3727,1864,["l",2,3727],1863],
  [3729,3728,["l",2,7457],3728],
  [3730,3729,["l",2,7459],3729],
  [3733,1867,["l",2,3733],1866],
  [3739,1870,["l",2,3739],1869],
  [3739,3738,["l",2,7477],3738],
  [3741,3740,["l",2,7481],3740],
  [3744,3743,["l",2,7487],3743],
  [3745,3744,["l",2,7489],3744],
  [3750,3749,["l",2,7499],3749],
  [3754,3753,["l",2,7507],3753],
  [3759,3758,["l",2,7517],3758],
  [3761,1881,["l",2,3761],1880],
  [3762,3761,["l",2,7523],3761],
  [3765,3764,["l",2,7529],3764],
  [3767,1884,["l",2,3767],1883],
  [3769,1885,["l",2,3769],1884],
  [3769,3768,["l",2,7537],3768],
  [3771,3770,["l",2,7541],3770],
  [3774,3773,["l",2,7547],3773],
  [3775,3774,["l",2,7549],3774],
  [3779,1890,["l",2,3779],1889],
  [3780,3779,["l",2,7559],3779],
  [3781,3780,["l",2,7561],3780],
  [3782,1861,["s",4,61],3721],[3782,1891,["s",4,61],3721],
  [3782,3660,["s",4,61],3721],[3782,3720,["u",3,61],3721],
  [3787,3786,["l",2,7573],3786],
  [3789,3788,["l",2,7577],3788],
  [3792,3791,["l",2,7583],3791],
  [3793,1897,["l",2,3793],1896],
  [3795,3794,["l",2,7589],3794],
  [3796,3795,["l",2,7591],3795],
  [3797,1899,["l",2,3797],1898],
  [3802,3801,["l",2,7603],3801],
  [3803,1902,["l",2,3803],1901],
  [3804,3803,["l",2,7607],3803],
  [3811,3810,["l",2,7621],3810],
  [3820,3819,["l",2,7639],3819],
  [3821,1911,["l",2,3821],1910],
  [3822,3821,["l",2,7643],3821],
  [3823,1912,["l",2,3823],1911],
  [3825,3824,["l",2,7649],3824],
  [3833,1917,["l",2,3833],1916],
  [3835,3834,["l",2,7669],3834],
  [3837,3836,["l",2,7673],3836],
  [3841,3840,["l",2,7681],3840],
  [3844,3843,["l",2,7687],3843],
  [3846,3845,["l",2,7691],3845],
  [3847,1924,["l",2,3847],1923],
  [3850,3849,["l",2,7699],3849],
  [3851,1926,["l",2,3851],1925],
  [3852,3851,["l",2,7703],3851],
  [3853,1927,["l",2,3853],1926],
  [3859,3858,["l",2,7717],3858],
  [3862,3861,["l",2,7723],3861],
  [3863,1932,["l",2,3863],1931],
  [3864,3863,["l",2,7727],3863],
  [3871,3870,["l",2,7741],3870],
  [3877,1939,["l",2,3877],1938],
  [3877,3876,["l",2,7753],3876],
  [3879,3878,["l",2,7757],3878],
  [3880,3879,["l",2,7759],3879],
  [3881,1941,["l",2,3881],1940],
  [3889,1945,["l",2,3889],1944],
  [3895,3894,["l",2,7789],3894],
  [3897,3896,["l",2,7793],3896],
  [3907,1954,["l",2,3907],1953],
  [3909,3908,["l",2,7817],3908],
  [3911,1956,["l",2,3911],1955],
  [3912,3911,["l",2,7823],3911],
  [3915,3914,["l",2,7829],3914],
  [3917,1959,["l",2,3917],1958],
  [3919,1960,["l",2,3919],1959],
  [3921,3920,["l",2,7841],3920],
  [3923,1962,["l",2,3923],1961],
  [3927,3926,["l",2,7853],3926],
  [3929,1965,["l",2,3929],1964],
  [3931,1966,["l",2,3931],1965],
  [3934,3933,["l",2,7867],3933],
  [3937,3936,["l",2,7873],3936],
  [3939,3938,["l",2,7877],3938],
  [3940,3939,["l",2,7879],3939],
  [3942,3941,["l",2,7883],3941],
  [3943,1972,["l",2,3943],1971],
  [3947,1974,["l",2,3947],1973],
  [3951,3950,["l",2,7901],3950],
  [3954,3953,["l",2,7907],3953],
  [3960,3959,["l",2,7919],3959],
  [3961,3960,["l",2,7921],3960],
  [3964,3963,["l",2,7927],3963],
  [3967,1984,["l",2,3967],1983],
  [3967,3966,["l",2,7933],3966],
  [3969,3968,["l",2,7937],3968],
  [3975,3974,["l",2,7949],3974],
  [3976,3975,["l",2,7951],3975],
  [3982,3981,["l",2,7963],3981],
  [3989,1995,["l",2,3989],1994],
  [3997,3996,["l",2,7993],3996],
  [4001,2001,["l",2,4001],2000],
  [4003,2002,["l",2,4003],2001],
  [4005,4004,["l",2,8009],4004],
  [4006,4005,["l",2,8011],4005],
  [4007,2004,["l",2,4007],2003],
  [4009,4008,["l",2,8017],4008],
  [4013,2007,["l",2,4013],2006],
  [4019,2010,["l",2,4019],2009],
  [4020,4019,["l",2,8039],4019],
  [4021,2011,["l",2,4021],2010],
  [4027,2014,["l",2,4027],2013],
  [4027,4026,["l",2,8053],4026],
  [4030,4029,["l",2,8059],4029],
  [4035,4034,["l",2,8069],4034],
  [4041,4040,["l",2,8081],4040],
  [4044,4043,["l",2,8087],4043],
  [4045,4044,["l",2,8089],4044],
  [4047,4046,["l",2,8093],4046],
  [4049,2025,["l",2,4049],2024],
  [4051,2026,["l",2,4051],2025],
  [4051,4050,["l",2,8101],4050],
  [4056,4055,["l",2,8111],4055],
  [4057,2029,["l",2,4057],2028],
  [4059,4058,["l",2,8117],4058],
  [4062,4061,["l",2,8123],4061],
  [4073,2037,["l",2,4073],2036],
  [4074,4073,["l",2,8147],4073],
  [4079,2040,["l",2,4079],2039],
  [4081,4080,["l",2,8161],4080],
  [4084,4083,["l",2,8167],4083],
  [4086,4085,["l",2,8171],4085],
  [4090,4089,["l",2,8179],4089],
  [4091,2046,["l",2,4091],2045],
  [4093,2047,["l",2,4093],2046],[4095,3315,["u",7,4],3277]
    ,[4095,3641,["u",5,8],3641],
  [4095,3937,["l",12,2],4096],[4095,4033,["u",3,64],4096],
  [4096,4095,["l",2,8191],4095],
  [4097,3855,["u",4,16],3855],
  [4097,4095,["l",2,4096],2048],
  [4099,2050,["l",2,4099],2049],
  [4105,4104,["l",2,8209],4104],
  [4110,4109,["l",2,8219],4109],
  [4111,2056,["l",2,4111],2055],
  [4111,4110,["l",2,8221],4110],
  [4116,4115,["l",2,8231],4115],
  [4117,4116,["l",2,8233],4116],
  [4119,4118,["l",2,8237],4118],
  [4122,4121,["l",2,8243],4121],
  [4127,2064,["l",2,4127],2063],
  [4129,2065,["l",2,4129],2064],
  [4132,4131,["l",2,8263],4131],
  [4133,2067,["l",2,4133],2066],
  [4135,4134,["l",2,8269],4134],
  [4137,4136,["l",2,8273],4136],
  [4139,2070,["l",2,4139],2069],
  [4144,4143,["l",2,8287],4143],
  [4146,4145,["l",2,8291],4145],
  [4147,4146,["l",2,8293],4146],
  [4149,4148,["l",2,8297],4148],
  [4153,2077,["l",2,4153],2076],
  [4156,4155,["l",2,8311],4155],
  [4157,2079,["l",2,4157],2078],
  [4159,2080,["l",2,4159],2079],
  [4159,4158,["l",2,8317],4158],
  [4165,4164,["l",2,8329],4164],
  [4177,2089,["l",2,4177],2088],
  [4177,4176,["l",2,8353],4176],
  [4182,4181,["l",2,8363],4181],
  [4185,4184,["l",2,8369],4184],
  [4189,4188,["l",2,8377],4188],
  [4194,4193,["l",2,8387],4193],
  [4195,4194,["l",2,8389],4194],
  [4201,2101,["l",2,4201],2100],
  [4210,4209,["l",2,8419],4209],
  [4211,2106,["l",2,4211],2105],
  [4212,4211,["l",2,8423],4211],
  [4215,4214,["l",2,8429],4214],
  [4216,4215,["l",2,8431],4215],
  [4217,2109,["l",2,4217],2108],
  [4219,2110,["l",2,4219],2109],
  [4222,4221,["l",2,8443],4221],
  [4224,4223,["l",2,8447],4223],
  [4229,2115,["l",2,4229],2114],
  [4231,2116,["l",2,4231],2115],
  [4231,4230,["l",2,8461],4230],
  [4234,4233,["l",2,8467],4233],
  [4241,2121,["l",2,4241],2120],
  [4243,2122,["l",2,4243],2121],
  [4251,4250,["l",2,8501],4250],
  [4253,2127,["l",2,4253],2126],
  [4257,4256,["l",2,8513],4256],
  [4259,2130,["l",2,4259],2129],
  [4261,2131,["l",2,4261],2130],
  [4261,4260,["l",2,8521],4260],
  [4264,4263,["l",2,8527],4263],
  [4269,4268,["l",2,8537],4268],
  [4270,4269,["l",2,8539],4269],
  [4271,2136,["l",2,4271],2135],
  [4272,4271,["l",2,8543],4271],
  [4273,2137,["l",2,4273],2136],
  [4282,4281,["l",2,8563],4281],
  [4283,2142,["l",2,4283],2141],
  [4287,4286,["l",2,8573],4286],
  [4289,2145,["l",2,4289],2144],
  [4291,4290,["l",2,8581],4290],
  [4297,2149,["l",2,4297],2148],
  [4299,4298,["l",2,8597],4298],
  [4300,4299,["l",2,8599],4299],
  [4305,4304,["l",2,8609],4304],
  [4312,4311,["l",2,8623],4311],
  [4314,4313,["l",2,8627],4313],
  [4315,4314,["l",2,8629],4314],
  [4321,4320,["l",2,8641],4320],
  [4324,4323,["l",2,8647],4323],
  [4327,2164,["l",2,4327],2163],
  [4332,4331,["l",2,8663],4331],
  [4335,4334,["l",2,8669],4334],
  [4337,2169,["l",2,4337],2168],
  [4339,2170,["l",2,4339],2169],
  [4339,4338,["l",2,8677],4338],
  [4341,4340,["l",2,8681],4340],
  [4345,4344,["l",2,8689],4344],
  [4347,4346,["l",2,8693],4346],
  [4349,2175,["l",2,4349],2174],
  [4350,4349,["l",2,8699],4349],
  [4354,4353,["l",2,8707],4353],
  [4357,2179,["l",2,4357],2178],
  [4357,4356,["l",2,8713],4356],
  [4360,4359,["l",2,8719],4359],
  [4363,2182,["l",2,4363],2181],
  [4366,4365,["l",2,8731],4365],
  [4369,4095,["l",4,16],4367],
  [4369,4368,["l",2,8737],4368],
  [4371,4370,["l",2,8741],4370],
  [4372,3640,["s",16,3],3280],[4372,3660,["s",16,3],3280],
  [4372,3690,["s",16,3],3280],[4372,4356,["s",16,3],3280],
  [4373,2187,["l",2,4373],2186],
  [4374,4373,["l",2,8747],4373],
  [4377,4376,["l",2,8753],4376],
  [4380,4372,["s",16,3],3280],
  [4381,4380,["l",2,8761],4380],
  [4390,4389,["l",2,8779],4389],
  [4391,2196,["l",2,4391],2195],
  [4392,4391,["l",2,8783],4391],
  [4397,2199,["l",2,4397],2198],
  [4402,4401,["l",2,8803],4401],
  [4404,4403,["l",2,8807],4403],
  [4409,2205,["l",2,4409],2204],
  [4410,4409,["l",2,8819],4409],
  [4411,4410,["l",2,8821],4410],
  [4416,4415,["l",2,8831],4415],
  [4419,4418,["l",2,8837],4418],
  [4420,4419,["l",2,8839],4419],
  [4421,2211,["l",2,4421],2210],
  [4422,2245,["s",4,67],4489],[4422,2278,["s",4,67],4489],
  [4423,2212,["l",2,4423],2211],
  [4425,4424,["l",2,8849],4424],
  [4431,4430,["l",2,8861],4430],
  [4432,4431,["l",2,8863],4431],
  [4434,4433,["l",2,8867],4433],
  [4441,2221,["l",2,4441],2220],
  [4444,4443,["l",2,8887],4443],
  [4447,2224,["l",2,4447],2223],
  [4447,4446,["l",2,8893],4446],
  [4451,2226,["l",2,4451],2225],
  [4457,2229,["l",2,4457],2228],
  [4462,4461,["l",2,8923],4461],
  [4463,2232,["l",2,4463],2231],
  [4465,4464,["l",2,8929],4464],
  [4467,4466,["l",2,8933],4466],
  [4471,4470,["l",2,8941],4470],
  [4476,4475,["l",2,8951],4475],
  [4481,2241,["l",2,4481],2240],
  [4482,4481,["l",2,8963],4481],
  [4483,2242,["l",2,4483],2241],
  [4485,4484,["l",2,8969],4484],
  [4486,4485,["l",2,8971],4485],
  [4488,4423,["u",3,67],4489],
  [4493,2247,["l",2,4493],2246],
  [4500,4499,["l",2,8999],4499],
  [4501,4500,["l",2,9001],4500],
  [4504,4503,["l",2,9007],4503],
  [4506,4505,["l",2,9011],4505],
  [4507,2254,["l",2,4507],2253],
  [4507,4506,["l",2,9013],4506],
  [4513,2257,["l",2,4513],2256],
  [4515,4514,["l",2,9029],4514],
  [4517,2259,["l",2,4517],2258],
  [4519,2260,["l",2,4519],2259],
  [4521,4520,["l",2,9041],4520],
  [4522,4521,["l",2,9043],4521],
  [4523,2262,["l",2,4523],2261],
  [4525,4524,["l",2,9049],4524],
  [4530,4529,["l",2,9059],4529],
  [4534,4533,["l",2,9067],4533],
  [4546,4545,["l",2,9091],4545],
  [4547,2274,["l",2,4547],2273],
  [4549,2275,["l",2,4549],2274],
  [4552,4551,["l",2,9103],4551],
  [4555,4554,["l",2,9109],4554],
  [4556,2245,["s",4,67],4489],[4556,2278,["s",4,67],4489],
  [4556,4422,["s",4,67],4489],[4556,4488,["u",3,67],4489],
  [4561,2281,["l",2,4561],2280],
  [4564,4563,["l",2,9127],4563],
  [4567,2284,["l",2,4567],2283],
  [4567,4566,["l",2,9133],4566],
  [4569,4568,["l",2,9137],4568],
  [4576,4575,["l",2,9151],4575],
  [4579,4578,["l",2,9157],4578],
  [4581,4580,["l",2,9161],4580],
  [4583,2292,["l",2,4583],2291],
  [4587,4586,["l",2,9173],4586],
  [4591,2296,["l",2,4591],2295],
  [4591,4590,["l",2,9181],4590],
  [4594,4593,["l",2,9187],4593],
  [4597,2299,["l",2,4597],2298],
  [4600,4599,["l",2,9199],4599],
  [4602,4601,["l",2,9203],4601],
  [4603,2302,["l",2,4603],2301],
  [4605,4604,["l",2,9209],4604],
  [4611,4610,["l",2,9221],4610],
  [4614,4613,["l",2,9227],4613],
  [4620,4619,["l",2,9239],4619],
  [4621,2311,["l",2,4621],2310],
  [4621,4620,["l",2,9241],4620],
  [4629,4628,["l",2,9257],4628],
  [4637,2319,["l",2,4637],2318],
  [4639,2320,["l",2,4639],2319],
  [4639,4638,["l",2,9277],4638],
  [4641,4640,["l",2,9281],4640],
  [4642,4641,["l",2,9283],4641],
  [4643,2322,["l",2,4643],2321],
  [4647,4646,["l",2,9293],4646],
  [4649,2325,["l",2,4649],2324],
  [4651,2326,["l",2,4651],2325],
  [4656,4655,["l",2,9311],4655],
  [4657,2329,["l",2,4657],2328],
  [4660,4659,["l",2,9319],4659],
  [4662,4661,["l",2,9323],4661],
  [4663,2332,["l",2,4663],2331],
  [4669,4668,["l",2,9337],4668],
  [4671,4670,["l",2,9341],4670],
  [4672,4671,["l",2,9343],4671],
  [4673,2337,["l",2,4673],2336],
  [4675,4674,["l",2,9349],4674],
  [4679,2340,["l",2,4679],2339],[4681,4599,["l",5,8],4679]
    ,[4686,4685,["l",2,9371],4685],
  [4689,4688,["l",2,9377],4688],
  [4691,2346,["l",2,4691],2345],
  [4696,4695,["l",2,9391],4695],
  [4699,4698,["l",2,9397],4698],
  [4702,4701,["l",2,9403],4701],
  [4703,2352,["l",2,4703],2351],
  [4705,4704,["l",2,9409],4704],
  [4707,4706,["l",2,9413],4706],
  [4710,4709,["l",2,9419],4709],
  [4711,4710,["l",2,9421],4710],
  [4716,4715,["l",2,9431],4715],
  [4717,4716,["l",2,9433],4716],
  [4719,4718,["l",2,9437],4718],
  [4720,4719,["l",2,9439],4719],
  [4721,2361,["l",2,4721],2360],
  [4723,2362,["l",2,4723],2361],
  [4729,2365,["l",2,4729],2364],
  [4731,4730,["l",2,9461],4730],
  [4732,4731,["l",2,9463],4731],
  [4733,2367,["l",2,4733],2366],
  [4734,4733,["l",2,9467],4733],
  [4737,4736,["l",2,9473],4736],
  [4740,4739,["l",2,9479],4739],
  [4746,4745,["l",2,9491],4745],
  [4749,4748,["l",2,9497],4748],
  [4751,2376,["l",2,4751],2375],
  [4756,4755,["l",2,9511],4755],
  [4759,2380,["l",2,4759],2379],
  [4761,4760,["l",2,9521],4760],
  [4767,4766,["l",2,9533],4766],
  [4770,4769,["l",2,9539],4769],
  [4774,4773,["l",2,9547],4773],
  [4776,4775,["l",2,9551],4775],
  [4783,2392,["l",2,4783],2391],
  [4787,2394,["l",2,4787],2393],
  [4789,2395,["l",2,4789],2394],
  [4793,2397,["l",2,4793],2396],
  [4794,4793,["l",2,9587],4793],
  [4799,2400,["l",2,4799],2399],
  [4801,2401,["l",2,4801],2400],
  [4801,4800,["l",2,9601],4800],
  [4807,4806,["l",2,9613],4806],
  [4810,4809,["l",2,9619],4809],
  [4812,4811,["l",2,9623],4811],
  [4813,2407,["l",2,4813],2406],
  [4815,4814,["l",2,9629],4814],
  [4816,4815,["l",2,9631],4815],
  [4817,2409,["l",2,4817],2408],
  [4822,4821,["l",2,9643],4821],
  [4825,4824,["l",2,9649],4824],
  [4831,2416,["l",2,4831],2415],
  [4831,4830,["l",2,9661],4830],
  [4839,4838,["l",2,9677],4838],
  [4840,4839,["l",2,9679],4839],
  [4845,4844,["l",2,9689],4844],
  [4849,4848,["l",2,9697],4848],
  [4860,4859,["l",2,9719],4859],
  [4861,2431,["l",2,4861],2430],
  [4861,4860,["l",2,9721],4860],
  [4867,4866,["l",2,9733],4866],
  [4870,4869,["l",2,9739],4869],
  [4871,2436,["l",2,4871],2435],
  [4872,4871,["l",2,9743],4871],
  [4875,4874,["l",2,9749],4874],
  [4877,2439,["l",2,4877],2438],
  [4884,4883,["l",2,9767],4883],
  [4885,4884,["l",2,9769],4884],
  [4889,2445,["l",2,4889],2444],
  [4891,4890,["l",2,9781],4890],
  [4894,4893,["l",2,9787],4893],
  [4896,2610,["s",6,17],2456],
  [4896,4895,["l",2,9791],4895],
  [4902,4901,["l",2,9803],4901],
  [4903,2452,["l",2,4903],2451],
  [4906,4905,["l",2,9811],4905],
  [4909,2455,["l",2,4909],2454],
  [4909,4908,["l",2,9817],4908],
  [4915,4914,["l",2,9829],4914],
  [4917,4916,["l",2,9833],4916],
  [4919,2460,["l",2,4919],2459],
  [4920,4919,["l",2,9839],4919],
  [4926,4925,["l",2,9851],4925],
  [4929,4928,["l",2,9857],4928],
  [4930,2610,["s",6,17],2456],[4930,4896,["s",6,17],2456],
  [4930,4929,["l",2,9859],4929],
  [4931,2466,["l",2,4931],2465],
  [4933,2467,["l",2,4933],2466],
  [4936,4935,["l",2,9871],4935],
  [4937,2469,["l",2,4937],2468],
  [4942,4941,["l",2,9883],4941],
  [4943,2472,["l",2,4943],2471],
  [4944,4943,["l",2,9887],4943],
  [4951,2476,["l",2,4951],2475],
  [4951,4950,["l",2,9901],4950],
  [4954,4953,["l",2,9907],4953],
  [4957,2479,["l",2,4957],2478],
  [4962,4961,["l",2,9923],4961],
  [4965,4964,["l",2,9929],4964],
  [4966,4965,["l",2,9931],4965],
  [4967,2484,["l",2,4967],2483],
  [4969,2485,["l",2,4969],2484],
  [4971,4970,["l",2,9941],4970],
  [4973,2487,["l",2,4973],2486],
  [4975,4974,["l",2,9949],4974],
  [4984,4983,["l",2,9967],4983],
  [4987,2494,["l",2,4987],2493],
  [4987,4986,["l",2,9973],4986],
  [4993,2497,["l",2,4993],2496],
  [4999,2500,["l",2,4999],2499],
  [5003,2502,["l",2,5003],2501],
  [5009,2505,["l",2,5009],2504],
  [5011,2506,["l",2,5011],2505],
  [5021,2511,["l",2,5021],2510],
  [5023,2512,["l",2,5023],2511],
  [5039,2520,["l",2,5039],2519],
  [5051,2526,["l",2,5051],2525],
  [5059,2530,["l",2,5059],2529],
  [5077,2539,["l",2,5077],2538],
  [5081,2541,["l",2,5081],2540],
  [5087,2544,["l",2,5087],2543],
  [5099,2550,["l",2,5099],2549],
  [5101,2551,["l",2,5101],2550],
  [5107,2554,["l",2,5107],2553],
  [5113,2557,["l",2,5113],2556],
  [5119,2560,["l",2,5119],2559],
  [5147,2574,["l",2,5147],2573],
  [5153,2577,["l",2,5153],2576],
  [5167,2584,["l",2,5167],2583],
  [5171,2586,["l",2,5171],2585],
  [5179,2590,["l",2,5179],2589],
  [5189,2595,["l",2,5189],2594],
  [5197,2599,["l",2,5197],2598],
  [5209,2605,["l",2,5209],2604],
  [5227,2614,["l",2,5227],2613],
  [5231,2616,["l",2,5231],2615],
  [5233,2617,["l",2,5233],2616],
  [5237,2619,["l",2,5237],2618],
  [5261,2631,["l",2,5261],2630],
  [5273,2637,["l",2,5273],2636],
  [5279,2640,["l",2,5279],2639],
  [5281,2641,["l",2,5281],2640],
  [5297,2649,["l",2,5297],2648],
  [5303,2652,["l",2,5303],2651],
  [5309,2655,["l",2,5309],2654],
  [5323,2662,["l",2,5323],2661],
  [5333,2667,["l",2,5333],2666],
  [5347,2674,["l",2,5347],2673],
  [5351,2676,["l",2,5351],2675],
  [5381,2691,["l",2,5381],2690],
  [5387,2694,["l",2,5387],2693],
  [5393,2697,["l",2,5393],2696],
  [5399,2700,["l",2,5399],2699],
  [5407,2704,["l",2,5407],2703],
  [5413,2707,["l",2,5413],2706],
  [5417,2709,["l",2,5417],2708],
  [5419,2710,["l",2,5419],2709],
  [5431,2716,["l",2,5431],2715],
  [5437,2719,["l",2,5437],2718],
  [5441,2721,["l",2,5441],2720],
  [5443,2722,["l",2,5443],2721],
  [5449,2725,["l",2,5449],2724],
  [5471,2736,["l",2,5471],2735],
  [5477,2739,["l",2,5477],2738],
  [5479,2740,["l",2,5479],2739],
  [5483,2742,["l",2,5483],2741],
  [5501,2751,["l",2,5501],2750],
  [5503,2752,["l",2,5503],2751],
  [5507,2754,["l",2,5507],2753],
  [5519,2760,["l",2,5519],2759],
  [5521,2761,["l",2,5521],2760],
  [5527,2764,["l",2,5527],2763],
  [5531,2766,["l",2,5531],2765],
  [5557,2779,["l",2,5557],2778],
  [5563,2782,["l",2,5563],2781],
  [5569,2785,["l",2,5569],2784],
  [5573,2787,["l",2,5573],2786],
  [5581,2791,["l",2,5581],2790],
  [5591,2796,["l",2,5591],2795],
  [5623,2812,["l",2,5623],2811],
  [5639,2820,["l",2,5639],2819],
  [5641,2821,["l",2,5641],2820],
  [5647,2824,["l",2,5647],2823],
  [5651,2826,["l",2,5651],2825],
  [5653,2827,["l",2,5653],2826],
  [5657,2829,["l",2,5657],2828],
  [5659,2830,["l",2,5659],2829],
  [5669,2835,["l",2,5669],2834],
  [5683,2842,["l",2,5683],2841],
  [5689,2845,["l",2,5689],2844],
  [5693,2847,["l",2,5693],2846],
  [5701,2851,["l",2,5701],2850],
  [5711,2856,["l",2,5711],2855],
  [5717,2859,["l",2,5717],2858],
  [5737,2869,["l",2,5737],2868],
  [5741,2871,["l",2,5741],2870],
  [5743,2872,["l",2,5743],2871],
  [5749,2875,["l",2,5749],2874],
  [5779,2890,["l",2,5779],2889],
  [5783,2892,["l",2,5783],2891],
  [5791,2896,["l",2,5791],2895],
  [5801,2901,["l",2,5801],2900],
  [5807,2904,["l",2,5807],2903],
  [5813,2907,["l",2,5813],2906],
  [5821,2911,["l",2,5821],2910],
  [5827,2914,["l",2,5827],2913],
  [5839,2920,["l",2,5839],2919],
  [5843,2922,["l",2,5843],2921],
  [5849,2925,["l",2,5849],2924],
  [5851,2926,["l",2,5851],2925],
  [5857,2929,["l",2,5857],2928],
  [5861,2931,["l",2,5861],2930],
  [5867,2934,["l",2,5867],2933],
  [5869,2935,["l",2,5869],2934],
  [5879,2940,["l",2,5879],2939],
  [5881,2941,["l",2,5881],2940],
  [5897,2949,["l",2,5897],2948],
  [5903,2952,["l",2,5903],2951],
  [5923,2962,["l",2,5923],2961],
  [5927,2964,["l",2,5927],2963],
  [5939,2970,["l",2,5939],2969],
  [5953,2977,["l",2,5953],2976],
  [5981,2991,["l",2,5981],2990],
  [5987,2994,["l",2,5987],2993],
  [6007,3004,["l",2,6007],3003],
  [6011,3006,["l",2,6011],3005],
  [6029,3015,["l",2,6029],3014],
  [6037,3019,["l",2,6037],3018],
  [6043,3022,["l",2,6043],3021],
  [6047,3024,["l",2,6047],3023],
  [6053,3027,["l",2,6053],3026],
  [6067,3034,["l",2,6067],3033],
  [6073,3037,["l",2,6073],3036],
  [6079,3040,["l",2,6079],3039],
  [6089,3045,["l",2,6089],3044],
  [6091,3046,["l",2,6091],3045],
  [6101,3051,["l",2,6101],3050],
  [6113,3057,["l",2,6113],3056],
  [6121,3061,["l",2,6121],3060],
  [6131,3066,["l",2,6131],3065],
  [6133,3067,["l",2,6133],3066],
  [6143,3072,["l",2,6143],3071],
  [6151,3076,["l",2,6151],3075],
  [6163,3082,["l",2,6163],3081],
  [6173,3087,["l",2,6173],3086],
  [6197,3099,["l",2,6197],3098],
  [6199,3100,["l",2,6199],3099],
  [6203,3102,["l",2,6203],3101],
  [6211,3106,["l",2,6211],3105],
  [6217,3109,["l",2,6217],3108],
  [6221,3111,["l",2,6221],3110],
  [6229,3115,["l",2,6229],3114],
  [6247,3124,["l",2,6247],3123],
  [6257,3129,["l",2,6257],3128],
  [6263,3132,["l",2,6263],3131],
  [6269,3135,["l",2,6269],3134],
  [6271,3136,["l",2,6271],3135],
  [6277,3139,["l",2,6277],3138],
  [6287,3144,["l",2,6287],3143],
  [6299,3150,["l",2,6299],3149],
  [6301,3151,["l",2,6301],3150],
  [6311,3156,["l",2,6311],3155],
  [6317,3159,["l",2,6317],3158],
  [6323,3162,["l",2,6323],3161],
  [6329,3165,["l",2,6329],3164],
  [6337,3169,["l",2,6337],3168],
  [6343,3172,["l",2,6343],3171],
  [6353,3177,["l",2,6353],3176],
  [6359,3180,["l",2,6359],3179],
  [6361,3181,["l",2,6361],3180],
  [6367,3184,["l",2,6367],3183],
  [6373,3187,["l",2,6373],3186],
  [6379,3190,["l",2,6379],3189],
  [6389,3195,["l",2,6389],3194],
  [6397,3199,["l",2,6397],3198],
  [6421,3211,["l",2,6421],3210],
  [6427,3214,["l",2,6427],3213],
  [6449,3225,["l",2,6449],3224],
  [6451,3226,["l",2,6451],3225],
  [6469,3235,["l",2,6469],3234],
  [6473,3237,["l",2,6473],3236],
  [6481,3241,["l",2,6481],3240],
  [6491,3246,["l",2,6491],3245],
  [6521,3261,["l",2,6521],3260],
  [6529,3265,["l",2,6529],3264],
  [6547,3274,["l",2,6547],3273],
  [6551,3276,["l",2,6551],3275],
  [6552,4372,["s",16,3],3280],[6552,4380,["s",16,3],3280],
  [6553,3277,["l",2,6553],3276],
  [6558,4372,["s",16,3],3280],[6558,4380,["s",16,3],3280],
  [6558,6552,["s",16,3],3280],[6560,4921,["u",9,3],4921],
  [6563,3282,["l",2,6563],3281],
  [6564,4372,["s",16,3],3280],[6564,4380,["s",16,3],3280],
  [6564,6552,["s",16,3],3280],[6564,6558,["s",16,3],3280],
  [6564,6560,["u",9,3],4921],[6569,3285,["l",2,6569],3284]
    ,[6570,4372,["s",16,3],3280],
  [6570,4380,["s",16,3],3280],[6570,6552,["s",16,3],3280],
  [6570,6558,["s",16,3],3280],[6570,6564,["s",16,3],3280],
  [6571,3286,["l",2,6571],3285],
  [6577,3289,["l",2,6577],3288],
  [6581,3291,["l",2,6581],3290],
  [6599,3300,["l",2,6599],3299],
  [6607,3304,["l",2,6607],3303],
  [6619,3310,["l",2,6619],3309],
  [6637,3319,["l",2,6637],3318],
  [6653,3327,["l",2,6653],3326],
  [6659,3330,["l",2,6659],3329],
  [6661,3331,["l",2,6661],3330],
  [6673,3337,["l",2,6673],3336],
  [6679,3340,["l",2,6679],3339],
  [6689,3345,["l",2,6689],3344],
  [6691,3346,["l",2,6691],3345],
  [6701,3351,["l",2,6701],3350],
  [6703,3352,["l",2,6703],3351],
  [6709,3355,["l",2,6709],3354],
  [6719,3360,["l",2,6719],3359],
  [6733,3367,["l",2,6733],3366],
  [6737,3369,["l",2,6737],3368],
  [6761,3381,["l",2,6761],3380],
  [6763,3382,["l",2,6763],3381],
  [6779,3390,["l",2,6779],3389],
  [6781,3391,["l",2,6781],3390],
  [6791,3396,["l",2,6791],3395],
  [6793,3397,["l",2,6793],3396],
  [6803,3402,["l",2,6803],3401],
  [6823,3412,["l",2,6823],3411],
  [6827,3414,["l",2,6827],3413],
  [6829,3415,["l",2,6829],3414],
  [6833,3417,["l",2,6833],3416],
  [6840,3620,["s",6,19],3429],
  [6841,3421,["l",2,6841],3420],
  [6857,3429,["l",2,6857],3428],
  [6863,3432,["l",2,6863],3431],
  [6869,3435,["l",2,6869],3434],
  [6871,3436,["l",2,6871],3435],
  [6878,3620,["s",6,19],3429],[6878,6840,["s",6,19],3429],
  [6883,3442,["l",2,6883],3441],
  [6899,3450,["l",2,6899],3449],
  [6907,3454,["l",2,6907],3453],
  [6911,3456,["l",2,6911],3455],
  [6917,3459,["l",2,6917],3458],
  [6947,3474,["l",2,6947],3473],
  [6949,3475,["l",2,6949],3474],
  [6959,3480,["l",2,6959],3479],
  [6961,3481,["l",2,6961],3480],
  [6967,3484,["l",2,6967],3483],
  [6971,3486,["l",2,6971],3485],
  [6977,3489,["l",2,6977],3488],
  [6983,3492,["l",2,6983],3491],
  [6991,3496,["l",2,6991],3495],
  [6997,3499,["l",2,6997],3498],
  [7001,3501,["l",2,7001],3500],
  [7013,3507,["l",2,7013],3506],
  [7019,3510,["l",2,7019],3509],
  [7027,3514,["l",2,7027],3513],
  [7039,3520,["l",2,7039],3519],
  [7043,3522,["l",2,7043],3521],
  [7057,3529,["l",2,7057],3528],
  [7069,3535,["l",2,7069],3534],
  [7079,3540,["l",2,7079],3539],
  [7103,3552,["l",2,7103],3551],
  [7109,3555,["l",2,7109],3554],
  [7121,3561,["l",2,7121],3560],
  [7127,3564,["l",2,7127],3563],
  [7129,3565,["l",2,7129],3564],
  [7151,3576,["l",2,7151],3575],
  [7159,3580,["l",2,7159],3579],
  [7177,3589,["l",2,7177],3588],
  [7187,3594,["l",2,7187],3593],
  [7193,3597,["l",2,7193],3596],
  [7207,3604,["l",2,7207],3603],
  [7211,3606,["l",2,7211],3605],
  [7213,3607,["l",2,7213],3606],
  [7219,3610,["l",2,7219],3609],
  [7229,3615,["l",2,7229],3614],
  [7237,3619,["l",2,7237],3618],
  [7243,3622,["l",2,7243],3621],
  [7247,3624,["l",2,7247],3623],
  [7253,3627,["l",2,7253],3626],
  [7283,3642,["l",2,7283],3641],
  [7297,3649,["l",2,7297],3648],
  [7307,3654,["l",2,7307],3653],
  [7309,3655,["l",2,7309],3654],
  [7321,3661,["l",2,7321],3660],
  [7331,3666,["l",2,7331],3665],
  [7333,3667,["l",2,7333],3666],
  [7349,3675,["l",2,7349],3674],
  [7351,3676,["l",2,7351],3675],
  [7369,3685,["l",2,7369],3684],
  [7393,3697,["l",2,7393],3696],
  [7411,3706,["l",2,7411],3705],
  [7417,3709,["l",2,7417],3708],
  [7433,3717,["l",2,7433],3716],
  [7451,3726,["l",2,7451],3725],
  [7457,3729,["l",2,7457],3728],
  [7459,3730,["l",2,7459],3729],
  [7477,3739,["l",2,7477],3738],
  [7481,3741,["l",2,7481],3740],
  [7487,3744,["l",2,7487],3743],
  [7489,3745,["l",2,7489],3744],
  [7499,3750,["l",2,7499],3749],
  [7507,3754,["l",2,7507],3753],
  [7517,3759,["l",2,7517],3758],
  [7523,3762,["l",2,7523],3761],
  [7529,3765,["l",2,7529],3764],
  [7537,3769,["l",2,7537],3768],
  [7541,3771,["l",2,7541],3770],
  [7547,3774,["l",2,7547],3773],
  [7549,3775,["l",2,7549],3774],
  [7559,3780,["l",2,7559],3779],
  [7561,3781,["l",2,7561],3780],
  [7573,3787,["l",2,7573],3786],
  [7577,3789,["l",2,7577],3788],
  [7583,3792,["l",2,7583],3791],
  [7589,3795,["l",2,7589],3794],
  [7591,3796,["l",2,7591],3795],
  [7603,3802,["l",2,7603],3801],
  [7607,3804,["l",2,7607],3803],
  [7621,3811,["l",2,7621],3810],
  [7639,3820,["l",2,7639],3819],
  [7643,3822,["l",2,7643],3821],
  [7649,3825,["l",2,7649],3824],
  [7669,3835,["l",2,7669],3834],
  [7673,3837,["l",2,7673],3836],
  [7681,3841,["l",2,7681],3840],
  [7687,3844,["l",2,7687],3843],
  [7691,3846,["l",2,7691],3845],
  [7699,3850,["l",2,7699],3849],
  [7703,3852,["l",2,7703],3851],
  [7717,3859,["l",2,7717],3858],
  [7723,3862,["l",2,7723],3861],
  [7727,3864,["l",2,7727],3863],
  [7741,3871,["l",2,7741],3870],
  [7753,3877,["l",2,7753],3876],
  [7757,3879,["l",2,7757],3878],
  [7759,3880,["l",2,7759],3879],
  [7789,3895,["l",2,7789],3894],
  [7793,3897,["l",2,7793],3896],
  [7817,3909,["l",2,7817],3908],
  [7823,3912,["l",2,7823],3911],
  [7829,3915,["l",2,7829],3914],
  [7841,3921,["l",2,7841],3920],
  [7853,3927,["l",2,7853],3926],
  [7867,3934,["l",2,7867],3933],
  [7873,3937,["l",2,7873],3936],
  [7877,3939,["l",2,7877],3938],
  [7879,3940,["l",2,7879],3939],
  [7883,3942,["l",2,7883],3941],
  [7901,3951,["l",2,7901],3950],
  [7907,3954,["l",2,7907],3953],
  [7919,3960,["l",2,7919],3959],
  [7927,3964,["l",2,7927],3963],
  [7933,3967,["l",2,7933],3966],
  [7937,3969,["l",2,7937],3968],
  [7949,3975,["l",2,7949],3974],
  [7951,3976,["l",2,7951],3975],
  [7963,3982,["l",2,7963],3981],
  [7993,3997,["l",2,7993],3996],
  [8009,4005,["l",2,8009],4004],
  [8011,4006,["l",2,8011],4005],
  [8017,4009,["l",2,8017],4008],
  [8039,4020,["l",2,8039],4019],
  [8053,4027,["l",2,8053],4026],
  [8059,4030,["l",2,8059],4029],
  [8069,4035,["l",2,8069],4034],
  [8081,4041,["l",2,8081],4040],
  [8087,4044,["l",2,8087],4043],
  [8089,4045,["l",2,8089],4044],
  [8093,4047,["l",2,8093],4046],
  [8101,4051,["l",2,8101],4050],
  [8111,4056,["l",2,8111],4055],
  [8117,4059,["l",2,8117],4058],
  [8123,4062,["l",2,8123],4061],
  [8147,4074,["l",2,8147],4073],
  [8161,4081,["l",2,8161],4080],
  [8167,4084,["l",2,8167],4083],
  [8171,4086,["l",2,8171],4085],
  [8179,4090,["l",2,8179],4089],
  [8191,4096,["l",2,8191],4095],
  [8193,8191,["l",2,8192],4096],
  [8209,4105,["l",2,8209],4104],
  [8219,4110,["l",2,8219],4109],
  [8221,4111,["l",2,8221],4110],
  [8231,4116,["l",2,8231],4115],
  [8233,4117,["l",2,8233],4116],
  [8237,4119,["l",2,8237],4118],
  [8243,4122,["l",2,8243],4121],
  [8263,4132,["l",2,8263],4131],
  [8269,4135,["l",2,8269],4134],
  [8273,4137,["l",2,8273],4136],
  [8287,4144,["l",2,8287],4143],
  [8291,4146,["l",2,8291],4145],
  [8293,4147,["l",2,8293],4146],
  [8297,4149,["l",2,8297],4148],
  [8311,4156,["l",2,8311],4155],
  [8317,4159,["l",2,8317],4158],
  [8329,4165,["l",2,8329],4164],
  [8353,4177,["l",2,8353],4176],
  [8363,4182,["l",2,8363],4181],
  [8369,4185,["l",2,8369],4184],
  [8377,4189,["l",2,8377],4188],
  [8387,4194,["l",2,8387],4193],
  [8389,4195,["l",2,8389],4194],
  [8419,4210,["l",2,8419],4209],
  [8423,4212,["l",2,8423],4211],
  [8429,4215,["l",2,8429],4214],
  [8431,4216,["l",2,8431],4215],
  [8443,4222,["l",2,8443],4221],
  [8447,4224,["l",2,8447],4223],
  [8461,4231,["l",2,8461],4230],
  [8467,4234,["l",2,8467],4233],
  [8501,4251,["l",2,8501],4250],
  [8513,4257,["l",2,8513],4256],
  [8521,4261,["l",2,8521],4260],
  [8527,4264,["l",2,8527],4263],
  [8537,4269,["l",2,8537],4268],
  [8539,4270,["l",2,8539],4269],
  [8543,4272,["l",2,8543],4271],
  [8563,4282,["l",2,8563],4281],
  [8573,4287,["l",2,8573],4286],
  [8581,4291,["l",2,8581],4290],
  [8597,4299,["l",2,8597],4298],
  [8599,4300,["l",2,8599],4299],
  [8609,4305,["l",2,8609],4304],
  [8623,4312,["l",2,8623],4311],
  [8627,4314,["l",2,8627],4313],
  [8629,4315,["l",2,8629],4314],
  [8641,4321,["l",2,8641],4320],
  [8647,4324,["l",2,8647],4323],
  [8663,4332,["l",2,8663],4331],
  [8669,4335,["l",2,8669],4334],
  [8677,4339,["l",2,8677],4338],
  [8681,4341,["l",2,8681],4340],
  [8689,4345,["l",2,8689],4344],
  [8693,4347,["l",2,8693],4346],
  [8699,4350,["l",2,8699],4349],
  [8707,4354,["l",2,8707],4353],
  [8713,4357,["l",2,8713],4356],
  [8719,4360,["l",2,8719],4359],
  [8731,4366,["l",2,8731],4365],
  [8737,4369,["l",2,8737],4368],
  [8741,4371,["l",2,8741],4370],
  [8747,4374,["l",2,8747],4373],
  [8753,4377,["l",2,8753],4376],
  [8761,4381,["l",2,8761],4380],
  [8779,4390,["l",2,8779],4389],
  [8783,4392,["l",2,8783],4391],
  [8803,4402,["l",2,8803],4401],
  [8807,4404,["l",2,8807],4403],
  [8819,4410,["l",2,8819],4409],
  [8821,4411,["l",2,8821],4410],
  [8831,4416,["l",2,8831],4415],
  [8837,4419,["l",2,8837],4418],
  [8839,4420,["l",2,8839],4419],
  [8849,4425,["l",2,8849],4424],
  [8861,4431,["l",2,8861],4430],
  [8863,4432,["l",2,8863],4431],
  [8867,4434,["l",2,8867],4433],
  [8887,4444,["l",2,8887],4443],
  [8893,4447,["l",2,8893],4446],
  [8923,4462,["l",2,8923],4461],
  [8929,4465,["l",2,8929],4464],
  [8933,4467,["l",2,8933],4466],
  [8941,4471,["l",2,8941],4470],
  [8951,4476,["l",2,8951],4475],
  [8963,4482,["l",2,8963],4481],
  [8969,4485,["l",2,8969],4484],
  [8971,4486,["l",2,8971],4485],
  [8999,4500,["l",2,8999],4499],
  [9001,4501,["l",2,9001],4500],
  [9007,4504,["l",2,9007],4503],
  [9011,4506,["l",2,9011],4505],
  [9013,4507,["l",2,9013],4506],
  [9029,4515,["l",2,9029],4514],
  [9041,4521,["l",2,9041],4520],
  [9043,4522,["l",2,9043],4521],
  [9049,4525,["l",2,9049],4524],
  [9059,4530,["l",2,9059],4529],
  [9067,4534,["l",2,9067],4533],
  [9091,4546,["l",2,9091],4545],
  [9103,4552,["l",2,9103],4551],
  [9109,4555,["l",2,9109],4554],
  [9127,4564,["l",2,9127],4563],
  [9133,4567,["l",2,9133],4566],
  [9137,4569,["l",2,9137],4568],
  [9151,4576,["l",2,9151],4575],
  [9157,4579,["l",2,9157],4578],
  [9161,4581,["l",2,9161],4580],
  [9173,4587,["l",2,9173],4586],
  [9181,4591,["l",2,9181],4590],
  [9187,4594,["l",2,9187],4593],
  [9199,4600,["l",2,9199],4599],
  [9203,4602,["l",2,9203],4601],
  [9209,4605,["l",2,9209],4604],
  [9221,4611,["l",2,9221],4610],
  [9227,4614,["l",2,9227],4613],
  [9239,4620,["l",2,9239],4619],
  [9241,4621,["l",2,9241],4620],
  [9257,4629,["l",2,9257],4628],
  [9277,4639,["l",2,9277],4638],
  [9281,4641,["l",2,9281],4640],
  [9283,4642,["l",2,9283],4641],
  [9293,4647,["l",2,9293],4646],
  [9311,4656,["l",2,9311],4655],
  [9319,4660,["l",2,9319],4659],
  [9323,4662,["l",2,9323],4661],
  [9337,4669,["l",2,9337],4668],
  [9341,4671,["l",2,9341],4670],
  [9343,4672,["l",2,9343],4671],
  [9349,4675,["l",2,9349],4674],
  [9371,4686,["l",2,9371],4685],
  [9377,4689,["l",2,9377],4688],
  [9391,4696,["l",2,9391],4695],
  [9397,4699,["l",2,9397],4698],
  [9403,4702,["l",2,9403],4701],
  [9413,4707,["l",2,9413],4706],
  [9419,4710,["l",2,9419],4709],
  [9421,4711,["l",2,9421],4710],
  [9431,4716,["l",2,9431],4715],
  [9433,4717,["l",2,9433],4716],
  [9437,4719,["l",2,9437],4718],
  [9439,4720,["l",2,9439],4719],
  [9461,4731,["l",2,9461],4730],
  [9463,4732,["l",2,9463],4731],
  [9467,4734,["l",2,9467],4733],
  [9473,4737,["l",2,9473],4736],
  [9479,4740,["l",2,9479],4739],
  [9491,4746,["l",2,9491],4745],
  [9497,4749,["l",2,9497],4748],
  [9511,4756,["l",2,9511],4755],
  [9521,4761,["l",2,9521],4760],
  [9533,4767,["l",2,9533],4766],
  [9539,4770,["l",2,9539],4769],
  [9547,4774,["l",2,9547],4773],
  [9551,4776,["l",2,9551],4775],
  [9587,4794,["l",2,9587],4793],
  [9601,4801,["l",2,9601],4800],
  [9613,4807,["l",2,9613],4806],
  [9619,4810,["l",2,9619],4809],
  [9623,4812,["l",2,9623],4811],
  [9629,4815,["l",2,9629],4814],
  [9631,4816,["l",2,9631],4815],
  [9643,4822,["l",2,9643],4821],
  [9649,4825,["l",2,9649],4824],
  [9661,4831,["l",2,9661],4830],
  [9677,4839,["l",2,9677],4838],
  [9679,4840,["l",2,9679],4839],
  [9689,4845,["l",2,9689],4844],
  [9697,4849,["l",2,9697],4848],
  [9719,4860,["l",2,9719],4859],
  [9721,4861,["l",2,9721],4860],
  [9733,4867,["l",2,9733],4866],
  [9739,4870,["l",2,9739],4869],
  [9743,4872,["l",2,9743],4871],
  [9749,4875,["l",2,9749],4874],
  [9767,4884,["l",2,9767],4883],
  [9769,4885,["l",2,9769],4884],
  [9781,4891,["l",2,9781],4890],
  [9787,4894,["l",2,9787],4893],
  [9791,4896,["l",2,9791],4895],
  [9803,4902,["l",2,9803],4901],
  [9811,4906,["l",2,9811],4905],
  [9817,4909,["l",2,9817],4908],
  [9829,4915,["l",2,9829],4914],
  [9833,4917,["l",2,9833],4916],
  [9839,4920,["l",2,9839],4919],
  [9851,4926,["l",2,9851],4925],
  [9857,4929,["l",2,9857],4928],
  [9859,4930,["l",2,9859],4929],
  [9871,4936,["l",2,9871],4935],
  [9883,4942,["l",2,9883],4941],
  [9887,4944,["l",2,9887],4943],
  [9901,4951,["l",2,9901],4950],
  [9907,4954,["l",2,9907],4953],
  [9923,4962,["l",2,9923],4961],
  [9929,4965,["l",2,9929],4964],
  [9931,4966,["l",2,9931],4965],
  [9941,4971,["l",2,9941],4970],
  [9949,4975,["l",2,9949],4974],
  [9967,4984,["l",2,9967],4983],
  [9973,4987,["l",2,9973],4986] ];

RECOG.ListOfGroupsCrossCharRepsByBound :=
function(dimbound)
    local list, q, n;

    list:=[];
    for q in [4..2*dimbound+1] do
      if IsPrimePowerInt(q) then
        Add(list, ["l",2,q]);
      fi;
    od;

    for n in [3..LogInt(dimbound,2)] do
      for q in [2..RootInt(dimbound,2)] do
        if IsPrimePowerInt(q) and (q^n-q)/(q-1)-1 <= dimbound then
          Add(list, ["l",n,q]);
        fi;
      od;
    od;

    for n in [2..LogInt(dimbound,2)] do
      for q in [2..RootInt(dimbound,2)] do
        if IsPrimePowerInt(q) and (q mod 2 = 1) and (q^n-1)/2 <= dimbound then
          Add(list, ["s",2*n,q]);
        fi;
        if IsPrimePowerInt(q) and (q mod 2 = 0) and
            (q^n-1)*(q^n-q)/(2*(q+1)) <= dimbound then
          Add(list, ["s",2*n,q]);
        fi;
      od;
    od;

    for n in [3..LogInt(dimbound,2)] do
      for q in [2..RootInt(dimbound,2)] do
        if IsPrimePowerInt(q) and  (q^n-1)/(q+1) <= dimbound and
             [n,q]<>[3,2] then
          Add(list, ["u",n,q]);
        fi;
      od;
    od;

    for n in [4..LogInt(dimbound,2)] do
      if (2^n-1)*(2^(n-1)-1)/3 <= dimbound then
        Add(list, ["o+",2*n,2]);
      fi;
      if (3^n-1)*(3^(n-1)-1)/8 <= dimbound then
        Add(list, ["o+",2*n,3]);
      fi;
      for q in [4..RootInt(dimbound,2)] do
        if IsPrimePowerInt(q) and (q^n-1)*(q^(n-1)+q)/(q^2-1) <= dimbound then
          Add(list, ["o+",2*n,q]);
        fi;
      od;
    od;

    for n in [4..LogInt(dimbound,2)] do
      for q in [2..RootInt(dimbound,2)] do
        if IsPrimePowerInt(q) and (q^n+1)*(q^(n-1)-q)/(q^2-1)-1 <= dimbound then
          Add(list, ["o-",2*n,q]);
        fi;
      od;
    od;

    for n in [3..LogInt(dimbound,2)] do
      if (3^n-1)*(3^n-3)/8 <= dimbound then
        Add(list, ["o",2*n+1,3]);
      fi;
      for q in [5..RootInt(dimbound,2)] do
        if IsPrimePowerInt(q) and (q mod 2 = 1) and
            (q^(2*n)-1)/(q^2-1)-2 <= dimbound then
          Add(list, ["o",2*n+1,q]);
        fi;
      od;
    od;

    for n in [2..LogInt(dimbound,2)] do
      if (n mod 2 = 1) and (2^n-1)*RootInt(2^n/2) <= dimbound then
        Add(list, ["2B2",2^n]);
      fi;
    od;

    for n in [2..LogInt(dimbound,2)] do
      if (n mod 2 = 1) and (3^n-1)*3^n <= dimbound then
        Add(list, ["2G2",3^n]);
      fi;
    od;

    for q in [3..RootInt(dimbound,2)] do
      if IsPrimePowerInt(q) and (q mod 3 <> 0) and q^3-1 <= dimbound then
        Add(list, ["G2",q]);
      fi;
      if IsPrimePowerInt(q) and (q mod 3 = 0) and q^4+q^2 <= dimbound then
        Add(list, ["G2",q]);
      fi;
    od;

    for q in [2..RootInt(dimbound,2)] do
      if IsPrimePowerInt(q) and q^5-q^3+q-1 <= dimbound then
        Add(list, ["3D4",q]);
      fi;
    od;

    for n in [1..LogInt(dimbound,2)] do
      if (n mod 2 = 1) and (2^(5*n)-2^(4*n))*RootInt(2^n/2) <= dimbound then
        Add(list, ["2F4",2^n]);
      fi;
    od;

    for q in [2..RootInt(dimbound,2)] do
      if IsPrimePowerInt(q) and (q mod 2 = 1) and q^8+q^4-2 <= dimbound then
        Add(list, ["F4",q]);
      fi;
      if IsPrimePowerInt(q) and (q mod 2 = 0) and
         (q^3-1)*(q^8-q^7)/2 <= dimbound then
        Add(list, ["F4",q]);
      fi;
    od;

    for q in [2..RootInt(dimbound,2)] do
      if IsPrimePowerInt(q) and q^11-q^9 <= dimbound then
        Add(list, ["2E6",q]);
        Add(list, ["E6",q]);
      fi;
      if IsPrimePowerInt(q) and q^17-q^15 <= dimbound then
        Add(list, ["E7",q]);
      fi;
      if IsPrimePowerInt(q) and q^29-q^27 <= dimbound then
        Add(list, ["E8",q]);
      fi;
    od;

    # Note that the following are lower bounds for faithful *projective*
    # representations, because we recognise simple groups in their
    # projective representations!
    if dimbound >= 5 then Add(list,["M11"]); fi;
    if dimbound >= 6 then Add(list,["M12"]); fi;
    if dimbound >= 6 then Add(list,["M22"]); fi;
    if dimbound >= 11 then Add(list,["M23"]); fi;
    if dimbound >= 11 then Add(list,["M24"]); fi;
    if dimbound >= 20 then Add(list,["HS"]); fi;
    if dimbound >= 21 then Add(list,["McL"]); fi;
    if dimbound >= 22 then Add(list,["Co3"]); fi;
    if dimbound >= 22 then Add(list,["Co2"]); fi;
    if dimbound >= 24 then Add(list,["Co1"]); fi;
    if dimbound >= 6 then Add(list,["J2"]); fi;
    if dimbound >= 12 then Add(list,["Suz"]); fi;
    if dimbound >= 27 then Add(list,["Fi22"]); fi;
    if dimbound >= 253 then Add(list,["Fi23"]); fi;
    if dimbound >= 781 then Add(list,["Fi24'"]); fi;
    if dimbound >= 50 then Add(list,["He"]); fi;
    if dimbound >= 132 then Add(list,["HN"]); fi;
    if dimbound >= 248 then Add(list,["Th"]); fi;
    if dimbound >= 7 then Add(list,["J1"]); fi;
    if dimbound >= 9 then Add(list,["J3"]); fi;
    if dimbound >= 28 then Add(list,["Ru"]); fi;
    if dimbound >= 45 then Add(list,["O'N"]); fi;
    if dimbound >= 111 then Add(list,["Ly"]); fi;
    if dimbound >= 112 then Add(list,["J4"]); fi;
    if dimbound >= 4370 then Add(list,["B"]); fi;
    if dimbound >= 196882 then Add(list,["M"]); fi;

    return list;

end;

# The following function was used to convert
#  RECOG.TableOfTwoLargestGroupElementOrders
# into
#  RECOG.TableOfTwoLargestGroupElementOrdersWithDegreeBound:
# Then the following is used:
#   PrintTo("tmp.g",
#           "RECOG.TableOfTwoLargestGroupElementOrdersWithDegreeBound := \n",
#           RECOG.TableOfTwoLargestGroupElementOrders,";\n");
# Then copy the output "tmp.g" to this file.

RECOG.AddInformationAboutDegree := function()
  local i,j,l,ll,new,t;
  l := [];
  t := RECOG.TableOfTwoLargestGroupElementOrders;
  for i in [1..5000] do
      Print(i," \c");
      ll := Set(RECOG.ListOfGroupsCrossCharRepsByBound(i));
      new := Difference(ll,l);
      for j in [1..Length(t)] do
          if t[j][3] in new then
              Add(t[j],i);
          fi;
      od;
      l := ll;
  od;
  Print("\n");
end;

RECOG.MakePSL2Hint := function( name, G )
  local d,defchar,f,p,q;
  f := DefaultFieldOfMatrixGroup(G);
  q := Size(f);
  p := Characteristic(f);
  d := DimensionOfMatrixGroup(G);
  defchar := Factors(name[3])[1];
  if p = defchar then return fail; fi;
  Info(InfoRecog,2,"Making hint for group ",name,"...");
  # we are in cross characteristic.
  # to be made better...
  return rec( elordersstart := [defchar], numberrandgens := 1, tries := 10,
              triesforgens := 3*(name[3]+1),
              orblenlimit := 3*(name[3]+1) );
end;

FindHomMethodsProjective.TwoLargeElOrders := function(ri,G)
  local d,epsilon,hint,i,max,name,namecat,nr,o,p,res,tab;
  d := ri!.dimension;
  epsilon := 1/100;
  #nr := 16 * LogInt(3*d,2) * LogInt(2/epsilon,2);
  # A cheat: 16 is theoretically OK for up to degree 5000,
  # for smaller degrees, our 4 here is also OK.
  nr := 4 * LogInt(3*d,2) * LogInt(2/epsilon,2);
  Info(InfoRecog,2,"Looking at the projective orders of ",nr,
       " random elements...");
  max := [0,0];
  ri!.projelorders := [];
  for i in [1..nr] do
      o := ProjectiveOrder(PseudoRandom(G))[1];
      Add(ri!.projelorders,o);
      # FIXME: Changed
      #if o > 3*d then
      if o > 10000 then   # we would not know about such orders anyway...
          #Info(InfoRecog,2,"Saw large projective order ",o," > 3*",d,
          Info(InfoRecog,2,"Saw large projective order ",o," > 10000,",
               " giving up.");
          return false;
      fi;
      if o > max[1] then
          max[2] := max[1];
          max[1] := o;
      elif o < max[1] and o > max[2] then
          max[2] := o;
      fi;
  od;
  Info(InfoRecog,2,"Two largest projective orders found: ",max,".");
  tab := RECOG.TableOfTwoLargestGroupElementOrdersWithDegreeBound;
  p := PositionSorted(tab,max);
  if p > Length(tab) or tab[p]{[1..2]} <> max then
      Info(InfoRecog,2,"Those orders are not in table, giving up.");
      return false;
  fi;
  # Try all possibilities:
  while p <= Length(tab) and tab[p]{[1..2]} = max do
      if tab[p][4] <= d then   # ruled out by degree bound
          name := tab[p][3];
          Info(InfoRecog,2,"Trying ",name);
          if name[1] = "l" then  # Handle PSL specially
              if name[2] = 2 then
                  hint := RECOG.MakePSL2Hint(name,G);
                  if hint <> fail then
                      res := DoHintedLowIndex(ri,G,hint);
                  else   # we use Pete Brooksbank's methods
                      return SLCR.FindHom(ri,G,2,name[3]);
                  fi;
              else
                  return SLCR.FindHom(ri,G,name[2],name[3]);
              fi;
          else
              if Length(name) = 3 then
                  namecat := Concatenation(name[1],String(name[2]),
                                           "(",String(name[3]),")");
              else
                  namecat := name[1];
              fi;
              res := LookupHintForSimple(ri,G,namecat);
          fi;
          if res = true then return true; fi;
      else
          Info(InfoRecog,2,"Cannot be ",tab[p][3]," in degree ",d,".");
      fi;
      p := p + 1;
  od;
  Info(InfoRecog,2,"Did not succeed with hints, giving up...");
  return fail;
end;

#AddMethod( FindHomDbMatrix, FindHomMethodsMatrix.TwoLargeElOrds,
#           950, "TwoLargeElOrds",
#           "look at two largest element orders" );
